/*
 * Created on 30.12.2004
 *
 */
package construction;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;

import javax.media.j3d.Bounds;
import javax.media.j3d.Group;
import javax.media.j3d.Node;
import javax.media.j3d.Texture;
import javax.media.j3d.Transform3D;
import javax.media.j3d.TransformGroup;
import javax.vecmath.Point3f;
import javax.vecmath.Vector3f;

/**
 * Ein Block ist ein n-eckiger K&ouml;rper, wobei n eine gerade, ganzzahlige
 * Zahl gr&ouml;sser 5 ist. Ein Block besteht aus einem Boden und einer Decke.
 * Boden und Decke sind zwei ebene, nicht-konvexe Polygone. Beide Polygone haben
 * die selbe Anzahl an Ecken. Zwischen den jeweiligen Kanten der beiden Polygone
 * werden Vierecke aufgespannt, wobei alle vier Ecken in der selben Ebene
 * liegen. Die Seiten eines Blocks schneiden sich nicht.
 *
 * Ein Block kann positiv oder negativ sein. Wenn ein Block positiv ist, so kann
 * er von Au&szlig;en betrachtet werden. Das Innere des Blocks ist also nicht
 * sichtbar. Befindet man sich jedoch innerhalb des Blocks, kann man nach
 * Au&szlig;en blicken. Der Block selbst ist dann nicht sichtbar.
 * 
 * Ist ein Block negativ, so kann man von Au&szlig;en in ihn hineinblicken.
 * Befindet man sich innerhalb des Blocks, so sieht man nur den Block. Alles
 * au&szlig;erhalb des Blocks ist nicht sichtbar.
 * 
 * Mehrere Bl&ouml;cke k&ouml;nnen miteinander verbunden werden, wobei sie
 * jedoch unterschiedlich ausgerichtete (positiv/negativ) sein m&uuml;ssen.
 * Auf diese Weise ist es z.B. m&ouml;glich in einen positiven Block ein
 * Loch zu schneiden, indem er mit einem negativen Block verbunden wird.
 * 
 * Durch diese Methodik wird eine geringe CSG (Constructive Solid Geometry)
 * Funktionalit&auml;t erm&ouml;glicht.
 * 
 * @author Martin Hedler
 * @version 1.0 (2005/01/31)
 */
public class Block extends Group
{
    /**
     * Dieser Iterator erm&ouml;glicht den direkten Zugriff auf die Planes.
     * Allerding ist die {@linkplain Iterator#remove()} Methode nicht
     * implementiert bzw. hat keine Wirkung.
     */
    private class PlaneIterator implements Iterator
    {
        /** Index der aktuellen Plane. */
        private int index;

        /**
         * Initialisiert den Iterator mit der planeList des Blocks.
         * 
         * @param planeList Liste, die die TransformedPlanes enth&auml;lt.
         */
        public PlaneIterator()
        {
            index = 0;
        }

        /**
         * Diese Funktion hat keine Auswirkung auf die Elemente des Iterators.
         * 
         * @throws UnsupportedOperationException Wenn die Methode aufgerufen
         *         wird.
         */
        public void remove() throws UnsupportedOperationException
        {
            throw new UnsupportedOperationException();
        }

        /* (non-Javadoc)
         * @see java.util.Iterator#hasNext()
         */
        public boolean hasNext()
        {
            return index < planeCount;
        }

        /* (non-Javadoc)
         * @see java.util.Iterator#next()
         */
        public Object next()
        {
            try
            {
                Plane plane =
                    (Plane) ((TransformGroup) getChild(index)).getChild(0);
                ++index;
                return plane;
            } catch (IndexOutOfBoundsException e)
            {
                e.printStackTrace();
                throw new NoSuchElementException("Keine Plane gefunden.");
            }
        }
    }

    /** Der Block ist positiv. */
    public static final int POSITIVE = 1;

    /** Der Block ist negativ. */
    public static final int NEGATIVE = -1;

    /** Anzahl der Planes in diesem Block*/ 
    private int planeCount;

    /** Speichert den Volumenstatus (POSITIVE oder NEGATIVE) des Blocks. */
    private int volumeState;

    /**
     * Legt einen leeren Block ohne Geometrie (Boden, Decke, W&auml;nde) an.
     */
    private Block()
    {
        planeCount = 0;
    }

    /**
     * Legt einen Block mit Boden, Decke und W&auml;nden an. Ein Block ist
     * definiert durch die Eckpunkte seines Bodens und seiner Decke. Die
     * Richtung in der die Eckpunkte bei der Definition angeordnet sind, muss
     * bei beiden Punktlisten &uuml;bereinstimmen, also entweder beide im oder
     * gegen den Uhrzeigersinn angeordnet sein. Sind die Punkte gegen den
     * Uhrzeigersinn angeordnet, ist der Block negativ, sonst positiv. Diese
     * Definition gilt, wenn der Block von oben betrachtet wird.
     * <p />
     * Boden und Decke k&uuml;nnen unterschiedlich geform sein, jedoch
     * d&uuml;rfen sich die beiden Polygone nicht schneiden, ebensowenig wie die
     * W&auml;nde, die zwischen ihnen aufgespannt werden. Die W&auml;nde sind
     * Vierecke, deren vier Punkte alle in der selben Ebene liegen m&uuml;ssen.
     * 
     * @param floorPoint Punktarray, welche das Bodenpolygon beschreibt.
     * @param ceilingPoint Punktarray, welche das Deckenpolygon beschreibt.
     * @throws IllegalArgumentException Wenn die Anzahl der Punkte in den beiden
     *         Listen nicht &uuml;bereinstimmt.
     */
    public Block(Point3f[] floorPoint, Point3f[] ceilingPoint)
        throws IllegalArgumentException
    {
        this();
        this.setPoints(floorPoint, ceilingPoint);
    }

    /**
     * Gibt einen Iterator zur&uuml;ck, der alle Planes des Blocks enth&auml;lt.
     * 
     * @return Der Iterator.
     */
    public Iterator getPlanes()
    {
        return new PlaneIterator();
    }

    /**
     * Gibt die Plane mit dem &uuml;bergebenen Index zur&uuml;ck. Die Plane mit
     * dem Index 0 ist der Boden. Die Plane mit dem Index 1 ist die Decke. Alle
     * nachfolgenden Planes sind W&auml;nde.
     * 
     * @param index Index der gew&uuml;ndschten Plane.
     * @return Die Plane mit dem angegebenen Index.
     * @throws IndexOutOfBoundsException Wenn der Index ausserhalb des
     *         zul&auml;ssigen Bereichs ist.
     */
    public Plane getPlane(int index) throws IndexOutOfBoundsException
    {
        return (Plane) ((TransformGroup) this.getChild(index)).getChild(0);
    }

    /**
     * Gibt den Volumenstatus des Blocks zur&uuml;ck.
     * 
     * @return {@linkplain #POSITIVE}, wenn der Block von Au&szlig;en sichtbar
     *         oder
     *         {@linkplain #NEGATIVE}, wenn der Block von Innen sichtbar ist.
     */
    public int getVolumeState()
    {
        return this.volumeState;
    }

    /**
     * Legt Boden, Decke und W&auml;nde an. Dies geschieht durch Definition
     * der Eckpunkte des Bodens und der Decke. Die Richtung in der die
     * Eckpunkte bei der Definition angeordnet sind, muss bei beiden Punktlisten
     * &uuml;bereinstimmen, also entweder beide im oder gegen den Uhrzeigersinn
     * angeordnet sein. Sind die Punkte gegen den Uhrzeigersinn angeordnet, ist
     * der Block negativ, sonst positiv. Diese Definition gilt, wenn der Block
     * von oben betrachtet wird.
     * <p />
     * Boden und Decke k&uuml;nnen unterschiedlich geformt sein, jedoch
     * d&uuml;rfen sich die beiden Polygone nicht schneiden, ebensowenig wie die
     * W&auml;nde, die zwischen ihnen aufgespannt werden. Die W&auml;nde sind
     * Vierecke, deren vier Punkte alle in der selben Ebene liegen m&uuml;ssen.
     * 
     * @param floorPoint Punktarray, welche das Bodenpolygon beschreibt.
     * @param ceilingPoint Punktarray, welche das Deckenpolygon beschreibt.
     * @throws IllegalArgumentException Wenn die Anzahl der Punkte in den beiden
     *         Listen nicht &uuml;bereinstimmt.
     */
    private void setPoints(Point3f[] floorPoint, Point3f[] ceilingPoint)
        throws
            IllegalArgumentException
    {
        if (floorPoint.length != ceilingPoint.length)
            throw new IllegalArgumentException("Punktlisten sind unterschiedlich lang.");

        int cornerCount = floorPoint.length;

        Transform3D trans = new Transform3D();
        Point3f[] localePoint = new Point3f[cornerCount];

        TransformGroup currTrGrp;

        // Boden und Decke anlegen
        localize(floorPoint, localePoint, trans);
        currTrGrp = new TransformGroup(trans);
        currTrGrp.addChild(new Plane(localePoint));
        this.addChild(currTrGrp);

        localize(ceilingPoint, localePoint, trans);
        Collections.reverse(Arrays.asList(localePoint));
        currTrGrp = new TransformGroup(trans);
        currTrGrp.addChild(new Plane(localePoint));
        this.addChild(currTrGrp);

        Point3f[] wallPoint =
            new Point3f[] {
                new Point3f(),
                new Point3f(),
                new Point3f(),
                new Point3f()};

        // Variablen zur Ermittlung des Volumenstatus
        boolean ceilingAllSame = true;
        boolean floorAllSame = true;
        boolean ceilingDirectionPositiv = true;
        boolean floorDirectionPositiv = true;
        Vector3f wallVector = new Vector3f();
        Vector3f floorNormal = new Vector3f();
        Vector3f ceilingNormal = new Vector3f();
        ((Plane) ((TransformGroup) this.getChild(0)).getChild(0)).getNormal(
            floorNormal);
        ((Plane) ((TransformGroup) this.getChild(1)).getChild(0)).getNormal(
            ceilingNormal);

        // Waende erstellen
        for (int i = 0; i < cornerCount; ++i)
        {
            int j = (i + 1 < cornerCount) ? i + 1 : 0;

            wallPoint[0].set(floorPoint[i]);
            wallPoint[1].set(ceilingPoint[i]);
            wallPoint[2].set(ceilingPoint[j]);
            wallPoint[3].set(floorPoint[j]);

            // Algorithmus zur Feststellung des Volumenstatus - Start
            wallVector.sub(floorPoint[i], ceilingPoint[i]);
            if (i == 0)
            {
                floorDirectionPositiv = floorNormal.dot(wallVector) >= 0;
                ceilingDirectionPositiv = ceilingNormal.dot(wallVector) >= 0;
            } else
            {
                if (floorDirectionPositiv
                    != (floorNormal.dot(wallVector) >= 0))
                {
                    floorAllSame = false;
                }
                if (ceilingDirectionPositiv
                    != (ceilingNormal.dot(wallVector) >= 0))
                {
                    ceilingAllSame = false;
                }
            }
            // Algorithmus zur Feststellung des Volumenstatuses - Ende

            localize(wallPoint, wallPoint, trans);

            currTrGrp = new TransformGroup(trans);
            currTrGrp.addChild(new Plane(wallPoint));
            //planeList.add(currTrGrp);
            this.addChild(currTrGrp);
        } // for (int i = 0; i < this.cornerCount; ++i)

        // Algorithmus zur Feststellung des Volumenstatuses - Start
        // Auswertung
        assert floorAllSame || ceilingAllSame;
        if (floorAllSame)
        {
            if (floorDirectionPositiv)
                volumeState = POSITIVE;
            else
                volumeState = NEGATIVE;
        } else
            if (ceilingAllSame)
            {
                if (ceilingDirectionPositiv)
                    volumeState = NEGATIVE;
                else
                    volumeState = POSITIVE;
            }
        // Algorithmus zur Feststellung des Volumenstatuses - Ende
        
        this.planeCount = this.numChildren();
    }

    /**
     * Setzt alle Texturen des Blocks. Alle bereits gesetzten Texturen werden
     * &uuml;berschrieben. Um die Texturen f&uuml;r einzelne Planes zu setzen,
     * k&ouml;nnen die Funktionen {@linkplain #getPlanes()} und
     * {@linkplain #getPlane(int)} verwendet werden.
     * 
     * @param texture Array von Texturen. Die Anzahl der Texturen im Array
     *           bestimmt f&uuml;r was die Texturen verwendet werden:
     *           <ul>
     *              <li>Befindet sich nur eine Textur im Array, so wird die
     *                  Textur f&uuml;r alle W&auml;nde, den Boden und die
     *                  Decke verwendet.</li>
     *              <li>Befinden sich zwei Texturen im Array, so wird die erste
     *                  Textur f&uuml;r den Boden und die Decke verwendet und
     *                  die zweite Textur f&uuml;r alle W&auml;nde.</li>
     *              <li>Bei drei Texturen, wird die erste Textur f&uuml;r den
     *                  Boden, die zweite Textur f&uuml;r die Decke und die
     *                  dritte Textur f&uuml;r alle W&auml;nde verwendet.</li>
     *           </ul>
     *           Bei mehr als drei Texturen, findet der dritte Fall Anwendung
     *           und alle weiteren Texturen werden ignoriert.
     * @see #getPlanes()
     * @see #getPlane(int)
     */
    public void setPlaneTexture(Texture[] texture)
    {
        Texture wallTexture, floorTexture, ceilingTexture;

        // Nur eine Textur enthalten?
        if (texture.length == 1)
        {
            wallTexture = floorTexture = ceilingTexture = texture[0];
        } else
            // Zwei Texturen enthalten?
            if (texture.length == 2)
            {
                floorTexture = ceilingTexture = texture[0];
                wallTexture = texture[1];
            } else
                // Mindestens drei Texturen enthalten?
                if (texture.length >= 3)
                {
                    floorTexture = texture[0];
                    ceilingTexture = texture[1];
                    wallTexture = texture[2];
                } else
                {
                    wallTexture = floorTexture = ceilingTexture = null;
                }

        if (floorTexture != null)
        {
            ((Plane) ((TransformGroup) this.getChild(0)).getChild(0))
                .getAppearance()
                .setTexture(floorTexture);
        }
        if (ceilingTexture != null)
        {
            ((Plane) ((TransformGroup) this.getChild(1)).getChild(0))
                .getAppearance()
                .setTexture(ceilingTexture);
        }
        if (wallTexture != null)
            for (int i = 2; i < this.planeCount; ++i)
            {
                ((Plane) ((TransformGroup) this.getChild(i)).getChild(0))
                    .getAppearance()
                    .setTexture(wallTexture);
            }
    }

    /**
     * Verbindet zwei Bl&ouml;cke miteinander. Zwei Bl&ouml;cke werden dann
     * miteinander verbunden, wenn mindestens eine Plane dieses Blocks mit
     * mindestens einer Plane von <code>other</code> in Kontakt steht, d.h. dass
     * die beiden Planes parallel zueinander sind, in der selben Ebene
     * liegen und sich &uuml;berschneiden.</br>
     * Steht keine Plane mit einer anderen in Kontakt, kommt es nicht zu einer
     * Verbindung.
     * 
     * @param other Block, mit dem dieser Block verbunden werden soll.
     * @param otherTransform Transformation bez&uuml;glich des lokalen
     *        Koordinatensystems von <code>this</code>, welche auf
     *        <code>other</code> angewendet wird.
     * @return <code>true</code>, wenn die beiden Bl&ouml;cke miteinander
     *         verbunden wurden, sonst <code>false</code>. 
     */
    public boolean connect(Block other, Transform3D otherTransform)
        throws IllegalArgumentException
    {
        // Liste, in der die Planes gespeichert werden, die mit anderen in Kontakt sind.
        ArrayList connectedPlaneList = new ArrayList();

        // Status der beiden Sektoren ermitteln
        boolean flags =
            this.isContact(other, otherTransform, connectedPlaneList);

        // Gibt es keinen Kontakt?
        if (flags == false)
        {
            return false;
        }

        // Wenn es Verbindungen gibt, dann jetzt verbinden.
        //Plane thisPlane, otherPlane;
        for (Iterator it = connectedPlaneList.iterator(); it.hasNext();)
        {
            ContactPlanes cP = ((ContactPlanes) it.next());
            cP.thisPlane.subIntersection(
                cP.otherPlane,
                cP.otherToThisTransform);
        }

        return true;
    }

    /**
     * Ermittelt ob dieser Block mit dem &uuml;bergebenen Block in Kontakt
     * steht.
     * 
     * @param other Der zu &uuml;berpr&uuml;fende Block.
     * @param otherTransform Transformation bez&uuml;glich des lokalen
     *        Koordinatensystems von <code>this</code>, welche auf
     *        <code>other</code> angewendet wird.
     * @return <code>true</code>, wenn die beide Bl&ouml;cke miteinander in
     *         Kontakt stehen, sonst <code>false</code>.
     */
    public boolean isContact(Block other, Transform3D otherTransform)
    {
        // Liste, in welcher die sich beruehrenden Planes gespeichert werden.
        ArrayList planeContactList = new ArrayList();

        return isContact(other, otherTransform, planeContactList);
    }

    /**
     * Gibt die Anzahl der Planes des Blocks zur&uuml;ck.
     */
    public int numPlanes()
    {
        return planeCount;
    }

    /**
     * Ermittelt ob dieser Block mit dem &uuml;bergebenen Block in Kontakt
     * steht.
     * 
     * @param other Der zu &uuml;berpr&uuml;fende Block.
     * @param otherTransform Transformation bez&uuml;glich des lokalen
     *        Koordinatensystems von <code>this</code>, welche auf
     *        <code>other</code> angewendet wird.
     * @param connectedPlaneList Liste, in der die Planes gespeichert werden,
     *        die mit anderen in Kontakt stehen. Die Planepaare werden jeweil in
     *        einem {@linkplain ContactPlanes} Objekt gespeichert.
     * @return <code>true</code>, wenn die beide Bl&ouml;cke miteinander in
     *         Kontakt stehen, sonst <code>false</code>.
     */
    private boolean isContact(
        Block other,
        Transform3D otherTransform,
        List connectedPlaneList)
    {
        // Bounds ueberpruefen
        Bounds thisBounds = (Bounds) this.getBounds().clone();
        Bounds otherBounds = (Bounds) other.getBounds().clone();

        // Bounds transformieren
        //thisBounds.transform(thisTransform);
        otherBounds.transform(otherTransform);

        // Schneiden sich die Bounds?
        if (thisBounds.intersect(otherBounds))
        {
            /* connectedPlaneList:
             * Liste, in der die Planes gespeichert werden, die mit anderen
             * in Kontakt stehen.
             */
            connectedPlaneList.clear();

            // Jede Plane mit jeder pruefen
            for (int j = 0; j < other.planeCount; ++j)
            {
                for (int i = 0; i < this.planeCount; ++i)
                {
                    if (other.getChild(j) == this.getChild(i))
                        continue;

                    Transform3D thisPlaneTransform = new Transform3D();
                    ((TransformGroup) this.getChild(i)).getTransform(
                        thisPlaneTransform);
                    Transform3D otherPlaneTransform = new Transform3D();
                    ((TransformGroup) other.getChild(j)).getTransform(
                        otherPlaneTransform);

                    otherPlaneTransform.mul(
                        otherTransform,
                        otherPlaneTransform);
                    otherPlaneTransform.mulInverse(thisPlaneTransform);

                    int state =
                        (
                            (Plane)
                                (
                                    (TransformGroup) this.getChild(
                                        i)).getChild(
                                0)).getState(
                            (Plane)
                                (
                                    (TransformGroup) other.getChild(
                                        j)).getChild(
                                0),
                            otherPlaneTransform);

                    if (((state & Plane.IN) == Plane.IN
                        || (state & Plane.INCLUDING) == Plane.INCLUDING
                        || (state & Plane.EQUAL) == Plane.EQUAL
                        || (state & Plane.INTERSECTED) == Plane.INTERSECTED)
                        && (state & Plane.SAME_DIRECTION) == 0)
                    {
                        // Kontakt vorhanden -> speichern
                        ContactPlanes cP = new ContactPlanes();
                        cP.thisPlane =
                            (Plane)
                                (
                                    (TransformGroup) this.getChild(
                                        i)).getChild(
                                0);
                        cP.otherPlane =
                            (Plane)
                                (
                                    (TransformGroup) other.getChild(
                                        j)).getChild(
                                0);
                        cP.otherToThisTransform = otherPlaneTransform;
                        connectedPlaneList.add(cP);
                    }
                } // for (int i = 0; i < this.planeGroup.numChildren(); ++i)
            } // for (int j = 0; j < other.planeGroup.numChildren(); ++j)

            // Haben sich Planes beruehrt?
            if (connectedPlaneList.size() > 0)
            {
                return true;
            }
        } // if (thisBounds.intersect(otherBounds))

        return false;
    }

    /**
     * Ermittelt dem Mittelpunkt zu einer gegebenen Punktwolke und gibt die
     * Transformation zu diesem Mittelpunkt zur&uuml;ck. Au&szlig;erdem werden
     * die &uuml;bergebenen Punkte kopiert und transformiert.
     * 
     * @param srcPoint Array von Punkten, deren Mittelpunkt ermittelt werden
     *        soll.
     * @param destPoint Array in welches die transformierten Punkte geschrieben
     *        werden.
     * @param trans In dieses Objekt wird die Transformation geschrieben.
     */
    private void localize(
        Point3f[] srcPoint,
        Point3f[] destPoint,
        Transform3D trans)
    {
        Point3f min = new Point3f(srcPoint[0]), max = new Point3f(srcPoint[0]);

        for (int i = 1; i < srcPoint.length; ++i)
        {
            min.x = Math.min(srcPoint[i].x, min.x);
            min.y = Math.min(srcPoint[i].y, min.y);
            min.z = Math.min(srcPoint[i].z, min.z);
            max.x = Math.max(srcPoint[i].x, max.x);
            max.y = Math.max(srcPoint[i].y, max.y);
            max.z = Math.max(srcPoint[i].z, max.z);
        }

        Vector3f transVec = new Vector3f(min);
        transVec.add(max);
        transVec.scale(-0.5f);

        trans.set(transVec);

        for (int i = 0; i < srcPoint.length; ++i)
        {
            try
            {
                trans.transform(srcPoint[i], destPoint[i]);
            } catch (NullPointerException nE)
            {
                if (destPoint[i] == null)
                {
                    destPoint[i] = new Point3f();
                    trans.transform(srcPoint[i], destPoint[i]);
                } else
                    throw nE;
            }
        }

        trans.invert();
    }

    /**
     * Ein Objekt dieser Klasse repr&auml;sentiert ein Planepaar, welches
     * miteinander in Kontakt steht.
     */
    private class ContactPlanes
    {
        public Plane thisPlane, otherPlane;
        public Transform3D otherToThisTransform;
    }

    /* (non-Javadoc)
     * @see javax.media.j3d.Node#cloneNode(boolean)
     */
    public Node cloneNode(boolean forceDuplicate)
    {
        Block newBlock = new Block();
        newBlock.duplicateNode(this, forceDuplicate);
        return newBlock;
    }

    /* (non-Javadoc)
     * @see javax.media.j3d.Node#duplicateNode(javax.media.j3d.Node, boolean)
     */
    public void duplicateNode(Node orgNode, boolean forceDuplicate)
    {
        super.duplicateNode(orgNode, forceDuplicate);

        Block orgBlock = (Block) orgNode;

        // Nicht-Szenegraph Daten kopieren
        this.volumeState = orgBlock.volumeState;
        this.planeCount = orgBlock.planeCount;
    }
}
