/*
 * Created on 25.01.2005
 *
 */
package construction;

import javax.media.j3d.Group;
import javax.media.j3d.Node;
import javax.media.j3d.Texture;
import javax.media.j3d.TextureAttributes;
import javax.media.j3d.Transform3D;
import javax.media.j3d.TransformGroup;
import javax.vecmath.Point3f;
import javax.vecmath.Vector3f;

/**
 * Erstellt eine gerade Treppe.
 * 
 * @author Martin Hedler
 * @version 0.9 (2005/01/31)
 * 
 */
public class LinearStair extends Group
{
    /**
     * Erstellt ein Objekt dieser Klasse, jedoch ohne Geometrie. Diese muss mit
     * der Methode {@linkplain #set(float, int, float, float, float)} erstellt
     * werden.
     */
    public LinearStair()
    {
        super();

        stepCount = 0;
    }

    /**
     * Erstellt die Geometrie der Treppe.
     * 
     * @param width Die Breite der Treppe.
     * @param stepCount Die Anzahl der Stufen.
     * @param stepHeight Die H&ouml;he einer einzelnen Stufe.
     * @param stepDeep Die Tiefe einer einzelnen Stufe.
     * @param startHeight Die Starth&ouml;he der untersten Stufe.
     */
    public void set(
        float width,
        int stepCount,
        float stepHeight,
        float stepDeep,
        float startHeight)
    {
        // Alte Bloecke zuvor loeschen
        this.removeAllChildren();
        
        float length = ((float) stepCount) * stepDeep;
        this.stepCount = stepCount;
        float height = ((float) stepCount) * stepHeight + startHeight;
        float currentHeight = startHeight;
        Block stepBlock;
        Block lastBlock = null;
        Transform3D lastTrans = new Transform3D();
        Transform3D trans = new Transform3D();
        TransformGroup trGrp;
        Vector3f transVec = new Vector3f();
        Point3f ceilingPoints[] = new Point3f[4];
        Point3f floorPoints[] = new Point3f[4];

        // Array mit Objekten belegen
        for (int i = 0; i < 4; ++i)
        {
            ceilingPoints[i] = new Point3f();
            floorPoints[i] = new Point3f();
        }

        for (int i = 0; i < stepCount; ++i)
        {
            currentHeight += stepHeight;

            floorPoints[3].set(-stepDeep / 2, -currentHeight / 2, -width / 2);
            floorPoints[2].set(-stepDeep / 2, -currentHeight / 2, +width / 2);
            floorPoints[1].set(+stepDeep / 2, -currentHeight / 2, +width / 2);
            floorPoints[0].set(+stepDeep / 2, -currentHeight / 2, -width / 2);
            ceilingPoints[3].set(-stepDeep / 2, currentHeight / 2, -width / 2);
            ceilingPoints[2].set(-stepDeep / 2, currentHeight / 2, +width / 2);
            ceilingPoints[1].set(+stepDeep / 2, currentHeight / 2, +width / 2);
            ceilingPoints[0].set(+stepDeep / 2, currentHeight / 2, -width / 2);

            stepBlock = new Block(floorPoints, ceilingPoints);

            transVec.set(
                ((float) i + 0.5f) * stepDeep - length / 2,
                (currentHeight - height) / 2,
                0);
            trans.setTranslation(transVec);
            trGrp = new TransformGroup(trans);

            trGrp.addChild(stepBlock);
            this.addChild(trGrp);

            if (lastBlock != null)
            {
                lastTrans.mulInverse(trans, lastTrans);
                lastBlock.connect(stepBlock, lastTrans);
            }
            lastBlock = stepBlock;
            lastTrans.set(trans);

            transVec.scale(0.2f);
//            transVec.x = -transVec.x;
            transVec.y = -transVec.y;
            trans.setTranslation(transVec);
            this.setTextureTransform(stepBlock, trans);
        }
    }

    private int stepCount;

    /**
     * Setzt die Texturen, welche die Stufen bei ihrer Erstellung erhalten
     * sollen.<br/>
     * Um f&uuml;r einzelne Stufen die Texturen zu setzen, muss auf das
     * Block Objekt direkt zugegriffen werden. Siehe hierzu die Methode
     * {@linkplain Block#setPlaneTexture(Texture[])}.
     * 
     * @param texture Array von Texturen. Die Anzahl der Texturen im Array
     *           bestimmt f&uuml;r was die Texturen verwendet werden:
     *           <ul>
     *              <li>Befindet sich nur eine Textur im Array, so wird die
     *                  Textur f&uuml;r alle W&auml;nde, den Boden und die
     *                  Decke verwendet.</li>
     *              <li>Befinden sich zwei Texturen im Array, so wird die erste
     *                  Textur f&uuml;r den Boden und die Decke verwendet und
     *                  die zweite Textur f&uuml;r alle W&auml;nde.</li>
     *              <li>Bei drei Texturen, wird die erste Textur f&uuml;r den
     *                  Boden, die zweite Textur f&uuml;r die Decke und die
     *                  dritte Textur f&uuml;r alle W&auml;nde verwendet.</li>
     *           </ul>
     *           Bei mehr als drei Texturen, findet der dritte Fall Anwendung
     *           und alle weiteren Texturen werden ignoriert.
     * @see Block#setPlaneTexture(Texture[])
     */
    public void setTexture(Texture[] texture)
    {
        for (int i = 0; i < stepCount; ++i)
        {
            getBlock(i).setPlaneTexture(texture);
        }
    }

    private void setTextureTransform(Block block, Transform3D transform)
    {
        for (int i = 2; i < block.numPlanes(); ++i)
        {
            TextureAttributes tA = new TextureAttributes();
            tA.setTextureTransform(transform);
            block.getPlane(i).getAppearance().setTextureAttributes(tA);
        }
    }

    /**
     * Gibt das Block Objekte an der Position <code>index</code> zur&uuml;ck.
     * 
     * @return Der Iterator, welcher die Block Objekte enth&auml;lt.
     */
    public Block getBlock(int index)
    {
        return (Block) ((TransformGroup) getChild(index)).getChild(0);
    }
    
    /**
     * Gibt die Anzahl der Stufen der Treppe zur&uml;ck.
     * 
     * @return Die Anzahl der Block Objekte.
     */
    public int numBlocks()
    {
        return numChildren();
    }
    
    /* (non-Javadoc)
     * @see javax.media.j3d.Node#cloneNode(boolean)
     */
    public Node cloneNode(boolean arg0)
    {
        LinearStair newStair = new LinearStair();
        newStair.duplicateNode(this, arg0);
        return newStair;
    }

    /* (non-Javadoc)
     * @see javax.media.j3d.Node#duplicateNode(javax.media.j3d.Node, boolean)
     */
    public void duplicateNode(Node arg0, boolean arg1)
    {
        super.duplicateNode(arg0, arg1);
        
        this.stepCount = ((LinearStair) arg0).stepCount;
    }
}
