/*
 * Created on 28.09.2004
 *
 */
package construction;

import java.util.ArrayList;
import java.util.Iterator;

import javax.media.j3d.BoundingBox;
import javax.media.j3d.BoundingSphere;
import javax.media.j3d.Bounds;
import javax.media.j3d.BranchGroup;
import javax.media.j3d.CapabilityNotSetException;
import javax.media.j3d.Group;
import javax.media.j3d.Node;
import javax.media.j3d.NodeReferenceTable;
import javax.media.j3d.Texture;
import javax.media.j3d.Transform3D;
import javax.media.j3d.TransformGroup;
import javax.vecmath.AxisAngle4f;
import javax.vecmath.Point2f;
import javax.vecmath.Point3d;
import javax.vecmath.Point3f;
import javax.vecmath.Vector3f;
/**
 * Eine Map ist eine BranchGroup mit speziellen Funktionen zur Verwaltung von
 * Sektoren. Sie stellt ein globalen Koordinatensystem zur Verf&uuml;gung, in
 * welchem die Sektoren beschrieben werden k&ouml;nnen.
 * Sie erstellt aus den &uuml;bergebenen Daten die Sektoren, transformiert sie
 * an die richtige Position und gliedert sie in den Szenegraph ein.
 * 
 * @author Martin Hedler
 * @version 1.0 (2005/01/31)
 *
 */
public class Map extends BranchGroup
{

    /**
     * Erstellt eine leere Map.
     */
    public Map()
    {
        super();

        defaultTexture = null;

        sectorList = new ArrayList();
    }

    /**
     * Setzt die Texturen, welche die Sektoren bei ihrer Erstellung erhalten
     * sollen. Diese Texturen werden nur bei den Sektoren gesetzt, die nach dem
     * Aufruf dieser Methode erstellt werden.<br/>
     * Um f&uuml;r einzelne Sektoren die Texturen zu setzen, muss auf das
     * Sector Objekt direkt zugegriffen werden. Siehe hierzu die Methode
     * {@linkplain Block#setPlaneTexture(Texture[])}.
     * 
     * @param tex Array von Texturen. Die Anzahl der Texturen im Array
     *           bestimmt f&uuml;r was die Texturen verwendet werden:
     *           <ul>
     *              <li>Befindet sich nur eine Textur im Array, so wird die
     *                  Textur f&uuml;r alle W&auml;nde, den Boden und die
     *                  Decke verwendet.</li>
     *              <li>Befinden sich zwei Texturen im Array, so wird die erste
     *                  Textur f&uuml;r den Boden und die Decke verwendet und
     *                  die zweite Textur f&uuml;r alle W&auml;nde.</li>
     *              <li>Bei drei Texturen, wird die erste Textur f&uuml;r den
     *                  Boden, die zweite Textur f&uuml;r die Decke und die
     *                  dritte Textur f&uuml;r alle W&auml;nde verwendet.</li>
     *           </ul>
     *           Bei mehr als drei Texturen, findet der dritte Fall Anwendung
     *           und alle weiteren Texturen werden ignoriert.
     * @see Block#setPlaneTexture(Texture[])
     */
    public void setDefaultTextures(Texture[] tex)
    {
        if (tex == null)
            throw new NullPointerException("Texturearray ist nicht gesetzt.");
        this.defaultTexture = tex;
    }

    private Texture[] defaultTexture;

    private ArrayList sectorList;

    /**
     * Ermittelt das <code>Group</code> Objekt mit dem kleinsten
     * H&uml;hlk&ouml;rper, welches sich unterhalb von <code>group</code>
     * befindet und in die <code>newElement</code> vollst&auml;ndig hineinpasst.
     * <p/>
     * Dabei wird die Struktur des Subgraphen unter <code>group</code>
     * ver&auml;ndert, wenn andere Knoten ggf. unter <code>newElement</code>
     * eingef&uuml;gt werden.
     * 
     * @param newElement Das Element, was in den Graph hinzugef&uuml;gt werden
     *        soll.
     * @param newElementLclToGlbTransform Die globale Transformation von
     *        <code>newElement</code>.
     * @param group Der Subszenegraph in den <code>newElement</code>
     *        eingef&uuml;gt werden soll.
     * @param groupLclToGlbTransform Die globale Transformation von
     *        <code>group</code>.
     * @param volume Ein Array der L&auml;nge 1 in das das Volumen des
     *        H&uuml;llk&ouml;rpers der gefunden Gruppe geschrieben wird.
     * @param foundedGroupLclToGlbTransform Die globale Transformation der
     *        gefundenen Gruppe.
     * @param searchForConnections <code>true</code>, wenn ein Block
     *        eingef&uuml;gt wird und Verbindungen mit anderen Block Objekten
     *        hergestellt werden soll.
     * @return Das <code>Group</code> Objekt in welches <code>newElement</code>
     *         vollst&auml;ndig hineinpassen w&uuml;rde.
     */
    private Group getPotParent(
        Node newElement,
        Transform3D newElementLclToGlbTransform,
        Group group,
        Transform3D groupLclToGlbTransform,
        float[] volume,
        Transform3D foundedGroupLclToGlbTransform,
        boolean searchForConnections)
    {
        Group result = null;
        volume[0] = Float.MAX_VALUE;

        Bounds groupBounds, newBounds, tmpBounds;

        Transform3D childTransformLclToGlb = new Transform3D();
        if (group instanceof TransformGroup)
             ((TransformGroup) group).getTransform(childTransformLclToGlb);
        childTransformLclToGlb.mul(
            groupLclToGlbTransform,
            childTransformLclToGlb);

        groupBounds = (Bounds) group.getBounds().clone();
        groupBounds.transform(childTransformLclToGlb);
        newBounds = (Bounds) newElement.getBounds().clone();
        newBounds.transform(newElementLclToGlbTransform);

        tmpBounds = (Bounds) groupBounds.clone();
        tmpBounds.combine(newBounds);

        // Liegt newElement in group?
        if (tmpBounds.equals(groupBounds)
            || (searchForConnections && newBounds.intersect(groupBounds)))
        {
            if (searchForConnections && group instanceof Block)
            {
                Transform3D otherToNewTransform = new Transform3D();
                otherToNewTransform.mulInverse(
                    childTransformLclToGlb,
                    newElementLclToGlbTransform);
                ((Block) newElement).connect(
                    (Block) group,
                    otherToNewTransform);
            }

            if (tmpBounds.equals(groupBounds))
            {
                volume[0] = getVolume(groupBounds);
                result = group;
                foundedGroupLclToGlbTransform.set(childTransformLclToGlb);
            }

            float[] otherV = new float[1];
            Transform3D otherTrans = new Transform3D();
            Group otherResult;

            // newElement mit Kindern von group untersuchen
            for (int i = 0; i < group.numChildren(); ++i)
            {
                otherResult = null;
                if (group.getChild(i) instanceof Group)
                {
                    otherResult =
                        getPotParent(
                            newElement,
                            newElementLclToGlbTransform,
                            (Group) group.getChild(i),
                            childTransformLclToGlb,
                            otherV,
                            otherTrans,
                            searchForConnections);

                    if (otherResult != null && otherV[0] <= volume[0])
                    {
                        result = otherResult;
                        volume[0] = otherV[0];
                        foundedGroupLclToGlbTransform.set(otherTrans);
                    }

                    // Ist Untergruppe jetzt leer?
                    if (((Group) group.getChild(i)).numChildren() == 0)
                    {
                        group.removeChild(i);
                        --i;
                        continue;
                    }
                }

                if (otherResult == null
                    && newElement instanceof Group
                    && group.getChild(i) instanceof Block)
                {
                    otherResult =
                        getPotParent(
                            group.getChild(i),
                            childTransformLclToGlb,
                            (Group) newElement,
                            newElementLclToGlbTransform,
                            otherV,
                            otherTrans,
                            false);

                    if (otherResult != null)
                    {
                        move(
                            group.getChild(i),
                            childTransformLclToGlb,
                            group,
                            otherResult,
                            otherTrans);
                    }
                }
            }
        }

        return result;
    }

    private void move(
        Node element,
        Transform3D elementGlbTrans,
        Group srcGroup,
        Group destGroup,
        Transform3D destGroupGlbTrans)
    {
        if (srcGroup != null)
            srcGroup.removeChild(element);

        Transform3D destTrans = new Transform3D();

        destTrans.mul(destGroupGlbTrans, destTrans);

        if (elementGlbTrans.equals(destTrans))
        {
            destGroup.addChild(element);
        } else
        {
            //Transform3D newTrans = new Transform3D();
            destTrans.mulInverse(elementGlbTrans, destTrans);
            TransformGroup trGrp = new TransformGroup(destTrans);
            trGrp.addChild(element);
            destGroup.addChild(trGrp);
        }
    }

    /**
     * Liefert das Volumen des &uuml;bergebenen H&uuml;llk&ouml;rpers
     * zur&uuml;ck.
     * 
     * @param b Der H&uuml;llk&ouml;rper
     * @return Das Volumen.
     * @throws IllegalArgumentException Wenn der H&uuml;llk&ouml;rper Typ nicht
     *         unterst&uuml;tzt wird.
     */
    private float getVolume(Bounds b) throws IllegalArgumentException
    {
        if (b instanceof BoundingBox)
        {
            BoundingBox box = (BoundingBox) b;
            Point3d lower = new Point3d();
            Point3d upper = new Point3d();
            box.getLower(lower);
            box.getUpper(upper);
            upper.sub(lower);
            return (float) (upper.x * upper.y * upper.z);
        } else
            if (b instanceof BoundingSphere)
            {
                BoundingSphere sphere = (BoundingSphere) b;
                return (float)
                    (4
                        * Math.PI
                        * sphere.getRadius()
                        * sphere.getRadius()
                        * sphere.getRadius()
                        / 3);
            } else
            {
                throw new IllegalArgumentException();
            }
    }

    /**
     * Erstellt einen Ausdruck (auf der Standardausgabe) des Szenegraph der
     * &uuml;bergebenen Group.
     * 
     * @param g Die Group, von der ausgehend der Graph ausgedruckt werden soll.
     * @param all Gibt an, ob alle Objekte des Szenegraph angezeigt werden
     *           sollen, oder nur Sektoren. 
     * @throws CapabilityNotSetException Wenn die Group Teil eines lebenen
     *            Szenegraph ist und die ben&ouml;tigte Capability nicht
     *            gesetzt ist.
     */
    public static void printSceneGraph(Group g, boolean all)
        throws CapabilityNotSetException
    {
        Map.printSceneGraph(g, 0, all);
    }

    /**
     * Erstellt einen Ausdruck (auf der Standardausgabe) des Szenegraph der
     * &uuml;bergebenen Group.
     * 
     * @param g Die Group, von der ausgehend der Graph ausgedruckt werden soll.
     * @param level Die aktuelle Verschachtelungstiefe des Graphen. Wird zur
     *           Darstellung ben&ouml;tigt.
     * @param all Gibt an, ob alle Objekte des Szenegraph angezeigt werden
     *           sollen, oder nur Sektoren. 
     * @throws CapabilityNotSetException Wenn die Group Teil eines lebenen
     *            Szenegraph ist und die ben&ouml;tigte Capability nicht
     *            gesetzt ist.
     */
    private static void printSceneGraph(Group g, int level, boolean all)
        throws CapabilityNotSetException
    {
        StringBuffer sB = new StringBuffer();
        for (int i = 0; i < level; ++i)
        {
            sB.append(' ');
        }
        String s = sB.toString();

        for (int i = 0; i < g.numChildren(); ++i)
        {
            if (g.getChild(i) instanceof Group)
            {
                if (all || g.getChild(i) instanceof Block)
                {
                    System.out.println(s + "+ " + g.getChild(i));
                }
                printSceneGraph((Group) g.getChild(i), level + 2, all);
            } else
                if (all)
                    System.out.println(s + "- " + g.getChild(i));
        }
    }

    /**
     * Erstellt nach den angegebenen globalen Parametern einen Block und gibt
     * diesen und desen localToVW Transformation zur&uuml;ck. 
     * 
     * @param cornerPoints Die lobalen 2D Eckpunkte des Sektors.
     * @param floorHeight Die globale Bodenh&ouml;he.
     * @param ceilingHeight Die globale Deckenh&ouml;he.
     * @param floorAngle Die Neigungsachse des Bodens.
     * @param ceilingAngle Die Neigungsachse der Decke.
     * @param glbTrans In diesem Objekt wird die Transformation gespeichert.
     * @return Der erstellte Block.
     * @throws IllegalArgumentException Wenn die Punktliste
     *         <code>cornerPoints</code> fehlerhaft ist (siehe
     *  {@linkplain #createBlock(Point2f[], float, AxisAngle4f, AxisAngle4f)}).
     */
    private Block createSectorTransform(
        Point2f[] cornerPoints,
        float floorHeight,
        float ceilingHeight,
        AxisAngle4f floorAngle,
        AxisAngle4f ceilingAngle,
        Transform3D glbTrans)
        throws IllegalArgumentException
    {
        // Gueltigkeit des Punktarrays ueberpruefen.
        if (cornerPoints.length <= 0)
            throw new IllegalArgumentException("Lnge des Arrays ist 0.");
        // Mittelpunkt berechnen
        Point2f centerPoint = new Point2f();
        calculateCenterPoint(cornerPoints, centerPoint);
        // Punkte kopieren und transformieren
        Point2f[] localCornerPoints = new Point2f[cornerPoints.length];
        for (int i = 0; i < cornerPoints.length; ++i)
        {
            localCornerPoints[i] = new Point2f(cornerPoints[i]);
            localCornerPoints[i].sub(centerPoint);
        }

        // Hoehe des Sektors errechnen
        float height = ceilingHeight - floorHeight;
        // Sektor erstellen
        Block sector =
            createBlock(localCornerPoints, height, floorAngle, ceilingAngle);
        if (defaultTexture != null)
            sector.setPlaneTexture(defaultTexture);

        // Transformation erstellen
        Vector3f transVector =
            new Vector3f(centerPoint.x, floorHeight, centerPoint.y);
        glbTrans.setTranslation(transVector);

        return sector;
    }

    /**
     * Berechnet den Mittelpunkt einer Punktwolke.
     * 
     * @param points Die Punktwolke zu welcher der Mittelpunkt errechnet werden soll.
     * @param centerPoint In diesem Objekt werden die Werte des Mittelpunktes
     *           geschrieben.
     */
    private void calculateCenterPoint(Point2f[] points, Point2f centerPoint)
    {
        assert points.length >= 0;
        Point2f max = new Point2f(points[0]), min = new Point2f(points[0]);
        for (int i = 0; i < points.length; ++i)
        {
            max.x = Math.max(max.x, points[i].x);
            max.y = Math.max(max.y, points[i].y);
            min.x = Math.min(min.x, points[i].x);
            min.y = Math.min(min.y, points[i].y);
        }

        centerPoint.add(min, max);
        centerPoint.scale(0.5f);
    }

    /**
     * Erstellt einen neuen Sektor und f&uuml;gt diesen in den Szenegraphen ein.
     * 
     * @param cornerPoints Die <i>globalen</i> Koordinaten der Ecken des
     *        Sektors. Die Punkte sind im Uhrzeigersinn angeordnet, wenn ein
     *        positiver Sektor entstehen soll und gegen den Uhrzeigersinn, wenn
     *        ein negativer Sektor entstehen soll. Die Eckpunkte liegen alle in
     *        der xz-Ebene, d.h. das Attribut <code>y</code> von
     *        {@linkplain Point2f} entspricht <code>z</code>.
     * @param floorHeight Die <i>globale</i> H&ouml;he des Bodens.
     * @param ceilingHeight Die <i>globale</i> H&ouml;he der Decke.
     * @return Der erstellte Sektor.
     */
    public Block createSector(
        Point2f[] cornerPoints,
        float floorHeight,
        float ceilingHeight)
    {
        AxisAngle4f aa = new AxisAngle4f(0, 0, 1, 0);
        return createSector(cornerPoints, floorHeight, ceilingHeight, aa, aa);
    }

    /**
     * Erstellt einen neuen Sektor und f&uuml;gt diesen in den Szenegraphen ein.
     * 
     * @param cornerPoints Die <i>globalen</i> Koordinaten der Ecken des
     *        Sektors. Die Punkte sind im Uhrzeigersinn angeordnet, wenn ein
     *        positiver Sektor entstehen soll und gegen den Uhrzeigersinn, wenn
     *        ein negativer Sektor entstehen soll. Die Eckpunkte liegen alle in
     *        der xz-Ebene, d.h. das Attribut <code>y</code> von
     *        {@linkplain Point2f} entspricht <code>z</code>.
     * @param floorHeight Die <i>globale</i> H&ouml;he des Bodens.
     * @param ceilingHeight Die <i>globale</i> H&ouml;he der Decke.
     * @param floorAngle Die Neigung des Bodens. Die Neigungsachse muss in der
     *           xz-Ebene liegen.
     * @param ceilingAngle Die Neigung der Decke. Die Neigungsachse muss in der
     *           xz-Ebene liegen.
     * @return Der erstellte Sektor.
     */
    public Block createSector(
        Point2f[] cornerPoints,
        float floorHeight,
        float ceilingHeight,
        AxisAngle4f floorAngle,
        AxisAngle4f ceilingAngle)
    {
        Transform3D glbTrans = new Transform3D();
        
        // Sektor erstellen
        Block sector =
            createSectorTransform(
                cornerPoints,
                floorHeight,
                ceilingHeight,
                floorAngle,
                ceilingAngle,
                glbTrans);

        Transform3D resultTrans = new Transform3D();

        // Moegliche Einfuegeposition im Szenegraph finden
        Group result =
            getPotParent(
                sector,
                glbTrans,
                this,
                new Transform3D(),
                new float[1],
                resultTrans,
                true);

        if (result == null)
        {
            result = this;
        }

        // Einfuegen
        move(sector, glbTrans, null, result, resultTrans);

        sectorList.add(sector);
        return sector;
    }

    /**
     * Liefert einen Iterator aller Sektoren der Karte zur&uuml;ck.
     * 
     * @return Der Iterator.
     */
    public Iterator getSectors()
    {
        return sectorList.iterator();
    }

    /**
     * Erstellt die Geometrie (W&auml;nde, Boden, Decke) des Sektors.
     * 
     * @param corner Array von Eckpunkten in der xz-Ebene. Die Punkte sind im
     *        Uhrzeigersinn angeordnet, wenn ein positiver Sektor entstehen soll
     *        und gegen den Uhrzeigersinn, wenn ein negativer Sektor entstehen
     *        soll. Die Eckpunkte liegen alle in der xz-Ebene, d.h. das Attribut
     *        <code>y</code> von {@linkplain Point2f} entspricht <code>z</code>.
     * @param height H&ouml;he des Sektors.
     * @param angleFloor Neigungsachse und -winkel des Bodens.
     * @param angleCeiling Neigungsachse und -winkel der Decke.
     * @throws IllegalArgumentException Wenn die &uuml;bergebenen Liste der Eckpunkte
     *            fehlerhaft ist. Die Punktliste ist dann fehlerhaft, wenn:
     *            <ul>
     *               <li>weniger als 3 Punkte enthalten sind,
     *               <li>zwei oder mehr Punkte gleich sind, oder</li>
     *               <li>sich die zwischen den Ecken verlaufenden W&auml;nde
     *                   schneiden.</li>
     *            </ul>
     *            Die Exception wird au&szlig;erdem geworfen, wenn
     *            <ul> 
     *               <li>die H&ouml;he des Sektors kleiner oder gleich 0 ist,</li> 
     *               <li>die AxisAngles nicht in der xz-Ebene liegen, oder</li>
     *               <li>die Neigungswinkel nicht im Bereich ]-pi/2, pi/2[
     *                   liegen.</li>
     *            </ul>
     */
    private Block createBlock(
        Point2f[] corner,
        float height,
        AxisAngle4f angleFloor,
        AxisAngle4f angleCeiling)
        throws IllegalArgumentException
    {
        if (height <= 0)
        {
            throw new IllegalArgumentException("Hhe des Sektors muss grer als Null sein.");
        }

        if (angleFloor.y != 0 || angleCeiling.y != 0)
        {
            throw new IllegalArgumentException("Rotationsachse muss in xz-Ebene liegen.");
        }

        if (Math.abs(angleFloor.angle) >= Math.PI / 2
            || Math.abs(angleCeiling.angle) >= Math.PI / 2)
        {
            throw new IllegalArgumentException("Neigunswinkel muss im Bereich ]-pi/2, pi/2[ liegen.");
        }

        // Punkte berechnen
        Point3f[] floorPoints = new Point3f[corner.length];
        Point3f[] ceilingPoints = new Point3f[corner.length];
        transform2DTo3D(corner, floorPoints, 0, angleFloor);
        transform2DTo3D(corner, ceilingPoints, height, angleCeiling);

        return new Block(floorPoints, ceilingPoints);
    }

    private void transform2DTo3D(
        Point2f[] corners,
        Point3f[] points,
        float height,
        AxisAngle4f angle)
    {
        Point3f tmpPoint = new Point3f();
        Vector3f normal = new Vector3f();
        calculatePlaneNormal(angle, normal);

        // Eckpunkte der Ebene ermitteln
        for (int i = 0; i < corners.length; ++i)
        {
            points[i] = new Point3f();
            this.calculateCrossPoint(
                normal,
                corners[i].x,
                corners[i].y,
                tmpPoint);
            tmpPoint.y += height;
            try
            {
                points[i].set(tmpPoint);
            } catch (NullPointerException nE)
            {
                points[i] = new Point3f(tmpPoint);
            }
        }
    }

    /**
     * Ermittelt den Schnittpunkt der Geraden
     * <code>g: (x,0,z) + r * (0, 1, 0)</code> mit der
     * Ebene die durch die Normale <code>normal</code> gegeben ist.
     * 
     * @param normal Normale der Ebene.
     * @param x x-Komponente des St&uuuml;tzvektors der Geraden g.
     * @param z z-Komponente des St&uuuml;tzvektors der Geraden g.
     * @param destPoint Punkt, in welchem die Werte des Schnittpunktes gespeichert
     *           werden.
     */
    private void calculateCrossPoint(
        Vector3f normal,
        float x,
        float z,
        Point3f destPoint)
    {
        assert normal.y != 0;

        destPoint.set(x, - (normal.x * x + normal.z * z) / normal.y, z);
    }

    /**
     * Errechnet die Ebenennormale der xz-Ebene wenn sie um die Rotationchse
     * und -winkel <code>axisAngle</code> rotiert wird.
     * 
     * <b>Voraussetzung:</b> axisAngle liegt in der xz-Ebene. 
     * 
     * @param axisAngle
     * @param normal
     */
    private void calculatePlaneNormal(AxisAngle4f axisAngle, Vector3f normal)
    {
        assert axisAngle.y != 0;

        Vector3f helpVector =
            new Vector3f(axisAngle.x, axisAngle.y, axisAngle.z);
        normal.set(0, 1, 0);

        helpVector.cross(helpVector, normal);
        helpVector.normalize(); // Fehler, wenn helpVector = 0;
        helpVector.scale((float) Math.sin(axisAngle.angle));

        normal.scaleAdd((float) Math.cos(axisAngle.angle), normal, helpVector);
        normal.normalize();
    }

    /* (non-Javadoc)
     * @see javax.media.j3d.Node#cloneNode(boolean)
     */
    public Node cloneNode(boolean arg0)
    {
        Map newMap = new Map();
        newMap.duplicateNode(this, arg0);
        return newMap;
    }

    /* (non-Javadoc)
     * @see javax.media.j3d.Node#duplicateNode(javax.media.j3d.Node, boolean)
     */
    public void duplicateNode(Node arg0, boolean arg1)
    {
        super.duplicateNode(arg0, arg1);

        Map org = (Map) arg0;
        this.defaultTexture = new Texture[org.defaultTexture.length];
        System.arraycopy(
            org.defaultTexture,
            0,
            this.defaultTexture,
            0,
            org.defaultTexture.length);
        this.sectorList = (ArrayList) org.sectorList.clone();
    }

    /* (non-Javadoc)
     * @see javax.media.j3d.SceneGraphObject#updateNodeReferences(javax.media.j3d.NodeReferenceTable)
     */
    public void updateNodeReferences(NodeReferenceTable arg0)
    {
        super.updateNodeReferences(arg0);

        for (int i = 0; i < this.sectorList.size(); ++i)
        {
            this.sectorList.set(
                i,
                arg0.getNewObjectReference((Block) this.sectorList.get(i)));
        }
    }

}
