/*
 * Created on 12.09.2004
 *
 */
package construction;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;

import javax.media.j3d.Appearance;
import javax.media.j3d.BoundingBox;
import javax.media.j3d.Bounds;
import javax.media.j3d.CapabilityNotSetException;
import javax.media.j3d.Geometry;
import javax.media.j3d.IllegalSharingException;
import javax.media.j3d.IndexedGeometryArray;
import javax.media.j3d.Material;
import javax.media.j3d.Node;
import javax.media.j3d.PolygonAttributes;
import javax.media.j3d.RestrictedAccessException;
import javax.media.j3d.SceneGraphCycleException;
import javax.media.j3d.Shape3D;
import javax.media.j3d.TextureAttributes;
import javax.media.j3d.Transform3D;
import javax.media.j3d.TriangleArray;
import javax.vecmath.Color3f;
import javax.vecmath.Color4f;
import javax.vecmath.Point3d;
import javax.vecmath.Point3f;
import javax.vecmath.TexCoord2f;
import javax.vecmath.Vector3f;
import javax.vecmath.Vector4f;

import com.sun.j3d.utils.geometry.GeometryInfo;
import com.sun.j3d.utils.geometry.NormalGenerator;
import com.sun.j3d.utils.geometry.Stripifier;

/**
 * Eine Plane ist eine durch Eckpunkte begrenzte, ebene Fl&auml;che im Raum.
 * Sie kann mit anderen Fl&auml;chen so verbunden werden, dass die Schnittmenge
 * der beiden Fl&auml;chen herausgeschnitten wird.
 * <p/>
 * Dieses Objekt weisst eine wichtige Besonderheit auf:<br/>
 * Bei jeder &Auml;nderung, welche die Plane an der Geometrie des Objektes
 * vornimmt, wird das <code>Geometry</code> Objekt komplett neu erstellt. D.h.
 * alle &Auml;nderungen an der Geometrie, welche nicht von der Plane selbst
 * durchgef&uuml;hrt werden, gehen verloren.<br/>
 * Dies betrifft unter anderem auch das Setzen der Capabilities der Geometrie.
 * D.h., wenn zum Beispiel eine Plane mit einer anderen Plane verbunden wird,
 * <i>nachdem</i> die Capabilities der Geometrie gesetzt wurden, so sind nach
 * dem Verbinden alle Capabilities wieder gel&ouml;scht.</p>
 * 
 * Folgende Methoden der Plane erstellen die Geometrie neu (neben dem Konstruktor):
 * <ul>
 *    <li>{@linkplain #subIntersection(Plane, Transform3D)}</li>
 * </ul>
 *
 * Die Plane erm&ouml;glicht jedoch die Registrierung von EventListeners, die
 * direkt nach einer Neuerstellung der Geometrie, aber noch vor deren
 * Einf&uuml;gen in das Shape3D Objekt, aufgerufen werden. So dass hier die
 * &Auml;nderungen an der Geometrie durchgef&uuml;hrt werden k&ouml;nnen, die
 * erhalten bleiben sollen.
 *
 * @author Martin Hedler
 * @version 1.0 (2005/01/31)
 */
public class Plane extends Shape3D
{
    /** Objekt zum Vergleichen von Punkten auf einer Geraden. */
    private PointComperator comperator;

    /** Liste der Punkte, welche die Plane beschreiben. */
    private Point3f[] point;

    /** Der Normalenvektor der Ebene. */
    private Vector3f normal;

    /** Liste der L&ouml;cher (Point3f-Arrays). */
    private ArrayList holeList;

    /** Verwaltet die PlaneListener. */
    private ArrayList listenerList;

    /** Genauigkeit bei der Schnittpunktberechnung. */
    private static final float epsilon = 0.0001f;

    /** &Uuml;berpr&uuml;fte Plane ist innerhalb. */
    public static final int IN = 1;

    /** &Uuml;berpr&uuml;fte Plane enth&auml;lt diese Plane. */
    public static final int INCLUDING = 2;

    /** Die Planes sind identisch. */
    public static final int EQUAL = 4;

    /**
     * Die Planes liegen in der selben Ebene und
     * &uuml;berschneiden/&uuml;berlagern sich.
     */
    public static final int INTERSECTED = 8;

    /** Die Planes liegen in verschiedenen Ebenen und schneiden sich. */
    public static final int CROSSING = 16;

    /** Die Normalen der Planes haben die selbe Richtung */
    public static final int SAME_DIRECTION = 32;

    /** &Uuml;berpr&uuml;fte Plane ist ausserhalb. */
    public static final int OUT = 64;

    /**
     * Erzeugt eine Plane mit den angegebenen Punkten.
     *
     * @param point Punkte, welche die Fl&auml;che beschreiben.
     * @throws IllegalArgumentException IllegalArgumentException Wenn die
     *         &uuml;bergebene Punktliste fehlerhaft ist. Die Punktliste ist
     *         dann fehlerhaft, wenn:
     *            <ul>
     *               <li>weniger als 3 Punkte enthalten sind,
     *               <li>zwei oder mehr Punkte gleich sind,</li>
     *               <li>sich die zwischen den Punkten aufgespannten Kanten
     *                   schneiden oder</li>
     *               <li>nicht alle Punkte in der selben Ebene liegen.</li>
     *            </ul>.
     */
    public Plane(Point3f[] point) throws IllegalArgumentException
    {
        super();

        this.point = point;
        this.normal = new Vector3f();
        this.comperator = new PointComperator();
        this.holeList = new ArrayList();
        this.listenerList = new ArrayList();
        this.setBoundsAutoCompute(false);
        this.setBounds(null);
        this.setAppearance(createAppearance());

        this.set(point);
    }

    /**
     * Gibt einen Eckpunkt der Plane zur&uuml;ck.
     * 
     * @param index Index des gew&uuml;nschten Punktes.
     * @param point Wird mit den Werten des Punktes belegt.
     * @throws IndexOutOfBoundsException Wenn der Index ausserhalb des zul&auml;ssigen
     *            Bereichs ist.
     * @see #Plane(Point3f[])
     */
    public void getPoint(int index, Point3f point)
        throws IllegalStateException, IndexOutOfBoundsException
    {
        point.set(this.point[index]);
    }

    /**
     * Gibt die Anzahl der Eckpunkte zur&uuml;ck aus denen die Plane besteht. 
     * @return Die Anzahl der Eckpunkte.
     * @see #Plane(Point3f[])
     */
    public int numPoints()
    {
        if (this.point == null)
            return 0;

        return this.point.length;
    }

    /**
     * Setzt die Punkte der Ebene. Es wird eine vollst&auml;ndige tiefe Kopie
     * des Arrays durchgef&uuml;hrt.
     *
     * @param point Punkte, welche die Fl&auml;che beschreiben.
     * @throws IllegalArgumentException Wenn die &uuml;bergebene Punktliste fehlerhaft
     *            ist. Die Punktliste ist dann fehlerhaft, wenn:
     *            <ul>
     *               <li>weniger als 3 Punkte enthalten sind,
     *               <li>zwei oder mehr Punkte gleich sind,</li>
     *               <li>sich die zwischen den Punkten aufgespannten Kanten
     *                   schneiden oder</li>
     *               <li>nicht alle Punkte in der selben Ebene liegen.</li>
     *            </ul>
     */
    private void set(Point3f[] point) throws IllegalArgumentException
    {
        assert holeList.size() == 0;

        // Punkte auf Korrektheit pruefen
        if (isDataCorrect(point) == false)
        {
            throw new IllegalArgumentException("Die Punktliste ist nicht korrekt.");
        }

        // BoundingBox initialisieren
        Point3d pointDouble = new Point3d();
        BoundingBox bounds =
            new BoundingBox(
                new Point3d(-0.01, -0.01, -0.01),
                new Point3d(0.01, 0.01, 0.01));

        // Punkte kopieren
        Vector3f pointVector = new Vector3f();
        this.point = new Point3f[point.length];
        for (int i = 0; i < point.length; ++i)
        {
            pointDouble.set(point[i]);
            // BoundingBox vergroessern
            bounds.combine(pointDouble);

            // Kontrollieren, ob die Punkte alle in der Ebenen liegen
            pointVector.set(pointDouble);

            this.point[i] = new Point3f(pointVector);
        }

        try
        {
            // Geometry neu erstellen
            this.setGeometry(createGeometry());

            // BoundingBox setzen
            this.setBounds(bounds);
        } catch (GeometryNotCreated gE)
        {
            this.removeAllGeometries();
        }
    }

    /**
     * Schneidet aus beiden Planes den Durschnitt heraus. Dies ist nur
     * m&ouml;glich, wenn die Planes in der selben Ebene liegen, sich
     * gegenseitig &uuml;berlagern und die Normalen der beiden in
     * engegengesetzte Richtungen zeigen. Die sich &uuml;berlagernden Teile der
     * beiden Planes werden herausgeschnitten.
     *  
     * @param other Die Plane, mit der die Operation durchgef&uuml;rt werden
     *        soll.
     * @param otherTrans Transformation, um <code>other</code> in das lokale
     *        Koordinatensystem von <code>this</code> zu transformieren.
     * @return <code>true</code>, wenn die Operation erfolgreich war, sonst
     *         <code>false</code>.
     * @throws CapabilityNotSetException Wenn die Plane Teil eines lebenden
     *            Szenegraphen ist und folgende Capabilities <i>nicht</i>
     *            gesetzt sind:
     *            <ul>
     *               <li>{@linkplain Node#ALLOW_BOUNDS_READ}</li>
     *               <li>{@linkplain Shape3D#ALLOW_GEOMETRY_WRITE}</li>
     *            </ul>
     * @throws SceneGraphCycleException Wenn die Plane Teil eines zyklisch
     *            geschlossenen Szenegraph ist.
     * @throws IllegalArgumentException Wenn das Planeobjekt selbst &uuml;bergeben
     *            wurde.
     * @see #Plane(Point3f[])
     */
    public boolean subIntersection(Plane other, Transform3D otherTrans)
        throws
            IllegalArgumentException,
            CapabilityNotSetException,
            SceneGraphCycleException
    {
        // Soll die Plane mit sich selber verbunden werden?
        if (other == this)
        {
            throw new IllegalArgumentException("Plane kann nicht mit sich selber verbunden werden.");
        }

        // List von Punkten erstellen
        ArrayList thisPointList = new ArrayList();
        ArrayList otherPointList = new ArrayList();
        ArrayList entryPointList = new ArrayList();

        int state =
            this.getState(
                other,
                otherTrans,
                otherPointList,
                thisPointList,
                entryPointList,
                false);

        if ((state & Plane.CROSSING) == Plane.CROSSING || state == 0)
        {
            return false;
        }

        if ((state & Plane.SAME_DIRECTION) == Plane.SAME_DIRECTION)
        {
            return false;
        }

        Transform3D otherTransInverte = new Transform3D();
        otherTransInverte.invert(otherTrans);

        if ((state & Plane.INTERSECTED) == Plane.INTERSECTED
            || (state & Plane.IN) == Plane.IN
            || (state & Plane.INCLUDING) == Plane.INCLUDING
            || (state & Plane.EQUAL) == Plane.EQUAL)
        {
            // Schnitt mit bereits existierenden Loechern?
            if (areHolesIntersected(other, otherTrans)
                || other.areHolesIntersected(this, otherTransInverte))
            {
                return false;
            }
        }

        if ((state & Plane.INTERSECTED) == Plane.INTERSECTED)
        {

            //Schnittpolygone ermitteln
            ArrayList polygonList = new ArrayList();
            this.createANDPolygons(
                otherPointList,
                thisPointList,
                entryPointList,
                polygonList);

            // Loecher transformieren
            ArrayList otherHole = new ArrayList();
            other.transform(otherTransInverte, polygonList, otherHole);

            // Loecher speichern
            this.createHole(polygonList);
            other.createHole(otherHole);
            return true;
        } else
            if ((state & Plane.IN) == Plane.IN)
            {
                // other liegt in this -> other bestimmt das Loch

                // Loch in this schneiden
                createHole(other.point, otherTrans, true);

                // Loch in other schneiden
                other.createHole(other.point, null, false);

                return true;
            } else
                if ((state & Plane.INCLUDING) == Plane.INCLUDING)
                {
                    // this liegt in other -> this bestimmt das Loch

                    // Loch in other schneiden
                    other.createHole(this.point, otherTransInverte, true);

                    // Loch in this schneiden
                    createHole(this.point, null, false);

                    return true;
                } else
                    if ((state & Plane.EQUAL) == Plane.EQUAL)
                    {
                        // Loch in this schneiden
                        createHole(this.point, null, false);

                        // Loch in other schneiden
                        other.createHole(other.point, null, false);

                        return true;
                    }

        return false;
    }

    private boolean areHolesIntersected(Plane other, Transform3D otherTrans)
    {
        Point3f[] pointArray;
        //List polygonList;
        ArrayList holePointList = new ArrayList();
        ArrayList otherPointList = new ArrayList();
        Point3f[] otherPointArray = new Point3f[other.point.length];
        Point3f point;
        boolean turn = false;

        //Vector3f glbThisNormal = new Vector3f();
        Vector3f glbOtherNormal = new Vector3f();
        otherTrans.transform(other.normal, glbOtherNormal);

        // Ist die Punktliste von other im Urzeigersinn angeordnet?
        // Oder anders: Sind die Normalenvektoren entgegengesetzt?
        if (this.normal.dot(glbOtherNormal) < 0)
        {
            // Liste von other anders herum anordnen
            turn = true;
        }

        // Punktliste von other erstellen
        for (int i = 0; i < other.point.length; ++i)
        {
            point = new Point3f();
            otherTrans.transform(other.point[i], point);
            if (turn)
            {
                otherPointArray[other.point.length - i - 1] = point;
                otherPointList.add(0, point);
            } else
            {
                otherPointArray[i] = point;
                otherPointList.add(point);
            }
        }

        for (Iterator polygonLists = this.holeList.iterator();
            polygonLists.hasNext();
            )
        {
            pointArray = (Point3f[]) polygonLists.next();
            holePointList.clear();

            // Punkte der Loecher kopieren
            for (int i = 0; i < pointArray.length; ++i)
            {
                //thisTrans.transform(pointArray[i], point);
                //holePointArray[i] = new Point3f(point);
                holePointList.add(pointArray[i]);
            }

            // Schneiden sich die Listen?
            if (createIntersectionPointList(otherPointList,
                holePointList,
                null,
                true))
            {
                return true;
            }

            if (getPositionInfo(pointArray, otherPointArray) != OUT)
            {
                return true;
            } else
                if (getPositionInfo(otherPointArray, pointArray) != OUT)
                {
                    return true;
                }
        }
        return false;
    }

    /**
     * Gibt zur&uuml;ck, in welchem Verh&auml;ltnis die zu &uuml;berpr&uuml;fende
     * Plane zu dieser Plane steht.
     *
     * @param other Die Plane, die &uuml;berpr&uuml;ft werden soll.
     * @param otherTrans Transformation, um <code>other</code> in das lokale
     *        Koordinatensystem von <code>this</code> zu transformieren.
     * @return Folgende Bits k&ouml;nnen gesetzt sein:
     *            <ul>
     *               <li>IN, wenn other komplett in this liegt.</li>
     *               <li>INCLUDING, wenn this in other liegt.</li>
     *               <li>EQUAL, wenn die beiden Planes genau aufeinander liegen.</li>
     *               <li>CROSSING, wenn die beiden Planes in verschiedenen Ebenen
     *                   liegen und sich schneiden.</li>
     *               <li>INTERSECTED, wenn die Planes in der selben Ebene liegen
     *                   und sich teilweise &uuml;berlagern.</li>
     *            </ul>
     *            Wenn die Planes in keinem Kontakt zueinander stehen, ist kein
     *            Bit gesetzt, es wird also 0 zur&uuml;ckgegeben.
     * @throws CapabilityNotSetException Wenn die Plane Teil eines lebenden Szenegraphen
     *            ist und folgende Capabilities <i>nicht</i> gesetzt sind:
     *            <ul>
     *               <li>{@linkplain Node#ALLOW_BOUNDS_READ}</li>
     *            </ul>
     * @throws SceneGraphCycleException Wenn sich die Plane in einem zyklisch
     *            geschlossenen Szenegraph befindet.
     * @throws RestrictedAccessException Wenn die Plane Teil eines kompilierten, aber
     *            nicht lebendigen Szenegraphen ist.
     * @throws IllegalArgumentException Wenn die Plane selbst &uuml;bergeben wurde.
     * @see #Plane(Point3f[])
    */
    public int getState(Plane other, Transform3D otherTrans)
        throws
            CapabilityNotSetException,
            SceneGraphCycleException,
            RestrictedAccessException,
            IllegalArgumentException
    {
        return this.getState(other, otherTrans, null, null, null, true);
    }

    /**
     * Gibt zur&uuml;ck, in welchem Verh&auml;ltnis die zu &uuml;berpr&uuml;fende
     * Plane zu dieser Plane steht. Die &uuml;bergebenen Listen und der BOOL'che
     * Parameter werden an die Funktion 
     * {@linkplain #createIntersectionPointList(List, List, List, boolean)}
     * weitergegeben. Siehe dort f&uuml;r mehr Informationen! Statt der Listen
     * kann auch <code>null</code> &uuml;bergeben werden.
     *
     * @param other Die Plane, die &uuml;berpr&uuml;ft werden soll.
     * @param otherTrans Transformation, um <code>other</code> in das lokale
     *        Koordinatensystem von <code>this</code> zu transformieren.
     * @param otherPointList Liste von Punkten eines Polygons. Die Punkte
     *                       m&uuml;ssen gegen den Uhrzeigersinn angeordnet sein,
     *                       bez&uuml;glich des Normalenvektors dieser Plane
     *                       (Rechte-Hand-Regel). Die Liste wird durch die
     *                       Funktion ge&auml;ndert.
     * @param thisPointList Liste von Punkten eines Polygons. Die Punkte
     *                      m&uuml;ssen gegen den Uhrzeigersinn angeordnet sein,
     *                      bez&uuml;glich des Normalenvektors dieser Plane
     *                      (Rechte-Hand-Regel). Die Liste wird durch die
     *                      Funktion ge&auml;ndert.
     * @param entryList In dieser Liste werden die Eintrittpunkte von
     *                  <code>thisPointList</code> in <code>otherPointList</code>
     *                  gespeichert.
     * @param stopping <code>true</code> wenn die Funktion beim ersten Auftreten
     *                 eines Schnitts abbrechen soll. Die Funktion gibt in
     *                 diesem Fall <code>INTERSECTED</code> zur&uuml;ck. Die Liste
     *                 <code>entryList</code> ist leer.<br \>
     *                 Ist dieser Parameter <code>false</code> so wird die
     *                 Funktion komplett abgearbeitet.  
     * @return Folgende Bits k&ouml;nnen gesetzt sein:
     *            <ul>
     *               <li>IN, wenn other komplett in this liegt.</li>
     *               <li>INCLUDING, wenn this in other liegt.</li>
     *               <li>EQUAL, wenn die beiden Planes genau aufeinander liegen.</li>
     *               <li>CROSSING, wenn die beiden Planes in verschiedenen Ebenen
     *                   liegen und sich schneiden.</li>
     *               <li>INTERSECTED, wenn die Planes in der selben Ebene liegen
     *                   und sich teilweise &uuml;berlagern.</li>
     *               <li>SAME_DIRECTION, wenn die Ebenen parallel
     *               zueinander liegen und die Normalen die gleiche Richtung
     *               haben.</li>
     *            </ul>
     *            Wenn die Planes in keinem Kontakt zueinander stehen, ist kein
     *            Bit gesetzt, es wird also 0 zur&uuml;ckgegeben.
     * @throws CapabilityNotSetException Wenn die Plane Teil eines lebenden Szenegraphen
     *            ist und folgende Capability <i>nicht</i> gesetzt ist:
     *            <ul>
     *               <li>{@linkplain Node#ALLOW_BOUNDS_READ}</li>
     *            </ul>
     * @throws SceneGraphCycleException Wenn sich die Plane in einem zyklisch
     *            geschlossenen Szenegraph befindet.
     * @throws IllegalArgumentException Wenn die Plane selbst &uuml;bergeben wurde.
     * @see #Plane(Point3f[])
    */
    private int getState(
        Plane other,
        Transform3D otherTrans,
        List otherPointList,
        List thisPointList,
        List entryPointList,
        boolean stopping)
        throws
            CapabilityNotSetException,
            SceneGraphCycleException,
            IllegalArgumentException
    {
        int result = 0;
        Transform3D otherTransInverse = new Transform3D();

        // Soll die Plane die Beziehung mit sich selbst ueberpruefen?
        if (other == this)
        {
            throw new IllegalArgumentException("Plane kann nicht mit sich selber berprft werden.");
        }

        // Ist diese Plane oder other noch nicht initialisiert?
        if (this.point == null || other.point == null)
        {
            return 0;
        }

        // Schneiden sich die Bounds?
        if (this.boundsIntersected(other, otherTrans))
        {
            // Auf Parallelitaet pruefen
            if (this.isParallel(other, otherTrans))
            {
                // Liegen die Planes in der selben Ebene?
                if (this.isParallelEqualLevel(other, otherTrans))
                {
                    // List von Punkten erstellen
                    if (thisPointList == null)
                        thisPointList = new ArrayList();
                    if (otherPointList == null)
                        otherPointList = new ArrayList();
                    if (entryPointList == null)
                        entryPointList = new ArrayList();

                    this.putPointsInList(thisPointList, null);
                    other.putPointsInList(otherPointList, otherTrans);

                    Vector3f glbOtherNormal = new Vector3f();
                    otherTrans.transform(other.normal, glbOtherNormal);

                    // Ist die Punktliste von other im Urzeigersinn angeordnet?
                    // Oder anders: Sind die Normalenvektoren entgegengesetzt?
                    if (this.normal.dot(glbOtherNormal) < 0)
                    {
                        // Umlaufrichtung aendern
                        Collections.reverse(otherPointList);
                    } else
                    {
                        result = Plane.SAME_DIRECTION;
                    }

                    // Ueberschneiden sich die Polygone?
                    if (this
                        .createIntersectionPointList(
                            otherPointList,
                            thisPointList,
                            entryPointList,
                            stopping)
                        == true)
                    {
                        return INTERSECTED | result;
                    }

                    int state = this.getPositionInfo(other, otherTrans);

                    switch (state)
                    {
                        case IN :
                            // other liegt in this.
                            return IN | result;
                        case OUT :
                            // Noch ueberpruefen, ob this in other liegt
                            otherTransInverse.invert(otherTrans);
                            if (other.getPositionInfo(this, otherTransInverse)
                                == IN)
                            {
                                // this liegt in other
                                return INCLUDING | result;
                            }
                            break;
                        case EQUAL :
                            // other und this stimmen ueberein.
                            return EQUAL | result;
                        default :
                            assert false;
                    }
                }
            } else // if (this.isParallel(other))
                {
                // Planes auf Ueberschneidungen pruefen
                ArrayList otherCrossPointList = new ArrayList();
                ArrayList thisCrossPointList = new ArrayList();
                Point3f startPoint, endPoint, otherCenterPoint, thisCenterPoint;
                float otherRadius, thisRadius, centerPointDistance;
                startPoint = new Point3f();
                endPoint = new Point3f();
                otherCenterPoint = new Point3f();
                thisCenterPoint = new Point3f();
                if (checkForCrossing(other, otherTrans, otherCrossPointList)
                    == CROSSING)
                {
                    otherTransInverse.invert(otherTrans);
                    if (other
                        .checkForCrossing(
                            this,
                            otherTransInverse,
                            thisCrossPointList)
                        == CROSSING)
                    {
                        // Segment vergleichen
                        for (Iterator otherPoints =
                            otherCrossPointList.iterator();
                            otherPoints.hasNext();
                            )
                        {
                            startPoint = (Point3f) otherPoints.next();
                            endPoint = (Point3f) otherPoints.next();
                            // Mittelpunkt und Radius berechnen                            
                            otherCenterPoint.add(startPoint, endPoint);
                            otherCenterPoint.scale(0.5f);
                            otherRadius = startPoint.distance(endPoint) / 2;
                            for (Iterator thisPoints =
                                thisCrossPointList.iterator();
                                thisPoints.hasNext();
                                )
                            {
                                startPoint = (Point3f) thisPoints.next();
                                endPoint = (Point3f) thisPoints.next();
                                // Mittelpunkt und Radius berechnen
                                thisCenterPoint.add(startPoint, endPoint);
                                thisCenterPoint.scale(0.5f);

                                // Punkt nach this transformieren
                                otherTrans.transform(thisCenterPoint);

                                thisRadius = startPoint.distance(endPoint) / 2;

                                centerPointDistance =
                                    otherCenterPoint.distance(thisCenterPoint);
                                // Abstand der Mittelpunkte kleiner als die Summe der Radien?
                                if (centerPointDistance
                                    < thisRadius + otherRadius)
                                {
                                    return CROSSING; // Planes schneiden sich
                                }
                            } // for (Enumeration thisSegments =...
                        } // for (Enumeration otherSegments =...
                    } // if (other.checkForCrossing(this, thisSegmentList) == CROSSING)
                } // if (checkForCrossing(other, otherSegmentList) == CROSSING)
            } // if (this.isParallel(other)) else
        } // if (this.boundsIntersected(other))

        return 0; // Kein Kontakt
    }

    /**
     * Ermittelt, ob sich der &uuml;bergebene Punkt innerhalb der Plane befindet.  
     * Algorithmus nach <a href="http://rw7.de/ralf/inffaq/polygon.html">Ralf
     * Wiebicke</a>.
     * 
     * @param point Punkt in lokalen Koordinaten der Plane.
     * @return <code>EQUAL</code>, wenn sich der Punkt auf einer Kante des
     *            Planepolygons befindet, <code>IN</code>, wenn sich der Punkt
     *            innerhalb der Plane befindet, sonst <code>OUT</code>.
     */
    public int isPointIn(Point3f point)
    {
        assert this.isDataCorrect(this.point);
        assert this.point != null;

        return isPointIn(this.point, point);
    }

    /**
     * Ermittelt, ob sich der &uuml;bergebene Punkt innerhalb des durch
     * <code>pointArray</code> definierten Polygons befindet. Die Punkte
     * in <code>pointArray</code> m&uuml;ssen alle in einer Ebene liegen, sonst
     * liefert die Funktion kein richtiges Ergebnis.  
     * Algorithmus nach <a href="http://rw7.de/ralf/inffaq/polygon.html">Ralf
     * Wiebicke</a>.
     * 
     * @param pointArray Array von Punkten, welche das Polygon beschreiben.
     * @param point Punkt in lokalen Koordinaten der Plane.
     * @return <code>EQUAL</code>, wenn sich der Punkt auf einer Kante des
     *            Polygons befindet, <code>IN</code>, wenn sich der Punkt
     *            innerhalb des Polygons befindet, sonst <code>OUT</code>.
     */
    private int isPointIn(Point3f[] pointArray, Point3f point)
    {
        assert this.isDataCorrect(pointArray);

        // Ein Segment des Polygons
        Segment polygonSegment = new Segment();

        // Kontrollieren, ob der Punkt auf einer Kante liegt
        int j;
        for (int i = 0; i < pointArray.length; ++i)
        {
            j =
                (i + 1 >= pointArray.length)
                    ? (i + 1 - pointArray.length)
                    : i + 1;
            polygonSegment.set(pointArray[i], pointArray[j]);
            // Liegt der Punkt auf der Kante?
            if (polygonSegment.isPointIn(point))
            {
                return EQUAL;
            }
        }

        // Richtungsvektor ermitteln
        Vector3f dirVec = new Vector3f();
        dirVec.sub(pointArray[1], pointArray[0]);

        // Strahl / Gerade erstellen
        Segment line;
        line = new Segment(point, dirVec);

        // Mit Punkt beginnen, der nicht auf line liegt
        int startIndex;
        for (startIndex = 0; startIndex < pointArray.length; ++startIndex)
        {
            if (Float.isNaN(line.getScaleFactor(pointArray[startIndex])))
            {
                // Punkt liegt nicht auf der Geraden -> Schleife beenden
                break;
            }
        }

        float[] scaleFactor = new float[2];
        Point3f crossPoint = new Point3f();
        float s;

        // Durch alle Punkte des Polygons wandern
        int counter = 0;
        int lastIndex = startIndex;
        int currIndex;
        boolean ignore = false;
        for (int i = startIndex + 1; i <= pointArray.length + startIndex; ++i)
        {
            if (i >= pointArray.length)
                currIndex = i - pointArray.length;
            else
                currIndex = i;

            // Kante setzen
            polygonSegment.set(pointArray[lastIndex], pointArray[currIndex]);

            s = line.getScaleFactor(pointArray[currIndex]);

            // Wurde der letzte Punkt ignoriert?
            if (ignore)
            {
                // Liegt point[currIndex] nicht auf der Geraden line?
                if (Float.isNaN(s))
                {
                    float result =
                        line.getCrossInfo(
                            polygonSegment,
                            epsilon,
                            scaleFactor,
                            crossPoint);

                    // Schneidet die Strecke polygonSegment die Gerade line?
                    if ((result == Segment.LINE_CROSSING
                        || result == Segment.SEGMENT_CROSSING
                        && scaleFactor[1] > 0
                        && scaleFactor[1] < 1))
                    {
                        // Liegt point auf einer Kante?
                        counter++;
                    }
                    ignore = false;
                    lastIndex = currIndex;
                }
            } else // if (ignore)
                {
                // Liegt der Punkt point[currIndex] auf dem Strahl line?
                if (Float.isNaN(s) == false && s > 0)
                {
                    // Punkt ignorieren
                    ignore = true;
                } else
                {
                    int result =
                        line.getCrossInfo(
                            polygonSegment,
                            epsilon,
                            scaleFactor,
                            crossPoint);

                    // Schneidet die Strecke polygonSegment den Strahl line?
                    if ((result == Segment.LINE_CROSSING
                        || result == Segment.SEGMENT_CROSSING)
                        && scaleFactor[0] >= 0
                        && scaleFactor[1] > 0
                        && scaleFactor[1] < 1)
                    {
                        counter++;
                    }
                    lastIndex = currIndex;
                } // if (Float.isNaN(s) == false && s > 0) else
            } // if (ignore) else
        } // for (int i = startIndex + 1; ...

        return (counter % 2 == 0) ? OUT : IN;
    }

    /**
     * Gibt die Fl&auml;chennormale der Plane zur&uuml;ck.
     * 
     * @param normal Vektor, in den die Werte der Normalen geschrieben werden.
     * @see #Plane(Point3f[])
     */
    public void getNormal(Vector3f normal) throws IllegalStateException
    {
        if (this.point == null)
            throw new IllegalStateException("Plane ist nicht initialisiert.");

        normal.set(this.normal);
    }

    /**
     * Registriert einen PlaneListener. Ein PlaneListener wird von der Plane
     * aufgerufen, wenn ein definiertes Ereignis eintritt.
     * 
     * @param listener Listener Objekt, welches bei der Plane registriert werden
     *           soll.
     * @throws NullPointerException Wenn <code>null</code> &uuml;bergeben wurde.
     */
    public void addPlaneListener(PlaneListener listener)
        throws NullPointerException
    {
        if (listener == null)
            throw new NullPointerException("Argument ist null.");

        this.listenerList.add(listener);
    }

    /**
     * L&ouml;scht einen PlaneLister, so dass er keine Ereignise mehr entf&auml;ngt.
     * 
     * @param listener Listener Objekt, welches gel&ouml;scht werden soll.
     * @throws NullPointerException Wenn <code>null</code> &uuml;bergeben wurde.
     */
    public void removePlaneListener(PlaneListener listener)
        throws NullPointerException
    {
        if (listener == null)
            throw new NullPointerException("Argument ist null.");

        this.listenerList.remove(listener);
    }

    /**
     * &Uuml;berpr&uuml;ft, ob sich other in this, ausserhalb von this, oder genau
     * auf this befindet.</br> 
     * Die Funktion liefert nur richtige Ergebnisse, wenn die Planes in einer Ebene
     * liegen und sich nicht &uuml;berschneiden.
     * 
     * @param other Die zu &uuml;berpr&uuml;fende Plane.
     * @param otherTrans Transformation, um <code>other</code> in das lokale
     *        Koordinatensystem von <code>this</code> zu transformieren.
     * @return Folgende Werte k&ouml;nnen zur&uuml;ck gegeben werden:
     *         <ul>
     *            <li>IN, wenn sich other in this befindet,</li>
     *            <li>OUT, wenn sich other ausserhalb von this befindet, oder</li>
     *            <li>EQUAL, wenn other ganau auf this liegt.</li>
     *         </ul>
     */
    private int getPositionInfo(Plane other, Transform3D otherTrans)
    {
        assert this.point != null && other.point != null;
        assert this.isParallel(other, otherTrans)
            && this.isParallelEqualLevel(other, otherTrans);

        Point3f otherPoint = new Point3f();

        int j;
        for (int i = 0; i < other.point.length; ++i)
        {
            j = (i + 1 == other.point.length) ? 0 : i + 1;
            // Mittelpunkt der Strecke ermitteln
            otherPoint.add(other.point[i], other.point[j]);
            otherPoint.scale(0.5f);

            // Punkt von other transformieren
            otherTrans.transform(otherPoint);

            // Position des Punktes abfragen
            switch (this.isPointIn(otherPoint))
            {
                case IN :
                    return IN; // Punkt liegt in this -> other liegt in this
                case OUT :
                    return OUT;
                    // Punkt liegt ausserhalb -> other liegt ausserhalb
            }
        }

        return EQUAL; // Alle Punkte liegen auf this -> Planes sind gleich
    }

    /**
     * &Uuml;berpr&uuml;ft, ob sich die Punktliste <code>point2</code> in,
     * ausserhalb, oder genau auf der Punktliste <code>point1</code> befindet.
     * </br> 
     * Die Funktion liefert nur richtige Ergebnisse, wenn die beiden Listen in
     * einer Ebene liegen und sich nicht &uuml;berschneiden.
     * 
     * @param point1 Liste von Punkten, die ein Polygon beschreiben.
     * @param point2 Liste von Punkten, die ein Polygon beschreiben.
     * @return Folgende Werte k&ouml;nnen zur&uuml;ck gegeben werden:
     *         <ul>
     *            <li>IN, wenn sich point2 in point1 befindet,</li>
     *            <li>OUT, wenn sich point2 ausserhalb von point1 befindet,
     *                oder</li>
     *            <li>EQUAL, wenn point2 ganau auf point1 liegt.</li>
     *         </ul>
     */
    private int getPositionInfo(Point3f[] point1, Point3f[] point2)
    {
        Point3f otherPoint = new Point3f();

        int j;
        for (int i = 0; i < point2.length; ++i)
        {
            j = (i + 1 == point2.length) ? 0 : i + 1;
            // Mittelpunkt der Strecke ermitteln
            otherPoint.add(point2[i], point2[j]);
            otherPoint.scale(0.5f);

            // Position des Punktes abfragen
            switch (this.isPointIn(point1, otherPoint))
            {
                case IN :
                    return IN; // Punkt liegt in this -> other liegt in this
                case OUT :
                    return OUT;
                    // Punkt liegt ausserhalb -> other liegt ausserhalb
            }
        }

        return EQUAL; // Alle Punkte liegen auf this -> Planes sind gleich
    }

    /**
     * <i>Kopiert</i> die Punkte der Plane in die Liste <code>pointList</code> und
     * transformiert das Ergebnis noch mit der Transformation <code>transform</code>.
     * 
     * @param pointList Die Liste, in welche die Punkte kopiert werden.
     * @param transform Die Transformation, die auf die Punkte angewendet wird.
     */
    private void putPointsInList(List pointList, Transform3D transform)
    {
        assert this.point != null;

        Point3f globalPoint;

        pointList.clear();
        for (int i = 0; i < this.point.length; ++i)
        {
            globalPoint = new Point3f(this.point[i]);
            if (transform != null)
                transform.transform(globalPoint);
            pointList.add(globalPoint);
        }
    }

    /**
     * &Uuml;berpr&uuml;ft, ob <code>other</code> die Ebene von <code>this</code>
     * schneidet.
     * (Idee: Wie Dreieck-Polygonschnitt)
     * 
     * @param other Die zu &uuml;berpr&uuml;fende Plane.
     * @param otherTrans Transformation, um <code>other</code> in das lokale
     *        Koordinatensystem von <code>this</code> zu transformieren.
     * @param crossPointList Liste, in der, bei Schneiden der Ebene, die
     *                    Schnittpunkte gespeichert werden. Der bisherige Inhalt
     *                    der Liste wird gel&ouml;scht.
     * @return <code>IN</code>, wenn other in der selben Ebene liegt wie this.</br>
     *         <code>CROSSING</code>, wenn other die Ebene schneidet.
     *         Es wird 0 zur&uuml;ckgegeben, wenn other die Ebene nicht schneidet
     *         und auch nicht in ihr liegt.
     */
    private int checkForCrossing(
        Plane other,
        Transform3D otherTrans,
        List crossPointList)
    {
        assert this.point != null && other.point != null;

        // Liste loeschen
        crossPointList.clear();

        /* Variablen *
         * --------- */
        Vector3f otherPointVector = new Vector3f();
        Point3f otherPoint = new Point3f();
        Point3f lastOtherPoint = new Point3f();
        Point3f crossPoint = new Point3f();

        Vector3f lastVector = new Vector3f();
        Vector3f nextVector = new Vector3f();
        Vector3f crossVector = new Vector3f();

        Vector4f plane = new Vector4f();
        float scaleFactor[] = new float[1];
        Segment segment = new Segment();
        float distance = 0, lastDistance = 0;

        /* Vorbereitung *
         * ------------ */

        // Ebenenwerte speichern
        plane.set(normal.x, normal.y, normal.z, 0);

        // Letzten Punkt von other in Ebenengleichung von this einsetzen
        // Punkt transformieren
        otherTrans.transform(other.point[other.point.length - 1], otherPoint);
        otherPointVector.set(otherPoint);

        // Punkt einsetzen und Entfernung zur Ebene ermitteln
        lastDistance = this.normal.dot(otherPointVector);
        lastDistance = (lastDistance == 0) ? 0 : ((lastDistance > 0) ? 1 : -1);

        /* Schnittpunkte ermitteln *
         * ----------------------- */
        // Alle Punkte von other in Ebenengleichung von this einsetzen
        for (int i = 0; i < other.point.length; ++i)
        {
            // Punkt transformieren
            otherTrans.transform(other.getPoint(i), otherPoint);
            otherPointVector.set(otherPoint);

            // Punkt einsetzen und Entfernung zur Ebene ermitteln
            distance = this.normal.dot(otherPointVector);
            distance = (distance == 0) ? 0 : ((distance > 0) ? 1 : -1);

            // Liegt der Punkt in der Ebene?
            if (distance == 0)
            {
                // Vektoren berechnen
                lastVector.sub(other.getPoint(i), other.getPoint(i - 1));
                nextVector.sub(other.getPoint(i + 1), other.getPoint(i));

                crossVector.cross(lastVector, nextVector);

                if (crossVector.dot(normal) < 0)
                {
                    // Punkt speichern
                    crossPointList.add(new Point3f(otherPointVector));
                }
            } else
            {
                // Lag der letzte Punkt auf der anderen Seite?
                if (lastDistance != 0 && distance != lastDistance)
                {
                    assert distance != 0;

                    // Schnittpunkt berechnen von Ebene und Gerade (lastOtherPoint <-> otherPoint)
                    otherPoint.set(otherPointVector);
                    // Punkt transformieren
                    otherTrans.transform(other.getPoint(i - 1), lastOtherPoint);

                    segment.set(lastOtherPoint, otherPoint);

                    int result =
                        segment.getCrossInfo(plane, scaleFactor, crossPoint);
                    // Ebene wird in diesem Fall immer geschnitten!
                    assert result == Segment.SEGMENT_CROSSING;

                    // Punkt speichern
                    crossPointList.add(new Point3f(crossPoint));
                } // if (lastDistance != 0 && distance != lastDistance)

                // Entfernung fuer naechsten Durchlauf sichern
                lastDistance = distance;
            } // if (distance == 0) else
        } // for (int i = 0; i < other.point.length; ++i)

        /* Nachbereitung *
         * ------------- */

        // Liegen alle Punkte von other in der Ebene von this?
        if (crossPointList.size() == 0 && distance == 0)
        {
            return IN; // other liegt in der Ebene von this
        }

        // Gibt es Schnittpunkte?
        if (crossPointList.size() > 0)
        {
            // Zwei Punkte aus Liste herausholen, um Schnittgerade zu ermitteln
            otherPoint.set((Point3f) crossPointList.get(0));
            lastOtherPoint.set(
                (Point3f) crossPointList.get(crossPointList.size() - 1));

            comperator.set(otherPoint, lastOtherPoint);

            // Punkte sortieren
            Collections.sort(crossPointList, comperator);

            assert crossPointList.size() % 2 == 0 : crossPointList.size();

            return CROSSING; // other kreuzt die Ebene von this
        }

        return 0; // Kein Kontakt
    }

    /**
     * &Uuml;berpr&uuml;ft die Punktliste auf Korrektheit. Eine Punktliste gilt
     * als Korrekt, wenn:
     * <ul>
     *    <li>mindestens 3 Punkte enthalten sind,</li>
     *    <li>sich keine Kanten schneiden.</li>
     * </ul>
     * Es wird nicht &uuml;berpr&uuml;ft, ob die Punkte alle in einer Ebene liegen.
     * Dies wird in der Methode {@linkplain #set(Point3f[])} kontrolliert, wo die
     * Punkte kopiert werden.
     * 
     * @param point Die zu &uuml;berpr&uuml;fende Liste von Punkten.
     * @return <code>true</code>, wenn die Liste korrekt ist, sonst <code>false</code>.
     */
    private boolean isDataCorrect(Point3f[] point)
    {
        // Zuwenig Punkte?
        if (point.length < 3)
        {
            return false;
        }

        Segment firstSegment = new Segment();
        Segment secondSegment = new Segment();
        float[] scaleFactor = new float[2];
        Point3f crossPoint = new Point3f();
        int iFirst, iSecond, jFirst, jSecond;

        // Jede Kante mit jeder vergleichen
        for (int i = 0; i < point.length; ++i)
        {
            iFirst = i;
            iSecond = (i + 1 < point.length) ? i + 1 : i + 1 - point.length;

            firstSegment.set(point[iFirst], point[iSecond]);
            for (int j = i; j < point.length + i; ++j)
            {
                jFirst = (j < point.length) ? j : j - point.length;
                jSecond = (j + 1 < point.length) ? j + 1 : j + 1 - point.length;

                secondSegment.set(point[jFirst], point[jSecond]);

                if (firstSegment
                    .getCrossInfo(
                        secondSegment,
                        epsilon,
                        scaleFactor,
                        crossPoint)
                    == Segment.SEGMENT_CROSSING)
                {
                    /*
                     * Ein Schneiden der beiden Segmente ist nur erlaubt, wenn
                     * sich die beiden Segmente jeweils an ihren Enden beruehren
                     * und die beiden Segmente in der Liste aufeinander folgen.
                     */
                    // Liegt ein unerlaubter Schnitt vor?
                    if (((scaleFactor[0] == 1
                        && scaleFactor[1] == 0
                        && iSecond == jFirst)
                        || (scaleFactor[0] == 0
                            && scaleFactor[1] == 1
                            && iFirst == jSecond))
                        == false)
                    {
                        // Punktliste ist nicht korrekt.
                        return false;
                    }
                }
            } // for (int j = i; j < point.length + i; ++j)
        } // for (int i = 0; i < point.length; ++i)

        // Punktliste ist korrekt
        return true;
    }

    /**
     * Erstellt ein Appearance Objekt in welchem die PolygonAttributes, das Material
     * und die TextureAttributes gesetzt sind. 
     */
    private Appearance createAppearance()
    {
        Appearance app = new Appearance();
        app.setPolygonAttributes(
            new PolygonAttributes(
                PolygonAttributes.POLYGON_FILL,
                PolygonAttributes.CULL_BACK,
                0,
                false));
        app.setMaterial(
            new Material(
                new Color3f(0.4f, 0.4f, 0.4f),
                new Color3f(0, 0, 0),
                new Color3f(1, 1, 1),
                new Color3f(0, 0, 0),
                128));
        app.setTextureAttributes(
            new TextureAttributes(
                TextureAttributes.MODULATE,
                new Transform3D(),
                new Color4f(),
                TextureAttributes.FASTEST));

        return app;
    }

    /**
     * Erstellt die Geometrie f&uuml; die Plane. 
     */
    private Geometry createGeometry() throws GeometryNotCreated
    {
        assert this.point != null && isDataCorrect(this.point);

        // Benoetigte lokale Objekte anlegen
        GeometryInfo gI = new GeometryInfo(GeometryInfo.POLYGON_ARRAY);
        NormalGenerator nG = new NormalGenerator();
        Stripifier s = new Stripifier();

        // Anzahl der Polygone ermitteln (Umrandung + Loecher)
        int stripCountsSize = 1 + holeList.size();

        // In diesem Array werden die Anzahl der Punkte je Polygon gespeichert
        int stripCounts[] = new int[stripCountsSize];

        // Liste aller Punkte.
        ArrayList allPointsVector = new ArrayList();

        // Punkte der Umrandung speichern
        allPointsVector.addAll(Arrays.asList(this.point));
        stripCounts[0] = this.point.length;
        int stripCountsIndex = 0;
        // Punkte der Loecher speichern
        for (Iterator polygonLists = holeList.iterator();
            polygonLists.hasNext();
            )
        {
            // Punktarray des akt. Polygons ermitteln
            Point3f[] pointArray = ((Point3f[]) polygonLists.next());
            ++stripCountsIndex;
            stripCounts[stripCountsIndex] = pointArray.length;
            // Punkte speichern
            allPointsVector.addAll(Arrays.asList(pointArray));
        }

        // In diesem Array werden alle Punkte gespeichert
        Point3f allPointsArray[] = new Point3f[allPointsVector.size()];
        // Punkte in Array kopieren
        allPointsVector.toArray(allPointsArray);

        // Koordinaten setzen (Daten werden NICHT kopiert)
        gI.setCoordinates(allPointsArray);
        gI.setContourCounts(new int[] { stripCountsSize });
        gI.setStripCounts(stripCounts);

        // Texturkoordinaten berechnen und setzen
        gI.setTextureCoordinateParams(1, 2);
        float[] texCoord = new float[2 * allPointsArray.length];
        calculateTextureCoordinates(texCoord, allPointsArray);
        gI.setTextureCoordinates(0, texCoord);

        try
        {
            // Alle erstellten Dreiecke auf Korrektheit ueberpruefen
            TriangleArray gA = (TriangleArray) gI.getGeometryArray();
            int vertexCount = gA.getVertexCount();
            Point3f newPointArray[] = new Point3f[vertexCount];
            TexCoord2f newTexCoordArray[] = new TexCoord2f[vertexCount];

            Point3f coordTripple[] =
                new Point3f[] { new Point3f(), new Point3f(), new Point3f()};
            TexCoord2f texCoordTripple[] =
                new TexCoord2f[] {
                    new TexCoord2f(),
                    new TexCoord2f(),
                    new TexCoord2f()};
            Vector3f a = new Vector3f();
            Vector3f b = new Vector3f();

            int writeIndex = 0;
            for (int i = 0; i < vertexCount; i += 3)
            {
                gA.getCoordinates(i, coordTripple);
                a.sub(coordTripple[1], coordTripple[0]);
                b.sub(coordTripple[2], coordTripple[0]);
                a.cross(a, b);

                // Bilden die Punkte ein Dreieck?
                if (a.lengthSquared() != 0)
                {
                    // Koordinaten und Texturkoordinaten kopieren
                    gA.getTextureCoordinates(0, i, texCoordTripple);
                    newPointArray[writeIndex] = new Point3f(coordTripple[0]);
                    newPointArray[writeIndex + 1] =
                        new Point3f(coordTripple[1]);
                    newPointArray[writeIndex + 2] =
                        new Point3f(coordTripple[2]);
                    newTexCoordArray[writeIndex] =
                        new TexCoord2f(texCoordTripple[0]);
                    newTexCoordArray[writeIndex + 1] =
                        new TexCoord2f(texCoordTripple[1]);
                    newTexCoordArray[writeIndex + 2] =
                        new TexCoord2f(texCoordTripple[2]);

                    writeIndex += 3;
                }
            }

            // Spannten die Punkte keine Flaeche auf?
            if (writeIndex == 0)
            {
                throw new GeometryNotCreated();
            } else
                if (writeIndex != vertexCount)
                {
                    // Neue Daten in gI einlesen
                    TriangleArray newTrAr =
                        new TriangleArray(
                            writeIndex,
                            TriangleArray.COORDINATES
                                | TriangleArray.TEXTURE_COORDINATE_2);
                    newTrAr.setCoordinates(0, newPointArray, 0, writeIndex);
                    newTrAr.setTextureCoordinates(
                        0,
                        0,
                        newTexCoordArray,
                        0,
                        writeIndex);

                    gI = new GeometryInfo(newTrAr);
                }
        } catch (ClassCastException cE)
        {
            // Wenn gI.getGeomeryArray kein TriangleArray zurueck gibt.
            // In diesem Fall sind die Normalen evt. fehlerhaft.
            assert false;
        }

        // Normalen berechnen
        nG.generateNormals(gI);

        // Dreiecke ggf. verbinden
        s.stripify(gI);

        IndexedGeometryArray iGA = gI.getIndexedGeometryArray(true);

        // Normale ermitteln und speichern
        Vector3f tmpNormal = new Vector3f();
        for (int i = 0; i < iGA.getIndexCount(); ++i)
        {
            if (i == 0)
                iGA.getNormal(iGA.getNormalIndex(i), this.normal);
            else
            {
                iGA.getNormal(iGA.getNormalIndex(i), tmpNormal);
                if (tmpNormal.equals(this.normal) == false)
                {
                    throw new GeometryNotCreated();
                }
            }
        }

        // Listener ueber Neuerstellung informieren
        fireGeometryCreatedEvent(iGA);

        return iGA;
    }

    /**
     * Benachrichtigt alle registrierten PlaneListener &uuml;ber die Neuerstellung
     * der Geoemtrie.
     * 
     * @param iGA Die neue Geoemtrie.
     */
    private void fireGeometryCreatedEvent(IndexedGeometryArray iGA)
    {
        for (Iterator listenerIterator = this.listenerList.iterator();
            listenerIterator.hasNext();
            )
        {
            ((PlaneListener) listenerIterator.next()).newGeometryCreated(iGA);
        }
    }

    /**
     * Errechnet fuer jeden &uuml;bergebenen Punkt die Texturekoordinaten.
     *  
     * @param texCoord Array, in dem die Texturekoordinaten gespeichert werden sollen.
     *           Da f&uuml;r jeden Punkt zwei Werte gespeichert werden, muss das
     *           Array doppelt so gross sein, wie das Punktarray <code>points</code>.
     * @param points Array, welches die Punkte enth&auml;lt, zu denen die
     *           Texturkoordinaten berechnet werden sollen.
     * 
     * @throws IllegalArgumentException Wenn die Punkte <i>keine</i>
     *            Ebene aufspannen.
     */
    private void calculateTextureCoordinates(
        float[] texCoord,
        Point3f[] points)
        throws
            IndexOutOfBoundsException,
            ArithmeticException,
            IllegalArgumentException
    {
        assert this.point != null;

        // Erste Achse fuer 2D Texturekoordinate
        Vector3f axisOne = new Vector3f();
        // Zweite Achse fuer 2D Texturekoordinate
        Vector3f axisTwo = new Vector3f();
        // Provisorische Normale
        Vector3f normal = new Vector3f();

        // Provisorische Normale der Ebene berechnen
        axisOne.sub(points[0], points[1]);
        for (int i = 1; i < points.length; ++i)
        {
            axisTwo.sub(
                points[i],
                points[(i + 1 == points.length) ? 0 : i + 1]);
            normal.cross(axisOne, axisTwo);
            if (normal.lengthSquared() != 0);
            break;
        }
        if (normal.lengthSquared() == 0)
            throw new IllegalArgumentException("Punkte spannen keine Ebene auf");

        Vector3f pointVector = new Vector3f();

        // Erste Ebene ist xz-Ebene -> n = (0, 1, 0)
        pointVector.set(0, 1, 0);
        axisOne.cross(normal, pointVector);
        // Liegt die Plane in der xz-Ebene?
        if (axisOne.lengthSquared() == 0)
        {
            // Zweite Ebene ist xy-Ebene -> n = (0, 0, 1)
            pointVector.set(0, 0, 1);
            axisOne.cross(normal, pointVector);
            assert axisOne.lengthSquared() == 0;
        }

        axisTwo.cross(normal, axisOne);
        axisOne.normalize();
        axisTwo.normalize();

        // Texturkoorinaten berechnen
        for (int j = 0; j < points.length; ++j)
        {
            pointVector.set(points[j]);
            texCoord[j * 2] = axisOne.dot(pointVector) / 5f;
            texCoord[j * 2 + 1] = axisTwo.dot(pointVector) / 5f;
        }
    }

    /**
     * Transformiert die in der List <code>srcPolygonList</code> enthaltenen
     * Punkte und speichert diese in der Liste <code>destPolygonList</code>.
     * Die Punkte sind dabei in mehreren Punktarrays gespeichert,
     * welche die Liste folgendermassen verwalten:
     * <pre>
     *    Point3f pointOne = ( (Point3f[]) srcPolygonList.get(0) ) [0];
     * </pre>
     * 
     * @param trans Die Transformation, die auf die Punkte angewendet wird.
     * @param srcPolygonList Liste, welche mehrere Punktarray enth&auml;lt, wie oben
     *           beschrieben. Die Punkte sind global.
     * @param destPolygonList Liste, welche nach Abarbeiten der Methode genau soviele
     *           Punktarrays enth&auml;lt wie <code>srcPolygonList</code>, jedoch
     *           sind die darin gespeicherten Punkte in das lokale Koordinatensystem
     *           von this transformiert wurden.
     * @throws RestrictedAccessException Wenn die Plane Teil eines kompilierten, aber
     *            nicht lebendigen Szenegraphen ist.
     * @throws ClassCastException Wenn die Liste <code>srcPolygonList</code> keine,
     *            oder nicht nur, <code>Point3f[]</code> Typen enth&auml;lt.
     */
    private void transform(
        Transform3D trans,
        List srcPolygonList,
        List destPolygonList)
        throws RestrictedAccessException, ClassCastException
    {
        Point3f[] srcPointArray, destPointArray;

        destPolygonList.clear();
        // Liste vollstaendig kopieren (tiefe Kopie)
        for (int i = 0; i < srcPolygonList.size(); ++i)
        {
            srcPointArray = (Point3f[]) srcPolygonList.get(i);
            destPointArray = new Point3f[srcPointArray.length];
            for (int j = 0; j < srcPointArray.length; ++j)
            {
                // Punkt kopieren
                destPointArray[j] = new Point3f(srcPointArray[j]);
                // Punkt transformieren
                trans.transform(destPointArray[j]);
            }
            // Array speichern
            destPolygonList.add(destPointArray);
        }
    }

    /**
     * Speichert die Punktarrayliste <code>polygonList</code> ab und erstellt
     * mit den neuen Werten die Geoemtrie neu. Die Punktarrays beschreiben
     * L&ouml;cher in der Plane. Die Punktarrays m&uuml;ssen in der selben
     * Richtung angeordnet sein, wie die Punkte von this.
     *  
     * @param polygonList Liste von Point3f[]. Also:
     *        <pre>
     *           Point3f[] pointArray = (Point3f[]) polygonList.get(0);
     *        </pre>
     *        Die Punktarrays m&uuml;ssen in der selben Richtung angeordnet
     *        sein, wie die Punkte von this.
     * @throws CapabilityNotSetException Wenn die Capability
     *            {@linkplain Shape3D#ALLOW_GEOMETRY_WRITE} <i>nicht</i> gesetzt ist
     *            und die Plane Teil eines lebenden Szenegraphen ist.
     */
    private void createHole(List polygonList) throws CapabilityNotSetException
    {
        assert this.point != null;

        holeList.addAll(polygonList);

        try
        {
            // Geometry neu erstellen
            if (this.numGeometries() > 0)
                this.setGeometry(createGeometry());
            else
                this.addGeometry(createGeometry());
        } catch (GeometryNotCreated gE)
        {
            this.removeAllGeometries();
        }
    }

    /**
     * Speichert das Punktarray <code>polygon</code> ab und erstellt mit den
     * neuen Werten die Geoemtrie neu. Das Punktarray beschreibt ein Loch in
     * der Plane.
     * 
     * @param polygon Punkteliste, welche das Loch beschreibt.
     * @param otherTrans Transformation, welche auf die Punktliste angewendet
     *        wird, um sie in das lokale Koordinatensystem von <code>this</code>
     *        zu transformieren.
     * @throws CapabilityNotSetException Wenn die Plane Teil eines lebenden
     *            Szenegraphen ist und folgende Capabilities <i>nicht</i>
     *            gesetzt sind:
     *            <ul>
     *               <li>{@linkplain Shape3D#ALLOW_GEOMETRY_WRITE}</li>
     *            <ul>
     * @throws RestrictedAccessException Wenn die Plane Teil eines kompilierten,
     *            aber nicht lebendigen Szenegraphen ist.
     * @throws IllegalSharingException Wenn <code>other</code> sich nicht im selben
     *            Szenegraphen befindet wie diese Plane.
     */
    private void createHole(
        Point3f[] polygon,
        Transform3D otherTrans,
        boolean turn)
        throws
            CapabilityNotSetException,
            RestrictedAccessException,
            IllegalSharingException
    {
        assert this.point != null;

        // Polygonliste erstellen
        ArrayList polygonList = new ArrayList();
        // Array kopieren
        Point3f[] pointArray = new Point3f[polygon.length];

        if (turn)
            for (int j = 0; j < polygon.length; ++j)
            {
                // Punkt kopieren
                pointArray[j] = new Point3f(polygon[polygon.length - j - 1]);
                // Punkt transformieren
                if (otherTrans != null)
                    otherTrans.transform(pointArray[j]);
            } else
            for (int j = 0; j < polygon.length; ++j)
            {
                // Punkt kopieren
                pointArray[j] = new Point3f(polygon[j]);
                // Punkt transformieren
                if (otherTrans != null)
                    otherTrans.transform(pointArray[j]);
            }

        // Array speichern
        polygonList.add(pointArray);

        createHole(polygonList);
    }

    /**
     * &Uuml;berpr&uuml;ft, ob diese Plane und die &uuml;bergebene Plane parallel
     * zueinander verlaufen.
     * 
     * @param other Die zu &uuml;berpr&uuml;fende Plane.
     * @param otherTrans Transformation, um <code>other</code> in das lokale
     *        Koordinatensystem von <code>this</code> zu transformieren.
     * @return <code>true</code>, wenn <code>other</code> parallel zu
     *         <code>this</code> verl&auml;uft, sonst <code>false</code>.
     */
    private boolean isParallel(Plane other, Transform3D otherTrans)
    {
        assert this.point != null && other.point != null;
        assert this.normal.lengthSquared() == 1
            && other.normal.lengthSquared() == 1;

        Vector3f tmpVector = new Vector3f();

        otherTrans.transform(other.normal, tmpVector);

        // Sind Planes parallel?
        float dot = this.normal.dot(tmpVector);
        if (dot == -1 || dot == 1)
            return true;

        return false;
    }

    /**
     * &Uuml;berpr&uuml;ft, ob die Bounds von <code>other</code> sich mit der
     * Bounds dieser Plane &uuml;berschneidet.
     * 
     * @param other Die Plane, deren Boundsobjekt auf &Uuml;berschneidung gepr&uuml;ft
     *        werden soll.
     * @param otherTrans Transformation, um <code>other</code> in das lokale
     *        Koordinatensystem von <code>this</code> zu transformieren.
     * @return <code>true</code>, wenn sich die Bounds &uuml;bershneiden, sonst
     *         <code>false</code>.
     * @throws CapabilityNotSetException Wenn die Plane Teil eines lebenden Szenegraphen
     *            ist und folgende Capabilities <i>nicht</i> gesetzt sind:
     *            <ul>
     *               <li>{@linkplain Node#ALLOW_BOUNDS_READ}</li>
     *            </ul>
     * @throws SceneGraphCycleException Wenn sich die Plane in einem zyklisch
     *            geschlossenen Szenegraph befindet.
     */
    private boolean boundsIntersected(Plane other, Transform3D otherTrans)
        throws CapabilityNotSetException, SceneGraphCycleException
    {
        assert this.point != null && other.point != null;

        // Bounds kopieren und transformieren
        Bounds otherBounds, thisBounds;

        thisBounds = (Bounds) this.getBounds().clone();
        otherBounds = (Bounds) other.getBounds().clone();

        //thisBounds.transform(thisTrans);
        otherBounds.transform(otherTrans);

        // Bounds preufen
        return thisBounds.intersect(otherBounds);
    }

    /**
     * Ermittelt, ob die angegebene Plane auf der selben Ebene liegt wie diese
     * Plane. Es wird davon ausgegangen, dass die beiden Planes parallel zueinander
     * sind.
     * 
     * @param other Die zu &uuml;berpr&uuml;fende Plane.
     * @param otherTrans Transformation, um <code>other</code> in das lokale
     *        Koordinatensystem von <code>this</code> zu transformieren.
     * @return <code>true</code>, wenn die beiden Planes auf der selben Ebene
     *         liegen, sonst <code>false</code>.
     */
    private boolean isParallelEqualLevel(Plane other, Transform3D otherTrans)
    {
        assert this.point != null && other.point != null;
        assert isParallel(other, otherTrans);

        Vector3f otherPointVector = new Vector3f();
        Point3f otherPoint = new Point3f();

        otherTrans.transform(other.point[0], otherPoint);
        otherPointVector.set(otherPoint);

        // Liegt der Punkt von other in der Ebene von this?
        if (this.normal.dot(otherPointVector) == 0)
        {
            return true;
        }

        return false;
    }

    /**
     * Ermittelt den Durchschnitt von zwei sich &uuml;berlagernden Polygonen.
     * Die Polygone, welche diesen Durchschnitt beschreiben, werden als Listen
     * in der Liste <code>polygonList</code> gespeichert.
     * <pre>
     *    Point3f[] firstPolygon = (Point3f[]) polygonList.get(0);
     *    Point3f firstPoint = (Point3f) ((Point3f[]) polygonList.get(0))[0];
     * </pre>
     * Es wird der Algorithmus nach Weiler & Atherton verwendet.</p>
     * 
     * Die Listen <code>otherPointList</code> und <code>thisPointList</code>
     * muessen die Punkte der zu &uum;berpr&uuml;fenden Polygone enthalten, plus
     * die Schnittpunkte der beiden. In beiden Listen m&uuml;ssen die Punkte
     * in der selben Richtung angeordnet sein, also bei linksherum, oder beide
     * rechtsherum.<\br>
     * Die Liste <code>entryList</code> muss die Punkte enthalten, bei denen das
     * Polygon, welches durch die Punkte in <code>thisPointList</code> beschrieben
     * ist, in das andere Polygon, beschrieben durch die Punkte in
     * <code>otherPointList</code>, eindringt.</p>
     *
     * Alle drei Listen k&ouml;nnen durch die Funktion
     * {@linkplain #createIntersectionPointList(List, List, List)} erstellt werden.
     * 
     * @param otherPointList Liste der Punkte eines Polygons plus der Schnittpunkte
     *           mit dem durch <code>thisPointList</code> beschriebenen
     *           Polygons. Die Punkte muessen global sein und in der selben
     *           Richtung angeordnet sein wie die Punkte in
     *           <code>thisPointList</code>.
     * @param thisPointList Liste der Punkte eines Polygons plus der Schnittpunkte
     *           mit dem durch <code>otherPointList</code> beschriebenen
     *           Polygons. Die Punkte muessen global sein in der selben
     *           Richtung angeordnet sein wie die Punkte in
     *           <code>otherPointList</code>.
     * @param entryList Lister der Eintrittspunkte.
     * @param polygonList In dieser Liste werden die resultierenden
     *           Druchschnittspolygone gespeichert, in folgendem Format:
     *           <pre>
     *              Point3f[] pointArray = (Point3f[]) polygonList.get(0);
     *           </pre>
     *           Die Punkte in den Arrays sind genau <i>andersherum</i>
     *           angeordnet, als die Punkte in den Listen
     *           <code>thisPointList</code> und <code>otherPointList</code>.
     * @see #createIntersectionPointList(List, List, List)
     */
    private void createANDPolygons(
        List otherPointList,
        List thisPointList,
        List entryList,
        List polygonList)
    {
        ArrayList polygonPointVector;
        Point3f[] polygonPointArray;
        List currentList;
        List waitingList;
        int currentIndex, entryListIndex, waitingIndex;
        Point3f startPoint, currentPoint;

        // Solange durch die Schleife laufen, bis alle Eintrittspunkte verarbeitet wurden
        while (entryList.size() > 0)
        {
            // Liste fuer neues Polygon erstellen
            polygonPointVector = new ArrayList();
            currentList = thisPointList;
            waitingList = otherPointList;
            startPoint = (Point3f) entryList.get(0);
            currentPoint = startPoint;
            currentIndex = currentList.indexOf(startPoint);
            do
            {
                // Aktuellen Punkt in Punktliste des Polygons einfuegen
                polygonPointVector.add(currentPoint);

                // Akt. Punkt auch in wartender Liste enthalten?
                waitingIndex = waitingList.indexOf(currentPoint);
                if (waitingIndex > -1)
                {
                    entryListIndex = entryList.indexOf(currentPoint);
                    // Eintrittspunkt von this?
                    if (entryListIndex > -1)
                    {
                        // Eintrittspunkt aus Liste loeschen
                        entryList.remove(entryListIndex);
                        // Zur Zeit in other?
                        if (currentList == otherPointList)
                        {
                            // Nach this wechseln
                            currentList = thisPointList;
                            waitingList = otherPointList;
                            currentIndex = waitingIndex;
                        }
                    } else
                    {
                        // Zur Zeit in this?
                        if (currentList == thisPointList)
                        {
                            // Nach other wechseln
                            currentList = otherPointList;
                            waitingList = thisPointList;
                            currentIndex = waitingIndex;
                        }
                    }
                }

                // Auf naechsten Punkt bewegen
                ++currentIndex;
                // Wenn am Ende, dann wieder zum Anfang
                if (currentIndex == currentList.size())
                    currentIndex = 0;
                currentPoint = (Point3f) currentList.get(currentIndex);
            } while (currentPoint.equals(startPoint) == false);
            // PolygonPointVector in PolygonListe als Array einfuegen
            polygonPointArray = new Point3f[polygonPointVector.size()];
            polygonList.add(polygonPointVector.toArray(polygonPointArray));
        }
    }

    /**
     * &Uuml;berpr&uuml;ft die beide Punktlisten auf &Uuml;berschneidungen.
     * Es wird der Algorithmus nach Weiler & Atherton verwendet.</p>
     * 
     * Die angegebenen Listen werden dabei ver&auml;dert, da Schnittpunkte
     * hinzugef&uuml;gt werden. In der Liste <code>entryList</code> werden die
     * Punkte gespeichert, bei denen das Polygon, welches durch die Punkte in
     * <code>thisPointList</code> beschrieben ist, in das andere Polygon, beschrieben
     * durch die Punkte in <code>otherPointList</code>, eindringt.
     * Wenn diese Liste also nach dem Abarbeiten der Funktion leer ist, schneiden
     * sich die beiden Polygone nicht.</p>
     * Alle drei Listen k&ouml;nnen nach Bearbeitung durch diese Funktion der Methode
     * {@link #createANDPolygons} &uuml;bergeben werden, die, bei vorhandenen
     * &Uuml;berschneidungen, eine Liste von Polygonen zur&uuml;ckgibt, welche den
     * Durchschnitt der beiden Polygone darstellen.</p>
     * 
     * Die Listen haben folgendes Format:  
     * <pre>
     *    Point3f firstThisPoint = (Point3f) thisPointList.get(0);
     *    Point3f firstOtherPoint = (Point3f) otherPointList.get(0);
     *    Point3f firstEntryPoint = (Point3f) entryList.get(0);
     * </pre>
     * 
     * @param otherPointList Liste von Punkten eines Polygons. Die Punkte
     *                       m&uuml;ssen gegen den Uhrzeigersinn angeordnet sein,
     *                       bez&uuml;glich des Normalenvektors dieser Plane
     *                       (Rechte-Hand-Regel). Die Liste wird durch die
     *                       Funktion ge&auml;ndert.
     * @param thisPointList Liste von Punkten eines Polygons. Die Punkte
     *                      m&uuml;ssen gegen den Uhrzeigersinn angeordnet sein,
     *                      bez&uuml;glich des Normalenvektors dieser Plane
     *                      (Rechte-Hand-Regel). Die Liste wird durch die
     *                      Funktion ge&auml;ndert.
     * @param entryList In dieser Liste werden die Eintrittpunkte von
     *                  <code>thisPointList</code> in <code>otherPointList</code>
     *                  gespeichert.
     * @param stopping <code>true</code> wenn die Funktion beim ersten Auftreten
     *                 eines Schnitts abbrechen soll. Die Funktion gibt in
     *                 diesem Fall <code>true</code> zur&uuml;ck. Die Liste
     *                 <code>entryList</code> ist leer.<br \>
     *                 Ist dieser Parameter <code>false</code> so wird die
     *                 Funktion komplett abgearbeitet.  
     * 
     * @see #createANDPolygons
     */
    private boolean createIntersectionPointList(
        List otherPointList,
        List thisPointList,
        List entryList,
        boolean stopping)
    {
        Segment otherSegment = new Segment();
        Segment thisSegment = new Segment();
        Point3f otherStartPoint, otherEndPoint, thisStartPoint, thisEndPoint;
        float[] scaleFactor = new float[2];
        Point3f crossPoint = new Point3f();

        int otherListIndex = 0;
        int thisListIndex = 0;

        if (stopping == false)
            entryList.clear();

        // Schleife solange aussfuehren, bis Liste am Ende ist
        for (otherListIndex = 0;
            otherListIndex < otherPointList.size();
            ++otherListIndex)
        {
            // Startpunkt aus Liste entnehmen
            otherStartPoint =
                (Point3f) getElement(otherPointList, otherListIndex);
            // Endpunkt aus Liste entnehmen
            otherEndPoint =
                (Point3f) getElement(otherPointList, otherListIndex + 1);
            // Segment erstellen
            otherSegment.set(otherStartPoint, otherEndPoint);

            // Schleife solange aussfuehren, bis Liste am Ende ist
            for (thisListIndex = 0;
                thisListIndex < thisPointList.size();
                ++thisListIndex)
            {
                // Startpunkt aus Liste entnehmen
                thisStartPoint =
                    (Point3f) getElement(thisPointList, thisListIndex);
                // Endpunkt aus Liste entnehmen
                thisEndPoint =
                    (Point3f) getElement(thisPointList, thisListIndex + 1);
                // Segment erstellen
                thisSegment.set(thisStartPoint, thisEndPoint);

                // Schneiden sich die beiden Segmente?
                if (thisSegment
                    .getCrossInfo(
                        otherSegment,
                        epsilon,
                        scaleFactor,
                        crossPoint)
                    == Segment.SEGMENT_CROSSING)
                {
                    // Beruehren sich die beiden Segmente jeweils am Anfang?
                    if (scaleFactor[0] == 0 && scaleFactor[1] == 0)
                    {
                        // akt. Startpunkt ist Schnittpunkt
                        boolean result =
                            isPenetrate(
                                (Point3f) getElement(otherPointList,
                                    otherListIndex - 1),
                                otherEndPoint,
                                thisStartPoint,
                                thisEndPoint,
                                this.normal);

                        // Dringt this in other ein?
                        if (result)
                        {
                            if (stopping)
                                return true;
                            // Punkt in Eintrittsliste speichern
                            entryList.add(thisStartPoint);
                        }
                    } else // if (scaleFactor[0] == 0 && scaleFactor[1] == 0)
                        {
                        // Wird otherSegment mittig geschnitten?
                        if (scaleFactor[1] > 0 && scaleFactor[1] < 1)
                        {
                            // Endpunkt neu belegen
                            otherEndPoint = new Point3f(crossPoint);
                            // Schnittpunkt in Liste zwischen akt. Start- und Endpunkt einfuegen
                            otherPointList.add(
                                otherListIndex + 1,
                                otherEndPoint);
                            // Segment neu erstellen
                            otherSegment.set(otherStartPoint, otherEndPoint);
                        }

                        // Wird thisSegment mittig geschnitten?
                        if (scaleFactor[0] > 0 && scaleFactor[0] < 1)
                        {
                            // Endpunkt neu belegen
                            thisEndPoint = new Point3f(crossPoint);
                            // Schnittpunkt in Liste zwischen akt. Start- und Endpunkt einfuegen
                            thisPointList.add(thisListIndex + 1, thisEndPoint);
                        }
                    } // if (scaleFactor[0] == 0 && scaleFactor[1] == 0) ELSE
                } // if (thisSegment.getCrossInfo(...))
            } // while (thisListIndex + 1 < thisResultList.size())
        } // while (otherListIndex + 1 < otherResultList.size())

        if (stopping == false && entryList.size() > 0)
            return true;
        else
            return false;
    }

    /**
     * &Uuml;berpr&uuml;ft, ob der Vektor 1 (nextThisPoint - crossPoint) auf der
     * jeweils linken Seite des Vektors 2 (nextOtherPoint - crossPoint) und des
     * Vektors 3 (crossPoint - lastOtherPoint) liegt.
     * Diese Funktion &uuml;berpr&uuml;ft also, ob der Vektor 1 in das Polygon
     * eindringt, welches unter anderem durch die Vektoren 2 und 3 definiert ist. 
     * 
     * @param lastOtherPoint Startpunkt von Vektor 3.
     * @param nextOtherPoint Endpunkt von Vektor 2.
     * @param crossPoint Endpunkt von Vektor 3 und Startpunkt von Vektor 1 und 2.
     *           Gemeinsamer Ber&uuml;hrungspunkt aller drei Vektoren.
     * @param nextThisPoint Endpunkt von Vektor 1.
     * @param normal Die Normale der Ebene, in der sich die Vektoren mit ihren
     *           Polygonen befinden. Die Punkte der Polygone m&uuml;ssen gegen den
     *           Uhrzeigersinn (Rechte-Hand-Regel) angeordnet sein.
     * @return <code>true</code>, wenn Vektor 1 in das Polygon von Vektor 2 und 3
     *         eindringt, sonst <code>false</code>.
     */
    private boolean isPenetrate(
        Point3f lastOtherPoint,
        Point3f nextOtherPoint,
        Point3f crossPoint,
        Point3f nextThisPoint,
        Vector3f normal)
    {
        boolean nextLeftSide, lastLeftSide, overPi;

        Vector3f lastOtherVector = new Vector3f();
        Vector3f nextOtherVector = new Vector3f();
        Vector3f thisVector = new Vector3f();
        Vector3f crossProdVector = new Vector3f();

        lastOtherVector.sub(crossPoint, lastOtherPoint);
        nextOtherVector.sub(nextOtherPoint, crossPoint);
        thisVector.sub(nextThisPoint, crossPoint);

        // Befindet sich der thisVector auf der linken Seite von nextOtherVector?
        crossProdVector.cross(nextOtherVector, thisVector);
        nextLeftSide = crossProdVector.dot(normal) > 0 ? true : false;

        // Befindet sich der thisVector auf der linken Seite von lastOtherVector?
        crossProdVector.cross(lastOtherVector, thisVector);
        lastLeftSide = crossProdVector.dot(normal) > 0 ? true : false;

        // Ist der Winkel zwischen lastOtherVector und nextOtherVector grosser als 180 Grad?
        crossProdVector.cross(lastOtherVector, nextOtherVector);
        overPi = crossProdVector.dot(normal) < 0 ? true : false;

        // Befindet sich thisVector auf der Innenseite?
        if (lastLeftSide
            && nextLeftSide
            || overPi
            && (lastLeftSide || nextLeftSide))
        {
            return true;
        }

        return false;
    }

    /**
     * Erm&ouml;glicht den zyklischen Zugriff auf das <code>point</code> Array, also
     * ohne das eine <code>IndexOutOfBoundsException</code> geworfen wird.
     * Wenn also z.B. der Index -1 &uuml;bergeben wird, wird das letzte Element
     * zur&uuml;ck gegeben.
     * 
     * @param index Der Index, welcher das Element adressiert.
     * @return Das engesprochene Element.
     * @throws NoSuchElementException Wenn das Array <code>this.point</code> leer ist.
     */
    private Point3f getPoint(int index) throws NoSuchElementException
    {
        assert this.point != null;

        if (this.point.length == 0)
            throw new NoSuchElementException("Das Array enthlt keine Punkte.");

        while (index < 0 || index >= this.point.length)
        {
            if (index >= this.point.length)
            {
                index = index - this.point.length;
            } else
                if (index < 0)
                {
                    index = index + this.point.length;
                }
        }

        return this.point[index];
    }

    /**
     * Erm&ouml;glicht den zyklischen Zugriff auf die Liste <code>list</code>, also
     * ohne das eine <code>IndexOutOfBoundsException</code> geworfen wird.
     * Wenn also z.B. der Index -1 &uuml;bergeben wird, wird das letzte Element
     * zur&uuml;ck gegeben.
     * 
     * @param list Die Liste, auf welche zugegriffen werden soll.
     * @param index Der Index, welcher das Element adressiert.
     * @return Das engesprochene Element.
     * @throws NoSuchElementException Wenn die Liste <code>list</code> leer ist.
     */
    private Object getElement(List list, int index)
        throws NoSuchElementException
    {
        if (list.size() == 0)
            throw new NoSuchElementException("Liste ist leer.");

        while (index < 0 || index >= list.size())
        {
            if (index >= list.size())
            {
                index = index - list.size();
            } else
                if (index < 0)
                {
                    index = index + list.size();
                }
        }

        return list.get(index);
    }

    /**
     * Klasse zum Vergleichen von Punkten, die alle auf einer Geraden liegen.
     */
    private class PointComperator implements Comparator
    {
        /** Startwert der Geraden, auf welcher die Punkte liegen. */
        private float startValue;

        /** Richtungswert der Geraden, auf welcher die Punkte liegen. */
        private float dirValue;

        /**
         * Index der Zeile (Zeilenvektor) welche zur Berechnung genutzt werden soll.
         * Da davon ausgegangen wird, dass alle Punkte auf einer Geraden liegen,
         * muss nur jeweils ein Wert (x, y oder z) f&uuml;r Richtungsvektor und
         * Startpunkt errechnet werden.
         */
        private int index;

        /**
         * Legt einen neuen uninitialisierten PointComperator an.
         * Er muss noch mit {@linkplain #set(Point3f, Point3f)} initialisiert werden,
         * bevor er genutzt werden kann.
         */
        public PointComperator()
        {
            startValue = dirValue = 0;
            index = -1;
        }

        /**
         * Setzt die Werte der Geraden, auf welcher die zu vergleichenden Punkte
         * liegen.
         * Der Richtungsvektor (<code>dirVec = endPoint - startPoint</code>) bestimmt
         * die Ordnung der Punkt. Je weiter ein Punkt in Richtung des Richtungsvektors
         * vom Startpunkt entfernt ist, desto gr&ouml;&szlig;er ist er.
         * 
         * @param startPoint Startpunkt der Geraden und des Richtungsvektors.
         * @param endPoint Endpunkt des Richtungsvektors.
         * @throws ArithmeticException Wird geworfen, wenn Start- und Endpunkt die
         *                             gleichen Werte haben.
         */
        public void set(Point3f startPoint, Point3f endPoint)
            throws ArithmeticException
        {
            Vector3f tmpVector = new Vector3f();
            Vector3f dirVec = new Vector3f();
            dirVec.sub(endPoint, startPoint);

            // Sind der Startpunkt und der Endpunkt identisch?
            if (dirVec.lengthSquared() == 0)
            {
                throw new ArithmeticException("Startpunkt und Endpunkt sind identisch.");
            }

            /* Da davon ausgegangen wird, dass alle Punkte auf einer Geraden liegen,
             * muss nur ein Wert errechnet werden.
             */
            tmpVector.absolute(dirVec);
            if (tmpVector.x > tmpVector.y && tmpVector.x > tmpVector.z)
            {
                index = 0;
                startValue = startPoint.x;
                dirValue = dirVec.x;
            } else
                if (tmpVector.y > tmpVector.z)
                {
                    index = 1;
                    startValue = startPoint.y;
                    dirValue = dirVec.y;
                } else
                {
                    index = 2;
                    startValue = startPoint.z;
                    dirValue = dirVec.z;
                }
        }

        /**
         * @throws IllegalStateException Wird geworfen, wenn das Objekt noch nicht
         *         initialisiert wurde. Initialisiert wird das Objekt mit
         *         {@linkplain #set(Point3f, Point3f)}.
         * @see #set(Point3f, Point3f)
         * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
         */
        public int compare(Object o1, Object o2) throws IllegalStateException
        {
            float p1, p2;

            switch (index)
            {
                case 0 :
                    p1 = ((Point3f) o1).x;
                    p2 = ((Point3f) o2).x;
                    break;
                case 1 :
                    p1 = ((Point3f) o1).y;
                    p2 = ((Point3f) o2).y;
                    break;
                case 2 :
                    p1 = ((Point3f) o1).z;
                    p2 = ((Point3f) o2).z;
                    break;
                default :
                    throw new IllegalStateException("Comperator wurde nicht initialisiert.");
            }

            if (p1 < p2)
                return -1;
            if (p1 > p2)
                return +1;

            return 0;
        }
    }

    /**
     * Wird geworfen, wenn keine Geometrie erstellt werden konnte.
     * 
     * @author Martin Hedler
     * @version 1.0 (2005/01/31)
     */
    private class GeometryNotCreated extends Exception
    {
        public GeometryNotCreated()
        {
            super();
        }

        public GeometryNotCreated(String message)
        {
            super(message);
        }

        public GeometryNotCreated(Throwable cause)
        {
            super(cause);
        }

        public GeometryNotCreated(String message, Throwable cause)
        {
            super(message, cause);
        }
    }

    /* (non-Javadoc)
     * @see javax.media.j3d.Node#cloneNode(boolean)
     */
    public Node cloneNode(boolean forceDuplicate)
    {
        Plane newPlane = new Plane(this.point);
        newPlane.duplicateNode(this, forceDuplicate);
        return newPlane;
    }

    /* (non-Javadoc)
     * @see javax.media.j3d.Node#duplicateNode(javax.media.j3d.Node, boolean)
     */
    public void duplicateNode(Node node, boolean forceDuplicate)
    {
        super.duplicateNode(node, forceDuplicate);
        Plane org = ((Plane) node);

        this.listenerList = (ArrayList) org.listenerList.clone();

        //Tiefe Kopie der Liste
        this.holeList = new ArrayList();
        for (Iterator it = org.holeList.iterator(); it.hasNext();)
        {
            Point3f[] orgHolePoints = (Point3f[]) it.next();
            Point3f[] newHolePoints = new Point3f[orgHolePoints.length];
            System.arraycopy(
                orgHolePoints,
                0,
                newHolePoints,
                0,
                orgHolePoints.length);
        }
    }
}
