/*
 * Created on 25.01.2005
 *
 */
package construction;

import javax.media.j3d.Group;
import javax.media.j3d.Node;
import javax.media.j3d.Texture;
import javax.media.j3d.TextureAttributes;
import javax.media.j3d.Transform3D;
import javax.media.j3d.TransformGroup;
import javax.vecmath.Point3f;
import javax.vecmath.Vector3f;

/**
 * Erstellt eine runde Treppe.
 * 
 * @author Martin Hedler
 * @version 0.9 (2005/01/31)
 *
 */
public class RoundStair extends Group
{
    /**
     * Erstellt ein Objekt dieser Klasse, jedoch ohne Geometrie. Diese muss mit
     * der Methode {@linkplain #set(float, float, float, float, int, float)}
     * erstellt werden.
     */
    public RoundStair()
    {
        super();
    }

    /**
     * Erstellt die Geometrie der Treppe
     * 
     * @param innerRadius Der innere Radius der Treppe.
     * @param outerRadius Der &auml;u&szlig;ere Radius der Treppe.
     * @param angle Der Winkel, welchen die Treppe aufspannen soll. Muss
     *        zwischen -2 * Pi und +2 * Pi liegen.
     * @param stepHeight H&ouml;he einer einzelnen Stufe.
     * @param stepCount Anzahl der Stufen.
     * @param startHeight Starth&ouml;he der untersten Stufe.
     * @throws IllegalArgumentException Wenn der Winkel au&szlig;erhalb des
     *         g&uuml;ltigen Bereichs liegt, oder wenn der &auml;u&szlig;ere
     *         Radius kleiner als der innere ist.
     */
    public void set(
        float innerRadius,
        float outerRadius,
        float angle,
        float stepHeight,
        int stepCount,
        float startHeight)
        throws IllegalArgumentException
    {
        // Alte Bloecke zuvor loeschen
        this.removeAllChildren();

        // Innerer Radius groesser als aeusserer Radius?
        if (innerRadius > outerRadius)
        {
            throw new IllegalArgumentException("Innerer Radius muss kleiner sein, als der uere Radius.");
        }

        if (Math.abs(angle) > 2 * Math.PI)
        {
            throw new IllegalArgumentException("Winkel muss zwischen -2 * Pi und +2 * Pi liegen.");
        }

        float deltaAngle = Math.abs(angle) / (float) stepCount;
        float height = ((float) stepCount) * stepHeight + startHeight;
        float currentHeight = startHeight;
        Block stepBlock;
        this.stepCount = stepCount;
        Block lastBlock = null;
        Transform3D trans = new Transform3D();
        Transform3D lastTrans = new Transform3D();
        TransformGroup trGrp;
        Vector3f transVec = new Vector3f();
        Point3f ceilingPoints[] = new Point3f[4];
        Point3f floorPoints[] = new Point3f[4];

        int direction = (angle >= 0) ? 1 : -1;

        // Array mit Objekten belegen
        for (int i = 0; i < 4; ++i)
        {
            ceilingPoints[i] = new Point3f();
            floorPoints[i] = new Point3f();
        }

        float xi =
            (float) Math.cos(deltaAngle / 2) * innerRadius
                - (innerRadius + outerRadius) / 2;
        float yi = (float) Math.sin(deltaAngle / 2) * innerRadius;
        float xo =
            (float) Math.cos(deltaAngle / 2) * outerRadius
                - (innerRadius + outerRadius) / 2;
        float yo = (float) Math.sin(deltaAngle / 2) * outerRadius;

        for (int i = 0; i < stepCount; ++i)
        {
            currentHeight += stepHeight;

            floorPoints[0].set(xi, -currentHeight / 2, +yi);
            floorPoints[1].set(xi, -currentHeight / 2, -yi);
            floorPoints[2].set(xo, -currentHeight / 2, -yo);
            floorPoints[3].set(xo, -currentHeight / 2, +yo);
            ceilingPoints[0].set(xi, currentHeight / 2, +yi);
            ceilingPoints[1].set(xi, currentHeight / 2, -yi);
            ceilingPoints[2].set(xo, currentHeight / 2, -yo);
            ceilingPoints[3].set(xo, currentHeight / 2, +yo);

            stepBlock = new Block(floorPoints, ceilingPoints);

            transVec.set(
                (float) Math.cos(
                    ((float) i + 0.5f) * deltaAngle * (float) direction)
                    * (innerRadius + outerRadius)
                    / 2,
                (currentHeight - height) / 2,
                - (float) Math.sin(
                    ((float) i + 0.5f) * deltaAngle * (float) direction)
                    * (innerRadius + outerRadius)
                    / 2);
            trans.rotY(((float) i + 0.5f) * deltaAngle * (float) direction);

            trans.setTranslation(transVec);
            trGrp = new TransformGroup(trans);

            trGrp.addChild(stepBlock);
            this.addChild(trGrp);

            if (lastBlock != null)
            {
                lastTrans.mulInverse(lastTrans, trans);
                lastBlock.connect(stepBlock, lastTrans);
            }
            lastBlock = stepBlock;
            lastTrans.set(trans);

            transVec.scale(0.2f);
            transVec.x = -transVec.x;
            transVec.y = -transVec.y;
            trans.setIdentity();
            trans.setTranslation(transVec);
            this.setTextureTransform(stepBlock, trans);
        }
    }

    private int stepCount;

    /**
     * Setzt die Texturen, welche die Stufen bei ihrer Erstellung erhalten
     * sollen.<br/>
     * Um f&uuml;r einzelne Stufen die Texturen zu setzen, muss auf das
     * Block Objekt direkt zugegriffen werden. Siehe hierzu die Methode
     * {@linkplain Block#setPlaneTexture(Texture[])}.
     * 
     * @param texture Array von Texturen. Die Anzahl der Texturen im Array
     *           bestimmt f&uuml;r was die Texturen verwendet werden:
     *           <ul>
     *              <li>Befindet sich nur eine Textur im Array, so wird die
     *                  Textur f&uuml;r alle W&auml;nde, den Boden und die
     *                  Decke verwendet.</li>
     *              <li>Befinden sich zwei Texturen im Array, so wird die erste
     *                  Textur f&uuml;r den Boden und die Decke verwendet und
     *                  die zweite Textur f&uuml;r alle W&auml;nde.</li>
     *              <li>Bei drei Texturen, wird die erste Textur f&uuml;r den
     *                  Boden, die zweite Textur f&uuml;r die Decke und die
     *                  dritte Textur f&uuml;r alle W&auml;nde verwendet.</li>
     *           </ul>
     *           Bei mehr als drei Texturen, findet der dritte Fall Anwendung
     *           und alle weiteren Texturen werden ignoriert.
     * @see Block#setPlaneTexture(Texture[])
     */
    public void setTexture(Texture[] texture)
    {
        for (int i = 0; i < stepCount; ++i)
        {
            getBlock(i).setPlaneTexture(texture);
        }
    }

    private void setTextureTransform(Block block, Transform3D transform)
    {
        for (int i = 2; i < block.numPlanes(); ++i)
        {
            TextureAttributes tA = new TextureAttributes();
            tA.setTextureTransform(transform);
            block.getPlane(i).getAppearance().setTextureAttributes(tA);
        }
    }

    /**
     * Gibt das Block Objekte an der Position <code>index</code> zur&uuml;ck.
     * 
     * @return Der Iterator, welcher die Block Objekte enth&auml;lt.
     */
    public Block getBlock(int index)
    {
        return (Block) ((TransformGroup) getChild(index)).getChild(0);
    }
    
    /**
     * Gibt die Anzahl der Stufen der Treppe zur&uml;ck.
     * 
     * @return Die Anzahl der Block Objekte.
     */
    public int numBlocks()
    {
        return numChildren();
    }
    
    /* (non-Javadoc)
     * @see javax.media.j3d.Node#cloneNode(boolean)
     */
    public Node cloneNode(boolean arg0)
    {
        RoundStair newStair = new RoundStair();
        newStair.duplicateNode(this, arg0);
        return newStair;
    }

    /* (non-Javadoc)
     * @see javax.media.j3d.Node#duplicateNode(javax.media.j3d.Node, boolean)
     */
    public void duplicateNode(Node arg0, boolean arg1)
    {
        super.duplicateNode(arg0, arg1);
        
        this.stepCount = ((RoundStair) arg0).stepCount;
    }
}
