/*
 * Created on 08.08.2004
 *
 */
package construction;

import javax.vecmath.Point3f;
import javax.vecmath.Vector3f;
import javax.vecmath.Vector4f;

/**
 * Repr&auml;sentiert eine Strecke im Raum. Eine Strecke ist ein endlicher Teilabschnitt
 * einer Geraden und definiert durch einen Start- und Endpunkt.<br />
 * Diese Klasse stellt Methoden bereit, um die Verh&auml;ltnisse zwischen
 * Strecken untereinander, Strecken und Ebenen und Strecken und Punkten zu
 * ermitteln.
 * 
 * @author Martin Hedler
 * @version 1.0.1 (2005/01/31)
 *
 */
public class Segment
{
    /** Startpunkt der Strecke */
    private Point3f startPoint;

    /** Endpunkt der Strecke */
    private Point3f endPoint;

    /** Richtungsvektor der Geraden (dirVec = endPoint - startPoint) */
    private Vector3f dirVec;

    /** Wenn kein Schnittpunkt existiert */
    public static final int NO_CROSSING = 0;

    /**
     * Wenn der Schnittpunkt innerhalb der &uuml;berpr&uuml;ften Segmente
     * liegt
     */
    public static final int SEGMENT_CROSSING = 1;

    /**
     * Wenn der Schnittpunkt bei mindestens einer Gerade nicht innerhalb des
     * Segments liegt.
     */
    public static final int LINE_CROSSING = 2;

    /**
     * Erstellt ein Segment. Die angegebenen Parameter werden kopiert.
     * @param startPoint Startpunkt des Segments.
     * @param endPoint Endpunkt des Segments.
     */
    public Segment(Point3f startPoint, Point3f endPoint)
    {
        this.startPoint = new Point3f();
        this.endPoint = new Point3f();
        this.dirVec = new Vector3f();
        this.set(startPoint, endPoint);
    }

    /**
     * Erstellt ein Segment mit dem Startpunt (0, 0, 0) und dem Endpunkt (1, 1, 1).
     */
    public Segment()
    {
        this.startPoint = new Point3f(0, 0, 0);
        this.endPoint = new Point3f(1, 1, 1);
        this.dirVec = new Vector3f();
        this.dirVec.sub(this.endPoint, this.startPoint);
    }

    /**
     * Erstellt ein Segment. Die angegebenen Parameter werden kopiert.
     * @param startPoint Startpunkt des Segments.
     * @param dirVec Vektor, der ausgehend vom Startpunkt auf den Endpunkt zeigt.
     *               Richtungsvektor der Geraden, auf der die Strecke liegt.
     */
    public Segment(Point3f startPoint, Vector3f dirVec)
    {
        this.startPoint = new Point3f();
        this.endPoint = new Point3f();
        this.dirVec = new Vector3f();
        this.set(startPoint, dirVec);
    }

    /**
     * Setzt die Werte des Segments neu. Die angegebenen Parameter werden kopiert.
     * @param startPoint Startpunkt des Segments.
     * @param endPoint Endpunkt des Segments.
     */
    public void set(Point3f startPoint, Point3f endPoint)
    {
        this.startPoint.set(startPoint);
        this.endPoint.set(endPoint);
        this.dirVec.sub(endPoint, startPoint);
    }

    /**
     * Setzt die Werte dieses Segments auf die Werte von <code>other</code>.
     * @param other Segments, dessen Werte &uuml;bernommen werden sollen.
     */
    public void set(Segment other)
    {
        this.startPoint.set(other.startPoint);
        this.endPoint.set(other.endPoint);
        this.dirVec.set(other.dirVec);
    }

    /**
     * Setzt die Werte des Segments neu. Die angegebenen Parameter werden kopiert.
     * @param startPoint Startpunkt des Segments.
     * @param dirVec Vektor, der ausgehend vom Startpunkt auf den Endpunkt zeigt.
     *               Richtungsvektor der Geraden, auf der die Strecke liegt.
     */
    public void set(Point3f startPoint, Vector3f dirVec)
    {
        this.startPoint.set(startPoint);
        this.dirVec.set(dirVec);
        this.endPoint.add(startPoint, dirVec);
    }

    /**
     * Liefert den Startpunkt des Segments.
     * @param startPoint Punkt, in dem die Werte des Startpunktes gespeichert werden.
     */
    public void getStartPoint(Point3f startPoint)
    {
        startPoint.set(this.startPoint);
    }

    /**
     * Liefert den Endpunkt des Segments.
     * @param endPoint Punkt, in dem die Werte des Endpunktes gespeichert werden.
     */
    public void getEndPoint(Point3f endPoint)
    {
        endPoint.set(this.endPoint);
    }

    /**
     * Liefert den Richtungsvektor des Segments. dirVec = endPoint - startPoint.
     * @param dirVec Vector, in dem die Werte des Richtungsvektors gespeichert werden.
     */
    public void getDirectionVector(Vector3f dirVec)
    {
        dirVec.set(this.dirVec);
    }

    /**
     * Liefert den Startpunkt und den Endpunkt des Segments.
     * @param startPoint Punkt, in dem die Werte des Startpunktes gespeichert werden.
     * @param endPoint Punkt, in dem die Werte des Endpunktes gespeichert werden.
     */
    public void get(Point3f startPoint, Point3f endPoint)
    {
        startPoint.set(this.startPoint);
        endPoint.set(this.endPoint);
    }

    /**
     * Gibt die L&auml;nge des Segments zur&uuml;ck.
     * @return Die L&auml;nge des Segments.
     */
    public float length()
    {
        return dirVec.length();
    }

    /**
     * Skaliert das Segments. Der Startpunkt bleibt erhalten.
     * @param s Skalierungsfaktor.
     */
    public void scale(float s)
    {
        this.dirVec.scale(s);
        this.endPoint.add(startPoint, dirVec);
    }

    /**
     * &Uuml;berpr&uuml;ft, ob der Punkt <code>point</code> in der Ebene
     * <code>plane</code> liegt.
     * 
     * @param plane Die Ebene.
     *              Die ersten drei Werte (x, y, z) sind die Werte des Normalenvektors
     *              der Ebene. Der vierte Wert (w) ist die Entfernung der Ebene vom
     *              Koordinatenursprung. Es gilt die Normalenform:
     *              <pre>
     *              E: x * px + y * py + z * py + w = 0;
     *              </pre>
     *              wobei <code>px, py, pz</code> die Punktkoordinaten sind.
     * @param point Der Punkt, welcher &uuml;berpr&uuml;ft werden soll.
     * @return <code>true</code>, wenn der Punkt in der Ebene liegt sonst
     *         <code>false</code>. 
     */
    private boolean isPointInPlane(Vector4f plane, Point3f point)
    {
        Vector4f pointVector4f = new Vector4f(point.x, point.y, point.z, 1);

        return (plane.dot(pointVector4f) == 0);
    }

    /**
     * Liefert Informationen &uuml;ber den Schnitt eines Segments mit einer
     * Ebene. Entweder schneidet die Gerade auf welcher das Segment liegt, die
     * Ebene nicht, oder die Gerade schneidet zwar die Ebene, jedoch nicht
     * innerhalb des Segments, oder das Segment schneidet die Ebene.
     * 
     * @param plane Die Ebene, mit der dieses Segment geschnitten werden soll.
     *              Die ersten drei Werte (x, y, z) sind die Werte des Normalenvektors
     *              der Ebene. Der vierte Wert (w) ist die Entfernung der Ebene vom
     *              Koordinatenursprung. Es gilt die Normalenform:
     *              <pre>
     *              E: x * px + y * py + z * py + w = 0;
     *              </pre>
     *              wobei <code>px, py, pz</code> die Punktkoordinaten sind.
     * @param scaleFactor Das Segmente liegt auf einer Geraden, welche wie folgt
     *        beschrieben ist:
     *        <pre>
     *        p:=this.startPoint + r * this.dirVec; 
     *        </pre>
     *        wobei <code>dirVec</code> berechnet wird:
     *        <pre>
     *        dirVec = endPoint - startPoint;
     *        </pre>
     *        In <code>scaleFactor[0]</code> wird nun der Wert von <code>r</code>
     *        gespeichert.
     * @param point Wenn sich die Ebene und die Gerade schneiden, dann wird hier der
     *        Schnittpunkt gespeichert.
     * @return 
     * <ul>
     *    <li>NO_CROSSING, wenn sich die Gerade und die Ebene nicht schneiden.
     *        </li>
     *    <li>LINE_CROSSING, wenn die Gerade, auf der das Segment liegt, die
     *        Ebene schneidet, nicht jedoch innerhalb des Segments. Der
     *        <code>scaleFactor</code> enth&auml;lt den Skalierungsfaktor.</li>
     *    <li>SEGMENT_CROSSING, wenn sich das Segmente und die Ebene schneiden.
     *        Der <code>scaleFactor</code> enth&auml;lt den Skalierungsfaktor.
     *        </li>
     * </ul>
     */
    public int getCrossInfo(Vector4f plane, float[] scaleFactor, Point3f point)
    {
        // Ist Segment ein Punkt?
        if (this.dirVec.lengthSquared() == 0)
        {
            // Liegt der Punkt in der Ebene?
            if (isPointInPlane(plane, this.startPoint))
            {
                point.set(this.startPoint);
                scaleFactor[0] = 0;
                return Segment.SEGMENT_CROSSING;
            } else
            {
                return Segment.NO_CROSSING;
            }
        }

        float nenner, r;
        Vector3f normal = new Vector3f();

        normal.set(plane.x, plane.y, plane.z);
        nenner = normal.dot(dirVec);

        // Segment paralell zu Ebene?
        if (nenner == 0)
        {
            return Segment.NO_CROSSING;
        }

        // scaleFactor berechnen
        Vector3f startPointVector = new Vector3f(this.startPoint);
        r = - (normal.dot(startPointVector) + plane.w) / nenner;
        scaleFactor[0] = r;

        // Schnittpunkt berechnen
        point.set(this.dirVec);
        point.scale(r);
        point.add(this.startPoint);

        if (r >= 0 && r <= 1)
        {
            // Segment schneidet Ebene
            return Segment.SEGMENT_CROSSING;
        }

        // Nur die Gerade schneidet die Ebene
        return Segment.LINE_CROSSING;
    }

    /**
     * Liefert Informationen &uuml;ber den Schnitt zweier Segmente. Dei beiden
     * Segmente k&ouml;nnen sich nicht schneiden, die Geraden, auf denen sich
     * die Segmente befinden k&ouml;nnen sich schneiden, oder die Segment
     * k&ouml;nnen sich schneiden.
     * 
     * @param other Das Segment, mit dem dieses Segment geschnitten werden soll.
     * @param epsilon Bestimmt die Genauigkeit, mit der entschieden wird, ob
     *        sich die beiden Segmente schneiden oder nicht.
     *        Nachdem <code>r</code> und <code>s</code> berechnet wurden, werden
     *        die daraus folgenden Schnittpunkte je Gerade berechnet und
     *        miteinander verglichen. Der Parameter wird f&uuml;r die Methode
     *        {@linkplain javax.vecmath.Tuple3f#epsilonEquals(javax.vecmath.Tuple3f, float)}
     *        verwendet.
     * @param scaleFactor Die Segmente liegen jeweils auf Geraden, welche wie folgt
     *        beschrieben werden:
     *        <pre>
     *        p:=this.startPoint + r * this.dirVec;
     *        p:=other.startPoint + s * other.dirVec; 
     *        </pre>
     *        wobei <code>dirVec</code> jeweils berechnet wird:
     *        <pre>
     *        dirVec = endPoint - startPoint;
     *        </pre>
     *        In <code>scaleFactor[0]</code> wird nun der Wert von <code>r</code>
     *        gespeichert, in <code>scaleFactor[1]</code> der Wert von <code>s</code>.
     * @param point Wenn sich die beiden Geraden schneiden, dann wird hier der
     *        Schnittpunkt gespeichert.
     * @return  
     * <ul>
     *    <li><code>NO_CROSSING</code>, wenn die Geraden sich nicht schneiden.
     *    <li><code>LINE_CROSSING</code>, wenn sich die Geraden, auf denen die
     *        Segmente liegen, schneiden, nicht jedoch innerhalb beider
     *        Segmente. Die <code>scaleFactor</code> enthalten den jeweiligen
     *        Skalierungsfaktor.</li>
     *    <li><code>SEGMENT_CROSSING</code>, wenn sich beide Segmente schneiden.
     *        Die <code>scaleFactor</code> enthalten den jeweiligen
     *        Skalierungsfaktor.</li>
     * </ul>
     * @see javax.vecmath.Tuple3f#epsilonEquals(javax.vecmath.Tuple3f, float)
     */
    public int getCrossInfo(
        Segment other,
        float epsilon,
        float[] scaleFactor,
        Point3f point)
    {
        // Ist this ein Punkt?
        if (this.dirVec.lengthSquared() == 0)
        {
            // Ist other auch ein Punkt?
            if (other.dirVec.lengthSquared() == 0)
            {
                // Sind die Punkte gleich?
                if (this.startPoint.equals(other.startPoint))
                {
                    point.set(this.startPoint);
                    scaleFactor[0] = scaleFactor[1] = 0;
                    return Segment.SEGMENT_CROSSING;
                } else
                    return Segment.NO_CROSSING;
            } else
            {
                // other ist ein Segment
                scaleFactor[1] = other.getScaleFactor(this.startPoint);

                // Liegt der Punkt this auf other?
                if (Float.isNaN(scaleFactor[1]) == false)
                {
                    scaleFactor[0] = 0;
                    point.set(this.startPoint);
                    if (scaleFactor[1] >= 0 && scaleFactor[1] <= 1)
                        return Segment.SEGMENT_CROSSING;
                    else
                        return Segment.LINE_CROSSING;
                } else
                {
                    return Segment.NO_CROSSING;
                }
            }
        } else // if (this.dirVec.lengthSquared() == 0)
            if (other.dirVec.lengthSquared() == 0)
            {
                // other ist ein Punkt
                scaleFactor[0] = this.getScaleFactor(other.startPoint);

                // Liegt der Punkt other auf this?
                if (Float.isNaN(scaleFactor[0]) == false)
                {
                    scaleFactor[1] = 0;
                    point.set(other.startPoint);
                    if (scaleFactor[0] >= 0 && scaleFactor[0] <= 1)
                        return Segment.SEGMENT_CROSSING;
                    else
                        return Segment.LINE_CROSSING;
                } else
                {
                    return Segment.NO_CROSSING;
                }
            }

        /*
         * Geradengleichungen:
         * p = this.p1 + r * this.dirVec
         * p = other.p1 + s * other.dirVec
         */

        float zaehler, nenner, nennerX, nennerY, nennerZ, s, r;

        nennerX =
            other.dirVec.z * this.dirVec.y - other.dirVec.y * this.dirVec.z;
        nennerY =
            other.dirVec.x * this.dirVec.z - other.dirVec.z * this.dirVec.x;
        nennerZ =
            other.dirVec.y * this.dirVec.x - other.dirVec.x * this.dirVec.y;

        if (Math.abs(nennerX) > Math.abs(nennerY)
            && Math.abs(nennerX) > Math.abs(nennerZ))
        {
            nenner = nennerX;
            zaehler =
                other.dirVec.z * (other.startPoint.y - this.startPoint.y)
                    - other.dirVec.y * (other.startPoint.z - this.startPoint.z);
        } else
            if (Math.abs(nennerY) > Math.abs(nennerZ))
            {
                nenner = nennerY;
                zaehler =
                    other.dirVec.x * (other.startPoint.z - this.startPoint.z)
                        - other.dirVec.z
                            * (other.startPoint.x - this.startPoint.x);
            } else
            {
                nenner = nennerZ;
                zaehler =
                    other.dirVec.y * (other.startPoint.x - this.startPoint.x)
                        - other.dirVec.x
                            * (other.startPoint.y - this.startPoint.y);
            }

        if (nenner == 0)
        {
            return Segment.NO_CROSSING;
        }

        r = zaehler / nenner;

        if (Math.abs(other.dirVec.x) > Math.abs(other.dirVec.y)
            && Math.abs(other.dirVec.x) > Math.abs(other.dirVec.z))
        {
            s =
                (this.startPoint.x - other.startPoint.x + r * this.dirVec.x)
                    / other.dirVec.x;
        } else
            if (Math.abs(other.dirVec.y) > Math.abs(other.dirVec.z))
            {
                s =
                    (this.startPoint.y
                        - other.startPoint.y
                        + r * this.dirVec.y)
                        / other.dirVec.y;
            } else
            {
                s =
                    (this.startPoint.z
                        - other.startPoint.z
                        + r * this.dirVec.z)
                        / other.dirVec.z;
            }

        // Gleicher Schnittpunkt?
        Point3f otherPoint = new Point3f(other.dirVec);
        otherPoint.scaleAdd(s, other.startPoint);

        point.scaleAdd(r, this.dirVec, this.startPoint);

        if (point.epsilonEquals(otherPoint, epsilon))
        {
            scaleFactor[0] = r;
            scaleFactor[1] = s;

            if (s >= 0 && s <= 1 && r >= 0 && r <= 1)
            {
                // Schnitt innerhalb beider Segmente
                return Segment.SEGMENT_CROSSING;
            }

            // Schnitt nur der Geraden
            return Segment.LINE_CROSSING;
        } else
        {
            // Geraden scheiden sich nicht
            return Segment.NO_CROSSING;
        }
    }

    /**
     * Gibt den Skalierunsfaktor zur&uuml;ck, der mit dem Richtungsvektor des
     * Segments multipliziert, und mit dem Startpunkt addiert, den Punkt
     * <code>point</code> ergibt. Es gilt:
     * <pre>
     * scaleFactor = getScaleFactor(point);
     * point = startPoint + scaleFactor * dirVec;
     * </pre>
     * Dies gilt jedoch nur f&uuml;r den Fall, dass sich <code>point</code> auf
     * der Geraden des Segments befindet. Liegt <code>point</code> nicht auf dieser
     * Geraden, wird <code>Float.NaN</code> zur&uuml;ckgegeben.
     * 
     * @param point Der Punkt, der &uuml;berpr&uuml;ft werden soll.
     * @return Der errechnete Skalierungsfaktor. Wenn der Punkt <code>point</code>
     *         nicht auf der Geraden des Segments liegt, dann wird
     *         <code>Float.NaN</code> zur&uuml;ckgegeben.
     */
    public float getScaleFactor(Point3f point)
    {
        float scaleFactor;

        /* 
         * Schnittpunkt mit der Ebene berechnen, die durch den Punkt point verlaeuft
         * und senkrecht zur Geraden liegt.
         */
        Vector3f pointVector = new Vector3f(point);
        pointVector.sub(startPoint);
        Vector3f pointVector2 = new Vector3f(pointVector);
        pointVector.scale(dirVec.lengthSquared());
        pointVector2.scale(pointVector2.dot(dirVec), dirVec);
        pointVector.sub(pointVector2);

        if (pointVector.lengthSquared() == 0)
        {
            pointVector.sub(point, startPoint);
            scaleFactor = dirVec.dot(pointVector) / dirVec.lengthSquared();

            return scaleFactor;
        }

        return Float.NaN;
    }

    /**
     * Ermittelt, ob ein Punkt sich innerhalb des Segments befindet.
     * @param point Der Punkt, der &uuml;berpr&uuml;ft werden soll.
     * @return <code>true</code>, wenn sich der Punkt innerhalb des Segments befindet,
     *         sonst <code>false</code>.
     */
    public boolean isPointIn(Point3f point)
    {
        float s;

        s = getScaleFactor(point);
        if (Float.isNaN(s) == false && s >= 0 && s <= 1)
            return true;
        else
            return false;
    }

    /*
     * (non-Javadoc)
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
        return "(" + startPoint + ", " + endPoint + ")";
    }

    /* (non-Javadoc)
     * @see java.lang.Object#clone()
     */
    public Object clone() throws CloneNotSupportedException
    {
        Segment segment = (Segment) super.clone();
        segment.startPoint = (Point3f) this.startPoint.clone();
        segment.endPoint = (Point3f) this.endPoint.clone();
        segment.dirVec = (Vector3f) this.dirVec.clone();

        return segment;
    }

}