/*
 * Created on 22.01.2005
 *
 */
package interaction;

import java.util.Enumeration;

import javax.media.j3d.Alpha;
import javax.media.j3d.Behavior;
import javax.media.j3d.Interpolator;
import javax.media.j3d.Node;
import javax.media.j3d.WakeupOnElapsedFrames;

/**
 * Klasse zur einfacheren bzw. erweiterten Steuerung eines Interpolators. Die
 * Richtung des Interpolators (bzw. des <code>Alpha</code> Objekts) kann
 * w&auml;hrend der Animation ge&auml;ndert werden.
 * 
 * Ist die Animation beendet, wird der Interpolator deaktiviert.
 * 
 * @author Martin Hedler
 * @version 1.0 (2005/04/25)
 */
public class InterpolatorSwitch extends Behavior
{
    /** Alpha ist 0. */
    public static final int ZERO = 0;

    /** Alpha ist 1. */
    public static final int ONE = 1;

    /** Alpha l&auml;uft von 1 nach 0. */
    public static final int CHANGING_TO_ZERO = 2;

    /** Alpha l&auml;uft von 0 nach 1. */
    public static final int CHANGING_TO_ONE = 3;

    /** Bei jedem Frame ausf&uuml;hren. */
    private WakeupOnElapsedFrames crit;

    /** Der Interpolator, der &uuml;berwacht und gesteuert werden soll. */
    private Interpolator inter;

    /**
     * Erstellt eine Instanz dieser Klasse zur Steuerung des &uuml;bergebenen
     * Interpolators.
     * 
     * @param inter Der Interpolator, der &uuml;berwacht und gesteuert werden
     *        soll.
     */
    public InterpolatorSwitch(Interpolator inter)
    {
        this.inter = inter;
        this.crit = new WakeupOnElapsedFrames(2);
    }

    /**
     * &Auml;ndert die Richtung der Interpolation bzw. des Alpha Objekts.
     */
    public void changeState()
    {
        if (inter.getAlpha().finished())
        {
            if (inter.getAlpha().value() == 0)
            {
                inter.getAlpha().setMode(Alpha.INCREASING_ENABLE);
            } else
                if (inter.getAlpha().value() == 1)
                {
                    inter.getAlpha().setMode(Alpha.DECREASING_ENABLE);
                }
            inter.getAlpha().setLoopCount(1);
            inter.getAlpha().setStartTime(System.currentTimeMillis());
            inter.setEnable(true);
            this.setEnable(true);
        } else
        {
            if ((inter.getAlpha().getMode() & Alpha.INCREASING_ENABLE)
                == Alpha.INCREASING_ENABLE)
            {
                // Alpha laueft von 1 nach 0
                long timeDiff =
                    getStartTimeDifference(
                        inter.getAlpha().value(),
                        inter.getAlpha().getDecreasingAlphaDuration(),
                        inter.getAlpha().getDecreasingAlphaRampDuration());
                inter.getAlpha().setMode(Alpha.DECREASING_ENABLE);
                inter.getAlpha().setStartTime(
                    System.currentTimeMillis()
                        - inter.getAlpha().getIncreasingAlphaDuration()
                        + timeDiff);
            } else
            {
                // Alpha laueft von 0 nach 1
                long timeDiff =
                    getStartTimeDifference(
                        inter.getAlpha().value(),
                        inter.getAlpha().getIncreasingAlphaDuration(),
                        inter.getAlpha().getIncreasingAlphaRampDuration());
                inter.getAlpha().setMode(Alpha.INCREASING_ENABLE);
                inter.getAlpha().setStartTime(
                    System.currentTimeMillis() - timeDiff);
            }
        }
    }

    /**
     * Liefert den Status des Alpha Objekts zur&uuml;ck.
     * 
     * @return Folgende Werte sind m&ouml;glich:
     *         <ul>
     *            <li><code>ZERO</code>, wenn Alpha 0 ist</li>
     *            <li><code>ONE</code>, wenn Alpha 1 ist</li>
     *            <li><code>CHANGING_TO_ZERO</code>, wenn Alpha von 1 nach 0
     *                l&auml;uft</li>
     *            <li><code>CHANGING_TO_ONE</code>, wenn Alpha von 0 nach 1
     *                l&auml;uft</li>
     *         </ul>
     */
    public int getState()
    {
        if (inter.getAlpha().finished())
            if (inter.getAlpha().value() == 0)
                return ZERO;
            else
                return ONE;
        else
            if ((inter.getAlpha().getMode() & Alpha.INCREASING_ENABLE)
                == Alpha.INCREASING_ENABLE)
                return CHANGING_TO_ONE;
            else
                return CHANGING_TO_ZERO;
    }

    /**
     * Errechnet wieviel Zeit vergehen muss, um bei der gegebenen
     * Animationsdauer und Beschleunigung den Wert <code>value</code> zu
     * erhalten. 
     * 
     * @param value Der Wert zwischen 0 und 1, f&uuml;r den die Zeit ermittelt
     *        werden soll.
     * @param wholeTime Die Anstiegs- bzw. Abstiegszeit.
     * @param accelerationTime Die Beschleunigungszeit.
     * @return Die Zeit, die bei den gegebenen Gr&ouml;&szlig;en vergehen muss
     *         um <code>value</code> zu erhalten.
     * @throws IllegalArgumentException Wenn <code>wholeTime</code> kleiner
     *         gleich 0 oder <code>value</code> nicht zwischen 0 und 1 liegt.
     */
    private long getStartTimeDifference(
        float value,
        long wholeTime,
        long accelerationTime)
        throws IllegalArgumentException
    {
        if (wholeTime <= 0 || value < 0 || value > 1)
            throw new IllegalArgumentException();

        // Beschleunigungszeit anpassen, wenn fehlerhaft
        if (accelerationTime < 0)
            accelerationTime = 0;
        else
            if (accelerationTime > wholeTime / 2)
                accelerationTime = wholeTime / 2;

        float accValue =
            (float) accelerationTime
                / (float) (2 * (wholeTime - accelerationTime));
        long diffTime;

        if (value < accValue)
        {
            diffTime =
                (long) Math.sqrt(
                    2
                        * value
                        * accelerationTime
                        * (wholeTime - accelerationTime));
        } else
            if (value <= 1 - accValue)
            {
                diffTime =
                    (long) (value * (wholeTime - accelerationTime)
                        + accelerationTime / 2);
            } else
            {
                // Hier Fehler, wenn value > 1
                diffTime =
                    wholeTime
                        - (long) Math.sqrt(
                            2
                                * accelerationTime
                                * (accelerationTime - wholeTime)
                                * (value - 1));
            }

        return diffTime;
    }

    /* (non-Javadoc)
     * @see javax.media.j3d.Behavior#initialize()
     */
    public void initialize()
    {
        this.setEnable(false);
        this.wakeupOn(crit);
    }

    /* (non-Javadoc)
     * @see javax.media.j3d.Behavior#processStimulus(java.util.Enumeration)
     */
    public void processStimulus(Enumeration arg0)
    {
        if (inter.getAlpha().finished())
        {
            this.setEnable(false);
            inter.setEnable(false);
        }

        this.wakeupOn(crit);
    }

    /* (non-Javadoc)
     * @see javax.media.j3d.Node#cloneNode(boolean)
     */
    public Node cloneNode(boolean arg0)
    {
        InterpolatorSwitch newInterSw = new InterpolatorSwitch(inter);
        newInterSw.duplicateNode(this, arg0);
        return newInterSw;
    }
}
