/*
 * Created on 05.02.2005
 *
 */
package interaction;

import java.awt.AWTEvent;
import java.awt.event.MouseEvent;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;

import javax.media.j3d.Behavior;
import javax.media.j3d.BranchGroup;
import javax.media.j3d.Canvas3D;
import javax.media.j3d.CapabilityNotSetException;
import javax.media.j3d.Node;
import javax.media.j3d.NodeReferenceTable;
import javax.media.j3d.SceneGraphPath;
import javax.media.j3d.WakeupCondition;
import javax.media.j3d.WakeupOnAWTEvent;

import com.sun.j3d.utils.picking.PickCanvas;
import com.sun.j3d.utils.picking.PickResult;

/**
 * Behavior, welches auf Mausereignisse wartet dementsprechend durch Picking
 * ein Objekt aus der Szene ausw&auml;hlt. Implementiert dieses Objekt oder die
 * Objekte oberhalb des ausgw&auml;hlten Objekts die Schnittstelle
 * {@linkplain interaction.PickListener} werden die entsprechenden Methoden der
 * Schnittstelle aufgerufen.
 * <p/>
 * Dies geschieht auch, wenn in den <code>userData</code> der Objekte
 * <code>PickListener</code> Objekte enthalten sind.
 * <p/>
 * In jedem Fall werden alle <code>PickListener</code> Objekte aufgerufen, die
 * bei dem Objekt der Klasse <code>PickBehavior</code> registriert sind. 
 * 
 * @author Martin Hedler
 * @version 1.0 (2005/04/25)
 */
public class PickBehavior extends Behavior
{
    private WakeupCondition wakeupPressed, wakeupReleased;
    private int pickCanvasMode;
    private BranchGroup pickBG;
    private ArrayList listeners;
    private boolean throwCapExc;

    /**
     * Erstellt eine Instanz der Klasse, welche bei Mausereignissen in der
     * BranchGroup <code>bg</code>pickt.
     * 
     * @param bg Die BranchGroup, deren Objekte gepickt werden sollen.
     * @param pickCanvasMode Der Pickmodus. Siehe
     *        {@linkplain com.sun.j3d.utils.picking.PickTool#BOUNDS},
     *        {@linkplain com.sun.j3d.utils.picking.PickTool#GEOMETRY} oder
     *        {@linkplain com.sun.j3d.utils.picking.PickTool#GEOMETRY_INTERSECT_INFO}.
     * @param throwCapExc <code>true</code>, wenn
     *        <code>CapabilityNotSetException</code>s geworfen werden sollen
     *        oder <code>false</code>, wenn diese ignoriert werden sollen.
     */
    public PickBehavior(
        BranchGroup bg,
        int pickCanvasMode,
        boolean throwCapExc)
    {
        this.pickBG = bg;
        this.pickCanvasMode = pickCanvasMode;
        this.listeners = new ArrayList();
        this.throwCapExc = throwCapExc;

        this.wakeupPressed = new WakeupOnAWTEvent(MouseEvent.MOUSE_PRESSED);
        this.wakeupReleased = new WakeupOnAWTEvent(MouseEvent.MOUSE_RELEASED);
    }

    /* (non-Javadoc)
     * @see javax.media.j3d.Behavior#initialize()
     */
    public void initialize()
    {
        this.wakeupOn(wakeupPressed);
    }

    /* (non-Javadoc)
     * @see javax.media.j3d.Behavior#processStimulus(java.util.Enumeration)
     */
    public void processStimulus(Enumeration enum)
    {
        WakeupOnAWTEvent crit;
        WakeupCondition wakeupCon;
        AWTEvent events[];

        wakeupCon = this.wakeupPressed;

        for (; enum.hasMoreElements();)
        {
            crit = (WakeupOnAWTEvent) enum.nextElement();
            events = crit.getAWTEvent();
            for (int i = 0; i < events.length; ++i)
            {
                assert events[i] instanceof MouseEvent;

                int id = events[i].getID();
                PickListener pL;

                // Gepicktes Objekt ermitteln
                PickCanvas pickC =
                    new PickCanvas((Canvas3D) events[i].getSource(), pickBG);
                pickC.setMode(pickCanvasMode);
                pickC.setShapeLocation((MouseEvent) events[i]);
                PickResult currentResult;
                try
                {
                    currentResult = pickC.pickClosest();
                } catch (CapabilityNotSetException e)
                {
                    if (throwCapExc)
                        throw e;
                    currentResult = null;
                }

                if (currentResult == null)
                    continue;

                SceneGraphPath sgp = currentResult.getSceneGraphPath();
                if (sgp != null)
                {
                    int si = -1;
                    Node n = sgp.getObject();
                    while (n != null)
                    {
                        pL = null;
                        if (n instanceof PickListener)
                        {
                            pL = (PickListener) n;

                        } else
                            if (n.getUserData() != null
                                && n.getUserData() instanceof PickListener)
                            {
                                pL = (PickListener) n.getUserData();
                            }

                        if (pL != null)
                        {
                            if (id == MouseEvent.MOUSE_PRESSED)
                                pL.pressed(
                                    currentResult,
                                    (MouseEvent) events[i]);
                            else
                                if (id == MouseEvent.MOUSE_RELEASED)
                                    pL.released(
                                        currentResult,
                                        (MouseEvent) events[i]);
                        }

                        ++si;
                        if (si < sgp.nodeCount())
                            n = sgp.getNode(si);
                        else
                            n = null;
                    }
                }

                if (id == MouseEvent.MOUSE_PRESSED)
                {
                    firePressed(currentResult, (MouseEvent) events[i]);
                    wakeupCon = this.wakeupReleased;
                } else
                    if (id == MouseEvent.MOUSE_RELEASED)
                    {
                        fireReleased(currentResult, (MouseEvent) events[i]);
                        wakeupCon = this.wakeupPressed;
                    }

            }
        }

        this.wakeupOn(wakeupCon);
    }

    /**
     * F&uuml;gt einen PickListener ein, der bei erfolgreichen Picking
     * aufgerufen wird.
     * 
     * @param l Der zu registrierende PickListener.
     */
    public void addPickListener(PickListener l)
    {
        this.listeners.add(l);
    }

    /**
     * Entfernt einen PickListener.
     * 
     * @param l Der zu entfernende PickListener.
     */
    public void removePickListener(PickListener l)
    {
        this.listeners.remove(l);
    }

    /**
     * Ruft die <code>pressed()</code> Methoden aller registrierten PickListener
     * auf.
     * 
     * @param pR Das <code>PickResult</code>, welches beim Picken ermittelt
     *        wurde.
     * @param event Das ausl&ouml;sende Mausereignis.
     */
    private void firePressed(PickResult pR, MouseEvent event)
    {
        for (Iterator it = listeners.iterator(); it.hasNext();)
        {
            ((PickListener) it.next()).pressed(pR, event);
        }
    }

    /**
     * Ruft die <code>released()</code> Methoden aller registrierten
     * PickListener auf.
     * 
     * @param pR Das <code>PickResult</code>, welches beim Picken ermittelt
     *        wurde.
     * @param event Das ausl&ouml;sende Mausereignis.
     */
    private void fireReleased(PickResult pR, MouseEvent event)
    {
        for (Iterator it = listeners.iterator(); it.hasNext();)
        {
            ((PickListener) it.next()).released(pR, event);
        }
    }

    /* (non-Javadoc)
     * @see javax.media.j3d.Node#cloneNode(boolean)
     */
    public Node cloneNode(boolean arg0)
    {
        PickBehavior newPickB =
            new PickBehavior(pickBG, pickCanvasMode, throwCapExc);
        newPickB.duplicateNode(this, arg0);
        return newPickB;
    }

    /* (non-Javadoc)
     * @see javax.media.j3d.Node#duplicateNode(javax.media.j3d.Node, boolean)
     */
    public void duplicateNode(Node arg0, boolean arg1)
    {
        super.duplicateNode(arg0, arg1);
        this.listeners = (ArrayList) ((PickBehavior) arg0).listeners.clone();
    }

    /* (non-Javadoc)
     * @see javax.media.j3d.SceneGraphObject#updateNodeReferences(javax.media.j3d.NodeReferenceTable)
     */
    public void updateNodeReferences(NodeReferenceTable arg0)
    {
        super.updateNodeReferences(arg0);

        this.pickBG = (BranchGroup) arg0.getNewObjectReference(this.pickBG);
    }
}
