/*
 * Created on 16.03.2005
 *
 */
package interaction;

import java.awt.AWTEvent;
import java.awt.Point;
import java.awt.event.MouseEvent;
import java.util.Enumeration;

import javax.media.j3d.Behavior;
import javax.media.j3d.Canvas3D;
import javax.media.j3d.Node;
import javax.media.j3d.NodeReferenceTable;
import javax.media.j3d.Transform3D;
import javax.media.j3d.TransformGroup;
import javax.media.j3d.WakeupCondition;
import javax.media.j3d.WakeupCriterion;
import javax.media.j3d.WakeupOnAWTEvent;
import javax.media.j3d.WakeupOr;
import javax.vecmath.AxisAngle4f;
import javax.vecmath.Vector3f;

/**
 * Erm&ouml;glicht die Rotation eines Objektes um sich selbst durch
 * Mausbewegungen des Nutzers.<p/>
 * 
 * Bei gedr&uuml;ckter linker Maustaste rotiert das Objekt um die Y-Achse bzw.
 * X-Achse, je nach Mausbewegung. Bei gedr&uuml;ckter rechter Maustaste rotiert
 * das Objekt um die X- bzw. Z-Achse.
 * 
 * @author Martin Hedler
 * @version 1.0 (2005/04/25)
 */
public class RotationModificator extends Behavior
{
    private TransformGroup target;
    private Transform3D deltaTransform, srcTransform;

    private WakeupCriterion wakeupMP;
    private WakeupCondition wakeupMDR;
    private int movingPlane;
    private float scalingFactor;

    private static final int XY_AXIS = 0;
    private static final int XZ_AXIS = 1;

    private Point lastPoint;
    private Vector3f rotatingVector;
    private Vector3f transVector;
    private Vector3f zero;

    /**
     * Erstellt einen RotationModificator, welcher die TransformGroup
     * <code>target</code> manipuliert.
     * 
     * @param target Die TransformGroup, welche manipuliert werden soll.
     */
    public RotationModificator(TransformGroup target)
    {
        super();

        if (target == null)
            throw new NullPointerException();

        this.target = target;
        deltaTransform = new Transform3D();
        srcTransform = new Transform3D();
        this.movingPlane = XY_AXIS;
        scalingFactor = 0.001f;
        rotatingVector = new Vector3f();
        transVector = new Vector3f();
        zero = new Vector3f(0, 0, 0);

        wakeupMP = new WakeupOnAWTEvent(MouseEvent.MOUSE_PRESSED);
        //wakeupMD = new WakeupOnAWTEvent(MouseEvent.MOUSE_DRAGGED);
        wakeupMDR =
            new WakeupOr(
                new WakeupCriterion[] {
                    new WakeupOnAWTEvent(MouseEvent.MOUSE_RELEASED),
                    new WakeupOnAWTEvent(MouseEvent.MOUSE_DRAGGED)});
    }

    /* (non-Javadoc)
     * @see javax.media.j3d.Behavior#initialize()
     */
    public void initialize()
    {
        this.wakeupOn(wakeupMP);
    }

    /* (non-Javadoc)
     * @see javax.media.j3d.Behavior#processStimulus(java.util.Enumeration)
     */
    public void processStimulus(Enumeration enum)
    {
        WakeupCondition wkcr = wakeupMP;
        
        for (; enum.hasMoreElements();)
        {
            AWTEvent[] events =
                ((WakeupOnAWTEvent) enum.nextElement()).getAWTEvent();
            for (int i = 0; i < events.length; ++i)
            {
                MouseEvent mouseEvent = (MouseEvent) events[i];

                switch (mouseEvent.getID())
                {
                    case MouseEvent.MOUSE_PRESSED :
                        start(mouseEvent);
                        wkcr = wakeupMDR;
                        break;
                    case MouseEvent.MOUSE_DRAGGED :
                        move(mouseEvent);
                        wkcr = wakeupMDR;
                        break;
                }
            }
        }
        
        this.wakeupOn(wkcr);
    }

    /**
     * Startet die Manipulation durch Speichern des Startpunktes des Mauscursors
     * und durch Ermittlung der Rotationsachse.
     *  
     * @param e Das ausl&ouml;sende Mausereignis.
     */
    private void start(MouseEvent e)
    {
        lastPoint = e.getPoint();
        if (e.getButton() == MouseEvent.BUTTON1)
            movingPlane = XY_AXIS;
        else
            if (e.getButton() == MouseEvent.BUTTON3)
                movingPlane = XZ_AXIS;
    }

    /**
     * Manipuliert die TransformGroup (Rotation)
     * 
     * @param e Das ausl&ouml;sende Mausereignis.
     */
    private void move(MouseEvent e)
    {
        Point currentPoint = e.getPoint();

        switch (movingPlane)
        {
            case XY_AXIS :
                rotatingVector.set(
                    currentPoint.y - lastPoint.y,
                    currentPoint.x - lastPoint.x,
                    0);
                break;
            case XZ_AXIS :
                rotatingVector.set(
                    currentPoint.y - lastPoint.y,
                    0,
                    currentPoint.x - lastPoint.x);
                break;
        }

        Transform3D localToVW = new Transform3D();

        ((Canvas3D) e.getSource()).getImagePlateToVworld(localToVW);

        localToVW.transform(rotatingVector);
        rotatingVector.scale(scalingFactor);

        lastPoint = currentPoint;
        deltaTransform.setIdentity();
        deltaTransform.setRotation(
            new AxisAngle4f(rotatingVector, rotatingVector.length()));

        target.getTransform(srcTransform);

        // Translation sichern
        srcTransform.get(transVector);

        srcTransform.setTranslation(zero);

        deltaTransform.mul(srcTransform);

        // Translation zurueckschreiben
        deltaTransform.setTranslation(transVector);

        target.setTransform(deltaTransform);
    }

    /* (non-Javadoc)
     * @see javax.media.j3d.SceneGraphObject#updateNodeReferences(javax.media.j3d.NodeReferenceTable)
     */
    public void updateNodeReferences(NodeReferenceTable arg0)
    {
        super.updateNodeReferences(arg0);

        this.target = (TransformGroup) arg0.getNewObjectReference(this.target);
    }

    /* (non-Javadoc)
     * @see javax.media.j3d.Node#cloneNode(boolean)
     */
    public Node cloneNode(boolean arg0)
    {
        RotationModificator newRotMod = new RotationModificator(target);
        newRotMod.duplicateNode(this, arg0);
        return newRotMod;
    }
}
