/*
 * Created on 24.01.2005
 *
 */
package interaction.navigation;

import java.util.Enumeration;

import javax.media.j3d.Behavior;
import javax.media.j3d.BranchGroup;
import javax.media.j3d.CapabilityNotSetException;
import javax.media.j3d.GeometryArray;
import javax.media.j3d.Node;
import javax.media.j3d.NodeReferenceTable;
import javax.media.j3d.PickRay;
import javax.media.j3d.PickSegment;
import javax.media.j3d.SceneGraphPath;
import javax.media.j3d.Shape3D;
import javax.media.j3d.Transform3D;
import javax.media.j3d.TransformGroup;
import javax.media.j3d.WakeupOnElapsedFrames;
import javax.vecmath.Point3d;
import javax.vecmath.Vector3d;
import javax.vecmath.Vector3f;

import com.sun.j3d.utils.picking.PickIntersection;
import com.sun.j3d.utils.picking.PickResult;
import com.sun.j3d.utils.picking.PickTool;
import com.sun.j3d.utils.timer.J3DTimer;

/**
 * Behavior, welches die Navigation durch die Szene unter Ber&uuml;cksichtigung
 * von Hindernissen erm&ouml;glicht. Ebenfalls implementiert ist eine
 * Terrainverfolgung.
 * 
 * @author Martin Hedler
 * @version 1.0 (2005/04/26)
 */
public class GravityBehavior extends Behavior
{
    /**
     * Lokale Variablen der Methode
     * {@linkplain GravityBehavior#checkForCollision()}
     */
    private class Stack_CheckForCollision
    {
        public Vector3d dirVec, sideVector, wallNormal, move, cross;
        public PickSegment pickSegment;
        public Point3d startPoint, tmpPoint;
        public PickTool pickTool;

        public Stack_CheckForCollision(BranchGroup bg)
        {
            dirVec = new Vector3d();
            sideVector = new Vector3d();
            wallNormal = new Vector3d();
            move = new Vector3d();
            cross = new Vector3d();
            startPoint = new Point3d();
            tmpPoint = new Point3d();
            pickSegment = new PickSegment();
            pickTool = new PickTool(bg);
            pickTool.setMode(PickTool.GEOMETRY_INTERSECT_INFO);
        }
    }

    /**
     * Lokale Variablen der Methode
     * {@linkplain GravityBehavior#processStimulus(Enumeration)}
     */
    private class Stack_ProcessStimulus
    {
        public Transform3D avatarTrans, rotTrans;
        public Vector3d posVec, dirVec;
        public Vector3f zero;
        public Point3d pos, posOld;
        public boolean callSetFloor;
        public float eyeHeight;
        public float timeCounter;

        public Stack_ProcessStimulus()
        {
            avatarTrans = new Transform3D();
            rotTrans = new Transform3D();
            posVec = new Vector3d();
            dirVec = new Vector3d();
            zero = new Vector3f(0, 0, 0);
            posOld = new Point3d();
            pos = new Point3d();
            callSetFloor = false;
            eyeHeight = 1.8f;
            timeCounter = 0;
        }
    }

    /**
     * Setzt die Capability-Flags, die ben&ouml;tigt werden, damit das
     * &uuml;bergebene Objekt von der Terrainverfolgung und der
     * Kollisionsvermeidung erkannt wird.
     * 
     * @param shape Das <code>Shape3D</code> Objekt, dessen Capability-Flags
     *        gesetzt werden sollen.
     */
    public static void setNeededCapabilities(Shape3D shape)
    {
        shape.setCapability(Shape3D.ALLOW_GEOMETRY_READ);
        shape.setPickable(true);
        for (Enumeration e = shape.getAllGeometries(); e.hasMoreElements();)
        {
            GeometryArray g = (GeometryArray) e.nextElement();
            g.setCapability(GeometryArray.ALLOW_INTERSECT);
            g.setCapability(GeometryArray.ALLOW_NORMAL_READ);
        }
    }

    private float angleSpeed;
    private BranchGroup bg;

    private Stack_CheckForCollision cFCStack;
    private WakeupOnElapsedFrames crit;
    private float current_a = 0;
    private float current_SpeedLimit = 0;
    private Vector3d dir;
    private float end_a = 0.00004f;

    private long lastTime;
    private float maxSpeed = 0.005f;
    private PickRay pickRay;

    private Stack_ProcessStimulus pSS;
    private float speed;
    private float start_a = 0.00002f;
    private boolean throwCapExc;
    private TransformGroup transGrp;

    /**
     * Erstellt eine Instanz dieser Klasse.
     * 
     * @param transGrp Die <code>TransformGroup</code>, welche manipuliert
     *        werden soll. Meistens die TransformGroup der
     *        <code>ViewPlatform</code>.
     * @param bg Die <code>BranchGroup</code> unter der sich die Hindernisse und
     *           das Terrain befinden.
     * @param throwCapExc <code>true</code>, wenn
     *        <code>CapabilityNotSetException</code>s geworfen werden sollen
     *        oder <code>false</code>, wenn diese ignoriert werden soll.
     */
    public GravityBehavior(
        TransformGroup transGrp,
        BranchGroup bg,
        boolean throwCapExc)
    {
        this.throwCapExc = throwCapExc;

        this.transGrp = transGrp;
        dir = new Vector3d(0, -1, 0);
        pickRay = new PickRay();
        this.bg = bg;
        speed = 0;
        angleSpeed = 0;

        cFCStack = new Stack_CheckForCollision(bg);
        pSS = new Stack_ProcessStimulus();
    }

    /**
     * &Uuml;berpr&uml;ft, ob sich auf dem Weg von der alten Position
     * <code>posOld</code> zur neuen Position <code>pos</code> ein Hindernis
     * befindet. Ist dies der Fall, wird die neue Position entsprechend
     * abge&auml;ndert, so dass keine Kollision erfolgt.
     * 
     * @param pos Die neue Position. Wird ggf. ver&auml;ndert.
     * @param posOld Die alte Position.
     */
    private void checkForCollision(Point3d pos, Point3d posOld)
    {
        cFCStack.dirVec.sub(pos, posOld);
        cFCStack.dirVec.normalize();

        cFCStack.sideVector.set(0, 1, 0);
        cFCStack.sideVector.cross(cFCStack.dirVec, cFCStack.sideVector);
        cFCStack.sideVector.normalize();

        cFCStack.startPoint.set(posOld);

        cFCStack.wallNormal.set(0, 0, 0);

        cFCStack.move.sub(pos, posOld);

        for (float s = -1f; s <= 1f; s = s + 1f)
        {
            cFCStack.startPoint.scaleAdd(s, cFCStack.sideVector, posOld);
            cFCStack.tmpPoint.set(pos);
            cFCStack.tmpPoint.add(cFCStack.dirVec);
            cFCStack.tmpPoint.scaleAdd(
                s,
                cFCStack.sideVector,
                cFCStack.tmpPoint);

            cFCStack.pickSegment.set(cFCStack.startPoint, cFCStack.tmpPoint);

            cFCStack.pickTool.setShape(
                cFCStack.pickSegment,
                cFCStack.startPoint);

            try
            {
                // Closest prueft hier auf den IntersectionPoint (laut API).
                PickResult p = cFCStack.pickTool.pickClosest();

                // Wurde eine Bounds beruehrt?
                if (p != null)
                {
                    PickIntersection pI =
                        p.getClosestIntersection(cFCStack.startPoint);

                    if (pI == null)
                        continue;

                    Vector3f n = pI.getPointNormal();
                    cFCStack.wallNormal.set(n);
                    p.getLocalToVworld().transform(cFCStack.wallNormal);
                    cFCStack.wallNormal.normalize();
                    cFCStack.cross.cross(cFCStack.dirVec, cFCStack.wallNormal);
                    if (cFCStack.cross.y * s >= 0)
                    {
                        double dot = cFCStack.move.dot(cFCStack.wallNormal);
                        if (dot < 0)
                        {
                            cFCStack.move.scaleAdd(
                                -dot,
                                cFCStack.wallNormal,
                                cFCStack.move);
                        }
                    }
                } // if (p != null)
            } catch (CapabilityNotSetException e)
            {
                if (throwCapExc)
                    throw e;
            }
        } // for (float s = -1f; s <= 1f; s = s + 1f)

        pos.add(posOld, cFCStack.move);
    }

    /**
     * Beendet eine Vorw&auml;rts- oder R&uuml;ckw&auml;rtsbewegung durch
     * gleichm&auml;&szlig;iges ausbremsen.
     * 
     * @see #startMovingBackward()
     * @see #startMovingForward()
     */
    public void endMoving()
    {
        if (speed > 0)
            current_a = -end_a;
        else
            if (speed < 0)
                current_a = end_a;

        current_SpeedLimit = 0;
    }

    /*
     *  (non-Javadoc)
     * @see javax.media.j3d.Behavior#initialize()
     */
    public void initialize()
    {
        Point3d pos = new Point3d();
        crit = new WakeupOnElapsedFrames(0);
        lastTime = J3DTimer.getValue();

        transGrp.getTransform(pSS.avatarTrans);
        pSS.avatarTrans.get(pSS.posVec);
        pos.set(pSS.posVec);

        setToFloor(pos, 1.8f);
        //TODO Returnwert auswerten! + Konstante einfuehren (1.8)

        pSS.posVec.set(pos);
        pSS.avatarTrans.setTranslation(pSS.posVec);
        transGrp.setTransform(pSS.avatarTrans);

        this.wakeupOn(crit);
    }

    /*
     *  (non-Javadoc)
     * @see javax.media.j3d.Behavior#processStimulus(java.util.Enumeration)
     */
    public void processStimulus(Enumeration arg0)
    {
        if (speed != 0 || angleSpeed != 0 || current_a != 0)
        {
            transGrp.getTransform(pSS.avatarTrans);
            pSS.avatarTrans.get(pSS.posVec);
            pSS.pos.set(pSS.posVec);
            pSS.avatarTrans.setTranslation(pSS.zero);
            pSS.dirVec.set(0, 0, -1);
            pSS.avatarTrans.transform(pSS.dirVec);
            pSS.dirVec.normalize();

            pSS.posOld.set(pSS.pos);

            // Zeit berechnen
            long time = (J3DTimer.getValue() - lastTime) / 1000000;

            // step berechnen
            float step =
                speed * (float) time + current_a * (float) (time * time);
            speed = current_a * (float) time + speed;
            if (current_a > 0 && speed > current_SpeedLimit)
            {
                speed = current_SpeedLimit;
                current_a = 0;
            } else
                if (current_a < 0 && speed < current_SpeedLimit)
                {
                    speed = current_SpeedLimit;
                    current_a = 0;
                }

            pSS.pos.scaleAdd(step, pSS.dirVec, pSS.pos);

            // step berechnen
            step = angleSpeed * (float) time;

            float yAngle = (float) Math.toRadians(step * 7f);

            if (angleSpeed != 0)
            {
                pSS.rotTrans.rotY(yAngle);
                pSS.avatarTrans.mul(pSS.rotTrans, pSS.avatarTrans);
            }

            // Augenhoehe berechnen //TODO Noch schoener programmieren
            pSS.timeCounter += time;
            pSS.eyeHeight =
                1.8f
                    + (float) Math.sin(((float) pSS.timeCounter) / 110) * 0.14f;

            // Wurde sich die Position aendern?
            if (!pSS.pos.equals(pSS.posOld))
            {
                // Auf Kollision pruefen
                checkForCollision(pSS.pos, pSS.posOld);
                pSS.callSetFloor = !setToFloor(pSS.pos, pSS.eyeHeight);
            }

            pSS.posVec.set(pSS.pos);
            pSS.avatarTrans.setTranslation(pSS.posVec);
            transGrp.setTransform(pSS.avatarTrans);
        }

        // Faellt der Koerper?
        if (pSS.callSetFloor)
        {
            pSS.callSetFloor = !setToFloor(pSS.pos, pSS.eyeHeight);
            pSS.posVec.set(pSS.pos);
            pSS.avatarTrans.setTranslation(pSS.posVec);
            transGrp.setTransform(pSS.avatarTrans);
        }

        lastTime = J3DTimer.getValue();
        this.wakeupOn(crit);
    }

    /**
     * Setzt die Winkelgeschwindigkeit der Drehung um die lokale Y-Achse der
     * zu manipulierenden <code>TransformGroup</code>.
     * 
     * @param speed Die Winkelgeschwindigkeit in 1 / s.
     */
    public void setAngleSpeed(float speed)
    {
        this.angleSpeed = speed / 1000f;
    }

    /**
     * Setzt die lineare Geschwindigkeit in Richtung der <i>negativen</i>
     * Z-Achse der zu manipulierenden <code>TransformGroup</code>.
     * 
     * @param speed Die Geschwindigkeit in m / s.
     */
    public void setLinearSpeed(float speed)
    {
        this.speed = speed / 1000f;
    }

    /**
     * Ermittelt den n&auml;chstliegenden Untergrund zu der &uuml;bergebenen
     * Position. Je nach Lage wird eine neue Position gesetzt. Ist der
     * Untergrund n&auml;her als die angegebene <code>eyeHeight</code>, wird
     * die neue Position soweit nach oben gesetzt, dass der Abstand
     * <code>eyeHeight</code> betr&auml;gt. In diesem Fall liefert die Funktion
     * <code>true</code> zur&uml;ck.
     * <p/>
     * Ist der Abstand gr&ouml;&szlig;er als <code>eyeHeight</code>, wird die
     * Position um 0.3 Meter nach unten gesetzt und die Funktion liefert
     * <code>false</code> zur&ueck.
     * 
     * @param pos Die aktuelle Position der Augen des Betrachters. Wird
     *            ggf. ver&auml;ndert.
     * @param eyeHeight Die Soll-H&ouml;he der Augen vom Untergrund.
     * @return <code>true</code>, wenn sich der Koerper nicht im Fall
     *         befindet. Also entweder auf festem Boden steht oder kein
     *         Untergrund vorhanden auf den man fallen koennte. Ansonsten wird
     *         <code>false</code> zur&uml;ckgegeben.
     */
    private boolean setToFloor(Point3d pos, float eyeHeight)
    {
        /*
         * locals:
         * pickRay* - PickRay
         * d[] - double[]
         */

        //        callSetFloor = false;

        pickRay.set(pos, dir);
        SceneGraphPath p[] = bg.pickAll(pickRay);
        double d[] = new double[] { 100 };
        double min = 1001;
        boolean intersect;
        if (p != null)
        {
            for (int i = 0; i < p.length; ++i)
            {
                Shape3D f = (Shape3D) p[i].getObject();
                try
                {
                    intersect = f.intersect(p[i], pickRay, d);
                } catch (CapabilityNotSetException e)
                {
                    if (throwCapExc)
                        throw e;
                    else
                        intersect = false;
                }
                if (intersect && d[0] < min)
                {
                    min = d[0];
                }
            }
        }

        if (min <= eyeHeight)
        {
            pos.y = pos.y - min + eyeHeight;
        } else
            if (min < 1000 && min > eyeHeight)
            {
                pos.y = pos.y - 0.3f;
                return false;
            }

        /*
         * Koerper befindet sich nicht im Fall. Also entweder auf festem Boden
         * oder es ist kein Untergrund vorhanden auf den man fallen koennte.
         */
        return true;
    }

    /**
     * Startet die Vorw&auml;rtbewegung mit einer gleichm&auml;&szlig;igen
     * Beschleunigung.
     */
    public void startMovingBackward()
    {
        current_a = -start_a;
        current_SpeedLimit = -maxSpeed;
        if (speed > 0)
            speed = 0;
    }

    /**
     * Startet die R&uml;ckw&auml;rtbewegung mit einer gleichm&auml;&szlig;igen
     * Beschleunigung.
     */
    public void startMovingForward()
    {
        current_a = start_a;
        current_SpeedLimit = maxSpeed;
        if (speed < 0)
            speed = 0;
    }

    /* (non-Javadoc)
     * @see javax.media.j3d.Node#cloneNode(boolean)
     */
    public Node cloneNode(boolean arg0)
    {
        GravityBehavior newGrBeh =
            new GravityBehavior(transGrp, bg, throwCapExc);
        newGrBeh.duplicateNode(this, arg0);
        return newGrBeh;
    }

    /* (non-Javadoc)
     * @see javax.media.j3d.SceneGraphObject#updateNodeReferences(javax.media.j3d.NodeReferenceTable)
     */
    public void updateNodeReferences(NodeReferenceTable arg0)
    {
        super.updateNodeReferences(arg0);

        this.bg = (BranchGroup) arg0.getNewObjectReference(bg);
        this.transGrp = (TransformGroup) arg0.getNewObjectReference(transGrp);
    }
}