/*
 * Created on 12.10.2004
 *
 */
package physics;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.media.j3d.Bounds;
import javax.media.j3d.BranchGroup;
import javax.media.j3d.CapabilityNotSetException;
import javax.media.j3d.GeometryArray;
import javax.media.j3d.Group;
import javax.media.j3d.IndexedGeometryStripArray;
import javax.media.j3d.Node;
import javax.media.j3d.PickBounds;
import javax.media.j3d.SceneGraphPath;
import javax.media.j3d.Shape3D;
import javax.media.j3d.Transform3D;
import javax.vecmath.Point3f;
import javax.vecmath.Tuple3f;
import javax.vecmath.Vector3f;

import physics.geometry.GeometryArrayHandler;
import physics.geometry.GeometryArrayHandlerFactory;
import physics.geometry.Polygon;

/**
 * 
 * Implementierung der CollisionHandler Schnittstelle welche die Geometrien
 * der Objekte auf Kollisionen pr&uuml;ft.
 * 
 * @author Martin Hedler
 * @version 1.0 (2005/04/25)
 */
public class DefaultCollisionHandler implements CollisionHandler
{
    /**
     * BranchGroup, die auf Kollisionen mit den sp&auml;ter &uuml;bergebenen
     * Objekten gepr&uuml;ft werden soll.
     */
    private BranchGroup rootBranchGroup;

    /**
     * Enth&auml;lt die {@linkplain Group} Objekte, die in einem
     * Durchgang bereits auf Kollisionen kontrolliert wurden.
     */
    private ArrayList scannedGroups;

    /**
     * Enth&auml;lt die lokalToVWorld Transformation der ersten von zwei
     * miteinander kollidierenden Shape3D Objekte.
     */
    private Transform3D shape1Transform = new Transform3D();

    /**
     * Enth&auml;lt die lokalToVWorld Transformation der zweiten von zwei
     * miteinander kollidierenden Shape3D Objekte.
     */
    private Transform3D shape2Transform = new Transform3D();

    /**
     * Aktuell zu pr&uuml;fendes Polygon des ersten von zwei
     * miteinander kollidierenden Shape3D Objekte. 
     */
    private Polygon polygon1 = new Polygon();

    /**
     * Aktuell zu pr&uuml;fendes Polygon des zweiten von zwei
     * miteinander kollidierenden Shape3D Objekte. 
     */
    private Polygon polygon2 = new Polygon();

    /**
     * PickBounds mit dem Bounds Objekt des ersten von zwei
     * miteinander kollidierenden Shape3D Objekte. 
     */
    private PickBounds shape1PickBounds = new PickBounds();

    // Liste der aktuellen Shape3D Kollisionen
    private HashMap collisionMap = new HashMap();

    // Mittelpunktberechner fuer Kollisionspunkt
    private CenterTupleCalculator collisionPointCalculator =
        new CenterTupleCalculator();

    // Mittelpunktberechner fuer Kollisionsnormale
    private CenterTupleCalculator collisionNormalCalculator =
        new CenterTupleCalculator();

    private Point3f polyCrossStartPoint = new Point3f();
    private Point3f polyCrossEndPoint = new Point3f();

    private GeometryArrayHandlerFactory fact1;
    private GeometryArrayHandlerFactory fact2;

    private boolean throwCapExc;

    /**
     * Erstellt eine Instanz dieser Klasse.
     * 
     * @param bg Die <code>BranchGroup</code>, deren Objekte auf Kollisionen mit
     *        den in {@linkplain #getAllCollisions(ArrayList, ArrayList)} und
     *        {@linkplain #getCollisions(Group, ArrayList)} &uuml;bergebenen
     *        Objekten gepr&uuml;ft werden sollen.
     * @param throwCapExc <code>true</code>, wenn
     *        <code>CapabilityNotSetException</code>s geworfen werden sollen
     *        oder <code>false</code>, wenn diese ignoriert werden sollen.
     */
    public DefaultCollisionHandler(BranchGroup bg, boolean throwCapExc)
    {
        scannedGroups = new ArrayList();
        rootBranchGroup = bg;
        this.throwCapExc = throwCapExc;

        fact1 = new GeometryArrayHandlerFactory();
        fact2 = new GeometryArrayHandlerFactory();
    }

    /**
     * &Uuml;berpr&uuml;ft, ob sich die Geometrien der beiden Shape3D Objekte
     * schneiden. Der Kollisionspunkt und die Kollisionsnormale wird
     * zur&uuml;ckgegeben.
     * 
     * @param s1 Erstes Shape3D Objekt.
     * @param shape1Transform Transformation des ersten Shape3D Objekts von
     *        Lokal nach Global.
     * @param s2 Zweites Shape3D Objekt.
     * @param shape2Transform Transformation des zweiten Shape3D Objekts von
     *        Lokal nach Global.
     * @param collisionPointCalculator Calculator, in dem der Kollisionspunkt
     *        gespeichert wird.
     * @param collisionNormalCalculator Calculator, in dem die Kollisionsnormale
     *        gespeichert wird.
     * @return <code>true</code>, wenn sich die Geometrien der beiden Shape3D
     *         Objekte schneiden, sonst <code>false</code>.
     */
    private boolean testShapeGeometryCollision(
        Shape3D s1,
        Transform3D shape1Transform,
        Shape3D s2,
        Transform3D shape2Transform,
        CenterTupleCalculator collisionPointCalculator,
        CenterTupleCalculator collisionNormalCalculator)
    {
        boolean result = false;
        GeometryArrayHandler geometryHandler1, geometryHandler2;
        collisionPointCalculator.reset();
        collisionNormalCalculator.reset();

        for (int i = 0; i < s1.numGeometries(); ++i)
        {
            geometryHandler1 =
                fact1.getGeometryArrayHandler(
                    (GeometryArray) s1.getGeometry(i));

            for (int j = 0; j < s2.numGeometries(); ++j)
            {
                geometryHandler2 =
                    fact2.getGeometryArrayHandler(
                        (GeometryArray) s2.getGeometry(j));

                if (geometryCollisionTest(collisionPointCalculator,
                    collisionNormalCalculator,
                    geometryHandler1,
                    geometryHandler2,
                    shape1Transform,
                    shape2Transform))
                {
                    result = true;
                }
            }
        }

        return result;
    }

    /**
     * &Uuml;berpr&uuml;ft alle Polygone in <code>geometryHandler1</code> auf
     * Schnitte mit Polygonen in <code>geometryHandler2</code>.
     * Die ben&ouml;tigten Transformation von lokal nach global f&uuml;r die
     * jeweiligen Geometrien m&uuml;ssen mit &uuml;bergeben werden.
     * Die Ergebnisse (Kollisionspunkt und Kollisionsnormale) werden in
     * <code>centerPointCalculator</code> und <code>normalCalculator</code>
     * gespeichert.
     * 
     * @param centerPointCalculator Um den Kollisionspunkt zu speichern.
     * @param normalCalculator Um die Kollisionsnormale zu speichern.
     * @param geometryHandler1 Handler f&uuml;r den Zugriff auf die Polygone.
     * @param geometryHandler2 Handler f&uuml;r den Zugriff auf die Polygone.
     * @param shape1Transform Transformation von lokal zu global f&uuml;r alle
     *        Polygone aus <code>geometryHandler1</code>.
     * @param shape2Transform Transformation von lokal zu global f&uuml;r alle
     *        Polygone aus <code>geometryHandler2</code>.
     * @return <code>true</code>, wenn sich Polygone schneiden, sonst
     *         <code>false</code>.
     */
    private boolean geometryCollisionTest(
        CenterTupleCalculator centerPointCalculator,
        CenterTupleCalculator normalCalculator,
        GeometryArrayHandler geometryHandler1,
        GeometryArrayHandler geometryHandler2,
        Transform3D shape1Transform,
        Transform3D shape2Transform)
    {
        /*
         * Verwendete globale Variablen:
         * polygon1,
         * polygon2,
         * polyCrossStartPoint,
         * polyCrossEndPoint,
         * polyCollisionNormal        
         */
        boolean result = false;

        geometryHandler1.reset();
        geometryHandler2.reset();

        for (; geometryHandler1.hasMorePolygons();)
        {
            geometryHandler1.getNextPolygon(polygon1);
            shape1Transform.transform(polygon1.a);
            shape1Transform.transform(polygon1.b);
            shape1Transform.transform(polygon1.c);
            shape1Transform.transform(polygon1.normal);
            polygon1.calculatePlaneNormalForm();

            for (geometryHandler2.reset(); geometryHandler2.hasMorePolygons();)
            {
                geometryHandler2.getNextPolygon(polygon2);
                shape2Transform.transform(polygon2.a);
                shape2Transform.transform(polygon2.b);
                shape2Transform.transform(polygon2.c);
                shape2Transform.transform(polygon2.normal);
                polygon2.calculatePlaneNormalForm();

                if (polygon1
                    .collided(
                        polygon2,
                        polyCrossStartPoint,
                        polyCrossEndPoint,
                        polyCollisionNormal))
                {
                    centerPointCalculator.addPoint(polyCrossStartPoint);
                    centerPointCalculator.addPoint(polyCrossEndPoint);
                    normalCalculator.addPoint(polyCollisionNormal);
                    result = true;
                }
            }
        }

        return result;
    }

    private Vector3f polyCollisionNormal = new Vector3f();

    /**
     * Ermittelt alle Kollisionen eines Shape3D Objekts mit den Shape3D Objekten
     * einer der Wurzel-BanchGroup.
     * 
     * @param s Das auf Kollisionen zu &uuml;berpr&uuml;fende Shape3D Objekt.
     * @param scannedGroups Liste von Groups, die bereits
     *        kontrolliert wurden und somit nicht noch einmal
     *        &uuml;berpr&uuml;ft werden m&uuml;ssen.
     * @param collisionMap In dieser HashMap werden die Kollisionen gespeichert.
     * @return <code>true</code>, wenn das Shape3D Objekt mit anderen Objekten
     *         kollidiert, sonst <code>false</code>.
     */
    private boolean testShapeCollision(
        Shape3D s,
        List scannedGroups,
        HashMap collisionMap)
    {
        /*
         * Verwendete globale Variablen:
         * shape1PickBounds
         * collisionPointCalculator
         * collisionNormalCalculator
         * rootBranchGroup
         */

        boolean result = false;
        Group dyTrGrp;
        Bounds shapeBounds;
        SceneGraphPath currentSGP;

        // Bounds der Shape ermitteln und transformieren
        s.getLocalToVworld(shape1Transform);
        shapeBounds = (Bounds) s.getBounds().clone();
        shapeBounds.set(s.getBounds());
        shapeBounds.transform(shape1Transform);

        // Alle Shapes picken, die mit shapeBounds kollidieren.
        shape1PickBounds.set(shapeBounds);
        SceneGraphPath sGP[] = rootBranchGroup.pickAll(shape1PickBounds);

        // Wurde etwas gefunden?
        if (sGP != null)
            // Die gefundenen Objekte auswerten
            for (int i = 0; i < sGP.length; ++i)
            {
                try
                {
                    currentSGP = sGP[i];

                    // Wurde ein anderes Objekt gefunden als das Shape selbst?
                    if (currentSGP.getObject() != s)
                    {
                        // Handelt es sich um ein Shape?
                        if (currentSGP.getObject() instanceof Shape3D)
                        {
                            dyTrGrp = null;
                            // Ist das gefundene Shapeobjekt Teil einer Group?
                            if (currentSGP.nodeCount() > 0)
                            {
                                dyTrGrp = (Group) currentSGP.getNode(0);

                                if (scannedGroups.contains(dyTrGrp))
                                {
                                    // Naechstes Objekt behandeln
                                    continue;
                                }
                            }

                            (
                                (Shape3D) currentSGP
                                    .getObject())
                                    .getLocalToVworld(
                                shape2Transform);

                            // Schneiden sich die Geometrien der beiden Shapeobjekte?
                            if (testShapeGeometryCollision(s,
                                shape1Transform,
                                (Shape3D) currentSGP.getObject(),
                                shape2Transform,
                                collisionPointCalculator,
                                collisionNormalCalculator))
                            {
                                result = true;

                                // Liste von Kollisionen mit dieser Group laden
                                List shapeCollisionList =
                                    getCollisionList(collisionMap, dyTrGrp);

                                // Neues Element anlegen
                                Collision collision = new Collision();
                                collision.node1 = s;
                                collision.node2 =
                                    (Shape3D) currentSGP.getObject();
                                collisionPointCalculator.getCenterPoint(
                                    collision.point);
                                collisionNormalCalculator.getCenterPoint(
                                    collision.normal);
                                collision.normal.normalize();
                                // Element speichern
                                shapeCollisionList.add(collision);
                            }
                        } // if (sGP[i].getObject() instanceof Shape3D)
                    } // if (sGP[i].getObject() != s)
                } // try
                catch (CapabilityNotSetException e)
                {
                    if (throwCapExc)
                        throw e;
                }
            } // for (int i = 0; i < sGP.length; ++i)

        return result;
    }

    /**
     * Setzt die ben&ouml;tigten Capability-Flags von <code>group</code>, damit
     * diese f&uuml;r die Kollision herangezogen wird.
     * 
     * @param group Der <code>Group</code> Knoten, dessen Capability-Flags
     *        gesetzt werden sollen.
     */
    public static void setNeededCapabilities(Group group)
    {
        group.setCapability(DynamicTransformGroup.ENABLE_PICK_REPORTING);
        group.setCapability(Group.ALLOW_CHILDREN_READ);
    }

    /**
     * Setzt die ben&ouml;tigten Capability-Flags von <code>shape</code>, damit
     * dieses f&uuml;r die Kollision herangezogen wird.
     * 
     * @param shape Der <code>Shape3D</code> Knoten, dessen Capability-Flags
     *        gesetzt werden sollen.
     */
    public static void setNeededCapabilities(Shape3D shape)
    {
        shape.setCapability(Shape3D.ALLOW_GEOMETRY_READ);
        shape.setCapability(Shape3D.ALLOW_LOCAL_TO_VWORLD_READ);
        shape.setCapability(Shape3D.ALLOW_BOUNDS_READ);
        shape.setPickable(true);
        for (Enumeration e = shape.getAllGeometries(); e.hasMoreElements();)
        {
            GeometryArray g = (GeometryArray) e.nextElement();
            g.setCapability(GeometryArray.ALLOW_COORDINATE_READ);
            g.setCapability(GeometryArray.ALLOW_COUNT_READ);
            g.setCapability(GeometryArray.ALLOW_NORMAL_READ);
            g.setCapability(GeometryArray.ALLOW_FORMAT_READ);

            if (g instanceof IndexedGeometryStripArray)
            {
                g.setCapability(
                    IndexedGeometryStripArray.ALLOW_COORDINATE_INDEX_READ);
                g.setCapability(
                    IndexedGeometryStripArray.ALLOW_NORMAL_INDEX_READ);
            }
        }
    }

    /**
     * Ermittelt, ob die Shape3D Objekte der &uuml;bergebene Group mit anderen
     * Shape3D Objekten in der Wurzel-BranchGroup kollidieren.
     * 
     * @param g Die zu &uuml;berpr&uuml;fende Group.
     * @param scannedGroups Liste von Groups, welche bereits
     *        kontrolliert wurden.
     * @param collisionMap In dieser HashMap werden die gefundenen Kollisionen
     *        gespeichert.
     * @return
     */
    private boolean testGroupCollision(
        Group g,
        List scannedGroups,
        HashMap collisionMap)
    {
        Node node;
        boolean result = false;

        for (int i = 0; i < g.numChildren(); ++i)
        {
            node = (Node) g.getChild(i);

            if (node instanceof Shape3D)
            {
                if (testShapeCollision((Shape3D) node,
                    scannedGroups,
                    collisionMap))
                {
                    result = true;
                }
            } else
                if (node instanceof Group)
                {
                    if (testGroupCollision((Group) node,
                        scannedGroups,
                        collisionMap))
                    {
                        result = true;
                    }
                }
        }

        return result;
    }

    private List getCollisionList(HashMap hashMap, Group dyTrGrp)
    {
        List collisionList;
        Object object;

        object = hashMap.get(dyTrGrp);

        // Ist kein CollisionMemory Objekt vorhanden?
        if (object == null)
        {
            // Liste erstellen
            collisionList = new ArrayList();

            // Unter angegebenen Schluessel einfuegen
            hashMap.put(dyTrGrp, collisionList);
        } else
        {
            collisionList = (List) object;
        }

        return collisionList;
    }

    /*
     *  (non-Javadoc)
     * @see physic.CollisionHandler#getAllCollisions(java.util.ArrayList)
     */
    synchronized public void getAllCollisions(
        ArrayList dynTraGrpList,
        ArrayList collisionList)
    {
        Group currentDynamic;

        collisionList.clear();
        this.scannedGroups.clear();

        // Alle gespeicherten Objekte durchgehen
        for (Iterator dyTrGrpIterator = dynTraGrpList.iterator();
            dyTrGrpIterator.hasNext();
            )
        {
            currentDynamic = (Group) dyTrGrpIterator.next();

            // In Liste der gescannten Gruppen aufnehmen
            scannedGroups.add(currentDynamic);
            collisionMap.clear();
            try
            {
                if (testGroupCollision(currentDynamic,
                    scannedGroups,
                    collisionMap))
                {
                    // Liste der Kollisionen erstellen
                    generateCollisionList(currentDynamic, collisionList);
                }
            } catch (CapabilityNotSetException e)
            {
                if (throwCapExc)
                    throw e;
            }
        }
    }

    /*
     * Die Kollisionen von verschiedenen Shapes einer Grp.
     * buendeln und danach die Kollisionsnormale berechnen.
     * Bei den Shapes, die keiner Dyn.Tra.Grp. angehoehren, werden die
     * Normalen sofort berechnet.
     */
    private void generateCollisionList(
        Group currDynTraGrp,
        ArrayList collisionList)
        throws RuntimeException
    {
        Collision shapeCollision = null;
        Collision collision = null;

        Set entrySet = collisionMap.entrySet();
        Map.Entry entry;
        List shapeCollisionList;
        Group dyTrGrp;

        for (Iterator entryIt = entrySet.iterator(); entryIt.hasNext();)
        {
            /*
             * Der Schluessel enthaelt die Referenz auf die Group. Der
             * zugehoerige Wert ist die Liste von Shape Collisions.
             */
            entry = (Map.Entry) entryIt.next();
            dyTrGrp = (Group) entry.getKey();
            shapeCollisionList = (List) entry.getValue();

            collisionPointCalculator.reset();
            collisionNormalCalculator.reset();

            // Alle ShapeCollisions durchgehen
            for (Iterator shapeCollIt = shapeCollisionList.iterator();
                shapeCollIt.hasNext();
                )
            {
                shapeCollision = (Collision) shapeCollIt.next();
                // Kollisionspunkte zusammenrechnen
                collisionPointCalculator.addPoint(shapeCollision.point);

                // Kollisionsnormalen zusammenrechnen
                collisionNormalCalculator.addPoint(shapeCollision.normal);
            }

            /*
             * Letztes Shapecollision Objekt verwenden um kein neues Objekt
             * anlegen zu muessen
             */
            collision = shapeCollision;
            collision.node1 = currDynTraGrp;
            collision.node2 = dyTrGrp;
            collisionPointCalculator.getCenterPoint(collision.point);
            // Entgueltige Normale berechnen
            collisionNormalCalculator.getCenterPoint(collision.normal);
            collision.normal.normalize();

            if (Float.isNaN(collision.normal.x))
                break; // Kollision nicht speichern

            // Kollision speichern
            collisionList.add(collision);
        } // for (Iterator entryIt = entrySet.iterator(); entryIt.hasNext();)
    }

    private class CenterTupleCalculator
    {
        private Point3f minTuple;
        private Point3f maxTuple;
        private boolean firstTuple;

        public CenterTupleCalculator()
        {
            minTuple = new Point3f();
            maxTuple = new Point3f();
            firstTuple = true;
        }

        public void reset()
        {
            firstTuple = true;
        }

        /**
         * F&uuml;gt einen weiteren Punkt hinzu.
         * 
         * @param point
         */
        public void addPoint(Tuple3f tuple)
        {
            if (firstTuple)
            {
                firstTuple = false;
                minTuple.set(tuple);
                maxTuple.set(tuple);
            } else
            {
                minTuple.x = Math.min(minTuple.x, tuple.x);
                minTuple.y = Math.min(minTuple.y, tuple.y);
                minTuple.z = Math.min(minTuple.z, tuple.z);
                maxTuple.x = Math.max(maxTuple.x, tuple.x);
                maxTuple.y = Math.max(maxTuple.y, tuple.y);
                maxTuple.z = Math.max(maxTuple.z, tuple.z);
            }
        }

        public void getCenterPoint(Tuple3f tuple)
        {
            tuple.add(maxTuple, minTuple);
            tuple.scale(0.5f);
        }
    }

    /* (non-Javadoc)
     * @see physic.CollisionHandler#getCollisions(physic.DynamicTransformGroup, java.util.ArrayList)
     */
    synchronized public void getCollisions(
        Group dyTrGrp,
        ArrayList collisionList)
    {
        collisionList.clear();
        this.scannedGroups.clear();
        this.scannedGroups.add(dyTrGrp);

        collisionMap.clear();
        try
        {
            if (testGroupCollision(dyTrGrp, scannedGroups, collisionMap))
            {
                // Liste der Kollisionen erstellen
                generateCollisionList(dyTrGrp, collisionList);
            }
        } catch (CapabilityNotSetException e)
        {
            if (throwCapExc)
                throw e;
        }
    }

    /* (non-Javadoc)
     * @see physics.CollisionHandler#setRootBranchGroup(javax.media.j3d.BranchGroup)
     */
    public void setRootBranchGroup(BranchGroup bg)
    {
        rootBranchGroup = bg;
    }
}
