package physics;
import javax.media.j3d.Node;
import javax.media.j3d.Transform3D;
import javax.media.j3d.TransformGroup;
import javax.vecmath.AxisAngle4f;
import javax.vecmath.Vector3f;

/*
 * Created on 30.06.2004
 *
 */

/**
 * Diese Klasse reprsentiert einen starren K&ouml;rper. Dieser K&ouml;rper ist
 * definiert durch eine Masse, einen Tr&auml;gheitsmoment, seine
 * Geschwindigkeit, seine Transformation, einen Sto&szlig;faktor und seine
 * Ausdehnung.
 * <p/>
 * Die Ausdehnung des K&ouml;rpers wird durch die Objekte bestimmt, welche
 * unter diese <code>DynanimcTransformGroup</code> eingef&uuml;gt werden. Neben
 * den bereits genannten Gr&ouml;&szlig;en kann auch die Beschleunigung, die auf
 * den K&ouml;rper wirkt, gesetzt werden.
 * <p/>
 * Der Sto&szlig;faktor gibt an, wieviel kinetische Energie bei einem Sto&szlig;
 * erhalten bleibt. Der Gesamtsto&szlig;faktor f&uuml; den Sto&szlig; ergibt
 * sich aus dem Produkt der Sto&szlig;faktoren der beiden K&ouml;rper. Ist
 * dieses Produkt 1, handelt es sich um einen vollst&auml;ndig elastischen
 * Sto&szlig;. Ist das Produkt 0 ist der Sto&szlig; vollst&auml;ndig
 * inelastisch. Der Sto&szlig;faktor kann zwischen 0 und 1 liegen.
 * <p/>
 * Ein K&ouml;rper kann schlafen gelegt werden, d.h. er wird nicht bewegt oder
 * von sich aus auf Kollisionen gepr&ueft, bis ein anderer K&ouml;rper mit ihm
 * kollidiert oder er manuell aufgeweckt wird (siehe
 * {@linkplain #setSleep(boolean)}).
 * 
 * @author Martin Hedler
 * @version 1.0 (05/04/25)
 */
public class DynamicTransformGroup extends TransformGroup
{
    /** Masse des K&ouml;rpers. */
    private float mass;

    /** Tr&auml;gheitsmoment des K&ouml;rpers. */
    private float I;

    /** Lineare Geschwindigkeit des K&ouml;rpers. */
    private Vector3f speed;

    /** Lineare Beschleunigung des K&ouml;rpers. */
    private Vector3f acceleration;

    /** Winkelbeschleunigung des K&ouml;rpers. */
    private Vector3f angleAcceleration;

    /** Winkelgeschwindigkeit des K&ouml;rpers. */
    private Vector3f angleSpeed;

    /** Sto&szlig;faktor des K&ouml;rpers. */
    private float energyPushFactor;

    /** Simulationsstatus des K&ouml;rpers. */
    private boolean sleeping;

    /** Tempor&auml;re Variable. */
    private Transform3D trans;

    /** Tempor&auml;re Variable. */
    private AxisAngle4f axis;

    /** Tempor&auml;re Variable. */
    private Vector3f position;

    /** Tempor&auml;re Variable. */
    private Vector3f rotation;

    /** Tempor&auml;re Variable. */
    private Transform3D addTrans;

    /** Nullvektor. */
    private Vector3f zero;

    /**
     * Erstellt einen starren K&ouml;rper, mit einer Masse von 1 kg und einem
     * Tr&auml;gheitsmoment von 0.4 kg * m^2 (Tr&auml;gheitsmoment einer Kugel
     * dieser Masse und einem Radius von einem Meter). Der Sto&szlig;faktor des
     * K&ouml;rpers ist 1. Die Geschwindigkeit und Beschleunigung des
     * K&ouml;rpers sind 0.
     */
    public DynamicTransformGroup()
    {
        this(
            new Transform3D(),
            1,
            0.4f,
            1,
            new Vector3f(),
            new Vector3f(),
            new Vector3f(),
            new Vector3f(),
            false);
    }

    /**
     * Liefert den Sto&szlig;faktor zur&uuml;ck.
     * 
     * @return Der Sto&szlig;faktor des K&ouml;rpers.
     */
    public float getEnergyPushFactor()
    {
        return energyPushFactor;
    }

    /**
     * Setzt den Sto&szlig;faktor des K&ouml;rpers.
     * Der Sto&szlig;faktor gibt an, wieviel kinetische Energie bei einem
     * Sto&szlig; erhalten bleibt. Der Gesamtsto&szlig;faktor f&uuml; den
     * Sto&szlig; ergibt sich aus dem Produkt der Sto&szlig;faktoren der beiden
     * K&ouml;rper. Ist dieses Produkt 1, handelt es sich um einen
     * vollst&auml;ndig elastischen Sto&szlig;. Ist das Produkt 0 ist der
     * Sto&szlig; vollst&auml;ndig inelastisch. Der Sto&szlig;faktor kann
     * zwischen 0 und 1 liegen.
     *      
     * @param value Faktor zwischen einschlie&szlig;lich 0 und 1.
     * @throws IllegalArgumentException Wenn der Wert nicht zwischen 0 und 1
     *         liegt.
     */
    synchronized public void setEnergyPushFactor(float value)
        throws IllegalArgumentException
    {
        if (value < 0f || value > 1f)
            throw new IllegalArgumentException("Wert muss zwischen einschlielich 0 und 1 liegen.");

        this.energyPushFactor = value;
    }

    /**
     * Erstellt einen starren K&ouml;rper, mit den angegebenen Werten.
     * 
     * @param transform Die Transformation des K&ouml;rpers.
     * @param mass Die Masse des K&ouml;rpers in Kilogramm.
     * @param I Der Tr&auml;gheitsmoment in kg * m^2.
     * @param energyPushFactor Wert zwischen 0 und 1 der angibt, wie elastisch
     *        sich ein K&ouml;rper bei einem Sto&szlig; verh&auml;lt.
     * @param speed Lineare Startgeschwindigkeit des K&ouml;rpers in m / s.
     * @param angleSpeed Startwinkelgeschwindigkeit in 1 / s.
     * @param acceleration Lineare Beschleunigung des K&ouml;rpers in m / s^2.
     * @param angleAcceleration Winkelbeschleunigung in 1 / s^2.
     * @param sleeping <code>true</code>, wenn der K&ouml;rper schlafen soll,
     *        sonst <code>false</code>.
     */
    public DynamicTransformGroup(
        Transform3D transform,
        float mass,
        float I,
        float energyPushFactor,
        Vector3f speed,
        Vector3f angleSpeed,
        Vector3f acceleration,
        Vector3f angleAcceleration,
        boolean sleeping)
    {
        super(transform);

        this.mass = mass;
        this.I = I; //Fuer Kugel: J = 0.4f * mass * radius^2;
        this.energyPushFactor = energyPushFactor;
        this.speed = new Vector3f(speed);
        this.angleSpeed = new Vector3f(angleSpeed);
        this.acceleration = new Vector3f(acceleration);
        this.angleAcceleration = new Vector3f(angleAcceleration);
        this.sleeping = sleeping;

        trans = new Transform3D();
        position = new Vector3f();
        axis = new AxisAngle4f();
        rotation = new Vector3f();
        addTrans = new Transform3D();
        zero = new Vector3f();

        this.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
        this.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
    }

    /**
     * Setzt die Kraft (in N), welche direkt auf den Schwerpunkt des
     * K&ouml;rpers einwirkt.
     * 
     * @param f Die einwirkende Kraft.
     */
    synchronized public void setLinearForce(Vector3f f)
    {
        this.acceleration.scale(1f / this.mass, f);
    }

    /**
     * Gibt die Kraft zur&uuml;ck, die auf den Schwerpunkt des K&ouml;rpers
     * wirkt.
     * 
     * @param f Vektor, in den die Kraft geschrieben wird.
     */
    public void getLinearForce(Vector3f f)
    {
        f.scale(mass, this.acceleration);
    }

    /**
     * Setzt den Drehmoment, der an den K&ouml;rper angreift in N * m.
     * 
     * @param f Der wirkende Drehmoment.
     */
    synchronized public void setAngleForce(Vector3f f)
    {
        this.angleAcceleration.scale(1f / I, f);
    }

    /**
     * Setzt die Winkelbeschleunigung des K&ouml;rpers in 1 / s^2.
     * 
     * @param a Die Winkelbeschleunigung des K&ouml;rpers.
     */
    synchronized public void setAngleAcceleration(Vector3f a)
    {
        this.angleAcceleration.set(a);
    }

    /**
     * Gibt die Winkelbeschleunigung des K&ouml;rpers zur&uuml;ck.
     * 
     * @param a Vektor, in den die Winkelbehschleunigung geschrieben wird.
     */
    public void getAngleAcceleration(Vector3f a)
    {
        a.set(angleAcceleration);
    }

    /**
     * Gibt den Drehmoment zur&uuml;ck, der auf den K&ouml;rper wirkt.
     * 
     * @param f Vektor, in den der einwirkende Drehmoment gespeichert wird.
     */
    public void getAngleForce(Vector3f f)
    {
        f.scale(I, this.angleAcceleration);
    }

    /**
     * Setzt die Kraft (in N), welche auf den K&ouml;rper an dem Punkt
     * <code>radius</code> wirkt.
     * 
     * @param f Die einwirkende Kraft.
     * @param radius Position des Punktes, an dem die Kraft angreift.
     */
    synchronized public void setForce(Vector3f f, Vector3f radius)
    {
        if (radius.lengthSquared() > 0)
        {
            Vector3f force = new Vector3f();
            force.cross(radius, f);
            this.setAngleForce(force);

            force.set(radius);
            force.normalize();
            force.scale(-f.dot(force), f);
            this.setLinearForce(force);
        } else
        {
            this.setLinearForce(f);
        }
    }

    /**
     * Ermittelt den Simulationsstatus des K&ouml;rpers.
     * 
     * @return <code>true</code>, wenn der K&ouml;rper schl&auml;ft, sonst
     *         <code>false</code>.
     */
    public boolean isSleep()
    {
        return sleeping;
    }

    /**
     * Setzt den Simulationsstatus des K&ouml;rpers.
     * 
     * @param s <code>true</code>, wenn sich der K&ouml;rper nicht bewegen, also
     *        schlafen soll, sonst <code>false</code>.
     */
    public void setSleep(boolean s)
    {
        sleeping = s;
    }

    /**
     * Erstellt einen starren K&ouml;rper, mit einer Masse von 1 kg und einem
     * Tr&auml;gheitsmoment von 0.4 kg * m^2 (Tr&auml;gheitsmoment einer Kugel
     * dieser Masse und einem Radius von einem Meter). Der Sto&szlig;faktor des
     * K&ouml;rpers ist 1. Die Geschwindigkeit und Beschleunigung des
     * K&ouml;rpers sind 0.
     * 
     * @param trans Die Position und Orientierung des K&ouml;rpers.
     */
    public DynamicTransformGroup(Transform3D trans)
    {
        this(
            trans,
            1,
            0.4f,
            1,
            new Vector3f(),
            new Vector3f(),
            new Vector3f(),
            new Vector3f(),
            false);
    }

    /**
     * Die Winkelgeschwindigkeit des K&ouml;rpers in 1 / s.
     * 
     * @param aS Die Winkelgeschwindigkeit.
     */
    synchronized public void setAngleSpeed(Vector3f aS)
    {
        this.angleSpeed.set(aS);
    }

    /**
     * Gibt die aktuelle Winkelgeschwindigkeit des K&ouml;rpers zur&uuml;ck.
     * 
     * @param aS Vektor, in den die Geschwindigkeit gespeichert wird.
     */
    public void getAngleSpeed(Vector3f aS)
    {
        aS.set(this.angleSpeed);
    }

    /**
     * Setzt die Beschleunigung des K&ouml;rpers in m / s^2.
     * 
     * @param a Die lineare Beschleunigung des K&ouml;rpers.
     */
    synchronized public void setAcceleration(Vector3f a)
    {
        this.acceleration.set(a);
    }

    /**
     * Gibt die Beschleunigung des K&ouml;rpers zur&uuml;ck.
     * 
     * @param a Vektor in den die lineare Beschleunigung geschrieben wird.
     */
    public void getAcceleration(Vector3f a)
    {
        a.set(this.acceleration);
    }

    /**
     * Bewegt den K&ouml;rper &uuml;ber der Zeit <code>deltaTime</code> auf
     * einer ballistischen Bahn.
     * 
     * @param deltaTime Der Zeitschritt, der die Dauer der Bewegung angibt. 
     */
    synchronized public void nextStep(float deltaTime)
    {
        // deltaTime in sekunden umrechnen
        deltaTime = deltaTime / 1000f;

        addTrans.setIdentity();
        super.getTransform(trans);
        trans.get(position);
        trans.setTranslation(zero);

        rotation.scale(deltaTime * deltaTime * 0.5f, angleAcceleration);
        rotation.scaleAdd(deltaTime, angleSpeed, rotation);
        axis.set(rotation, rotation.length());
        addTrans.setRotation(axis);

        addTrans.mul(trans);

        position.scaleAdd(deltaTime * deltaTime * 0.5f, acceleration, position);
        position.scaleAdd(deltaTime, speed, position);

        // Neue Geschwindigkeit berechnen
        speed.scaleAdd(deltaTime, acceleration, speed);

        addTrans.setTranslation(position);

        super.setTransform(addTrans);
    }

    /**
     * Liefert die Masse zur&uuml;ck.
     * 
     * @return Die Masse des K&ouml;rpers.
     */
    public float getMass()
    {
        return this.mass;
    }

    /**
     * Setzt die Masse des K&ouml;rpers in kg.
     * 
     * @param mass Die Masse.
     */
    synchronized public void setMass(float mass)
    {
        this.mass = mass;
    }

    /**
     * Gibt die lineare Geschwindigkeit des K&ouml;rpers zur&uuml;ck.
     * 
     * @param speed Vektor, in den die Geschwindigkeit geschrieben wird.
     */
    public void getSpeed(Vector3f speed)
    {
        speed.set(this.speed);
    }

    /**
     * Setzt die lineare Geschwindigkeit des K&ouml;rpers in m / s.
     * 
     * @param speed Die lineare Geschwindigkeit.
     */
    synchronized public void setSpeed(Vector3f speed)
    {
        this.speed.set(speed);
    }

    /**
     * Liefert den Dehmoment des K&ouml;rpers zur&uuml;ck.
     * 
     * @return Der Drehmoment.
     */
    public float getI()
    {
        return I;
    }

    /**
     * Setzt den Tr&auml;gheitsmoment des K&ouml;rpers in kg * m^2.
     * 
     * @param i Der Tr&auml;gheitsmoment.
     */
    synchronized public void setI(float i)
    {
        this.I = i;
    }

    /**
     * Setzt die Impuls&auml;nderung des K&ouml;rpers in kg * m / s.
     * 
     * @param deltaP Die Impuls&auml;nderung bzw. der Kraftsto&szlig;.
     * @param collisionRadius Der Position, bei der der Kraftsto&szlig;
     *        angreift.
     */
    synchronized public void setDeltaMomentum(
        Vector3f deltaP,
        Vector3f collisionRadius)
    {
        // Lineare Geschwindigkeit
        Vector3f deltaV = new Vector3f(deltaP);
        deltaV.scale(1f / this.mass);

        this.speed.add(deltaV);

        // Winkelgeschwindigkeit
        deltaV.cross(collisionRadius, deltaP);
        deltaV.scale(1f / this.getI());
        this.angleSpeed.add(deltaV);
    }

    /* (non-Javadoc)
     * @see javax.media.j3d.Node#cloneNode(boolean)
     */
    public Node cloneNode(boolean arg0)
    {
        this.getTransform(trans);
        DynamicTransformGroup newDynTrGrp =
            new DynamicTransformGroup(
                trans,
                mass,
                I,
                energyPushFactor,
                speed,
                angleSpeed,
                acceleration,
                angleAcceleration,
                sleeping);
        newDynTrGrp.duplicateNode(this, arg0);
        return newDynTrGrp;
    }

}
