/*
 * Created on 02.02.2005
 *
 */
package physics;

import java.util.ArrayList;
import java.util.Iterator;

import javax.media.j3d.Node;
import javax.media.j3d.Transform3D;
import javax.vecmath.Point3f;
import javax.vecmath.Vector3f;

import com.sun.j3d.utils.timer.J3DTimer;

/**
 * Diese Klasse f&uuml;hrt, bei Aufruf der entsprechenden Methode, einen
 * Simulationschritt durch. Dabei werden zwei unterschiedliche Verfahren zur
 * Berechnung dieses Schrittes angeboten.
 * <p/>
 * Zum einen das <code>ALL_STEP</code> Verfahren: In diesem Modus werden bei
 * jedem Aufruf alle registrierten Objekte bewegt, danach auf Kollisionen
 * gepr&uuml;ft und eventuelle Kollisionsreaktionen durchgef&uuml;hrt. Dies hat
 * den Vorteil, dass sich alle Objekte quasi-gleichzeitig bewegen. Der Nachteil
 * ist, dass ein Simulationsschritt erst beendet ist, wenn alle Objekte
 * abgearbeitet wurden, was bedeutet, dass der Thread, welcher diese Methode
 * aufruft, f&uuml;r unbestimmte Zeit blockiert wird.
 * <p/>
 * Im Simulationsmodus <code>SINGLE_STEP</code> wird versucht das Auftreten
 * dieses Problems zu verringern. Die Berechnung der Bewegung, das Pr&uuml;fen
 * auf Kollisionen und die daraus resultierende Kollisionsreaktionen werden fr
 * jedes Objekt einzeln durchgef&uuml;hrt. Wurde ein Objekt auf diese Weise
 * verarbeitet, wird die verstrichene Zeit &uuml;berpr&uuml;ft. &Uuml;bersteigt
 * diese einen festgesetzten Grenzwert, wird die weitere Bearbeitung
 * abgebrochen, so dass der aufrufende Thread mit hoher Wahrscheinlichkeit nicht
 * l&auml;nger als das vorgegebene Zeitlimit blockiert wird.
 * Beim n&auml;chsten Simulationsschritt wird bei dem n&auml;chsten noch nicht
 * bearbeiteten Objekt fortgesetzt. Auf diese Weise ist es auch m&ouml;glich,
 * dass w&auml;hrend eines Simulationsschrittes die K&ouml;rper mehrmals
 * abgearbeitet werden k&ouml;nnen, was die Genauigkeit der Simulation
 * erh&ouml;ht.
 * 
 * @author Martin Hedler
 * @version 1.0 (05/04/25)
 */
public class PhysicalEnvironment
{
    //  TODO Reibung ist noch nicht implementiert
    //  TODO Stroemungswiederstand nicht implementiert
    //  TODO Kraftfelder

    private ArrayList dynamicObjectList;
    private CollisionHandler collisionHandler;

    private PhysicalBody body1, body2;
    private long lastTime;
    private ArrayList collisionList;
    private int type;
    private float defaultEnergyPushFactor;
    private long maxDeltaTime;
    private long maxCalculationTime;
    private int single_stepCurrentIndex;

    private Transform3D tmpTransform;
    private Vector3f deltaImpulse;
    private Vector3f deltaVP2;
    // Geschwindigkeitsaenderung des Punktes P2
    private Vector3f deltaVP1;
    // Geschwindigkeitsaenderung des Punktes P1
    private Vector3f tmpVec;

    /** F&uuml;r den entsprechenden Simulationsmodus. */
    public static final int SINGLE_STEP = 0;

    /** F&uuml;r den entsprechenden Simulationsmodus. */
    public static final int ALL_STEP = 1;

    /**
     * Erstellt eine Instanz der Klasse mit dem angegebenen
     * {@linkplain CollisionHandler} und f&uuml;r den angegebenen
     * Simulationsmodus mit folgenden Werten:<br/>
     * <code>maxDeltaTime = Long.MAX_VALUE</code>,<br/>
     * <code>maxCalculationTime = 20</code>,<br/>
     * <code>defaultEnergyPushFactor = 1</code>.
     * 
     * @param ch Der verwendete {@linkplain CollisionHandler}.
     * @param simulationType <code>ALL_STEP</code> oder
     *        <code>SINGLE_STEP</code>.
     * @throws IllegalArgumentException Wenn kein g&uuml;ltiger Simulationsmodus
     *         angegeben wurde.
     */
    public PhysicalEnvironment(CollisionHandler ch, int simulationType)
        throws IllegalArgumentException
    {
        this(ch, simulationType, Long.MAX_VALUE, 20, 1);
    }

    /**
     * Erstellt eine Instanz dieser Klasse mit den angegebenen Werten.
     * 
     * @param ch Der verwendete {@linkplain CollisionHandler}.
     * @param type <code>ALL_STEP</code> oder <code>SINGLE_STEP</code>.
     * @param maxDeltaTime Maximale L&auml;nge eines Zeitschritts 
     *        (maxDeltaTime > 0).
     * @param maxCalcTime Maximale Berechnungszeit im Modus
     *        <code>SINGLE_STEP</code> (maxCalcTime > 0).
     * @param epf Sto&szlig;faktor f&uuml;r Objekte der Szene, mit denen die
     *        registrierten K&ouml;rper zwar kollidieren k&ouml;nnen, die jedoch
     *        selbst keine Instanzen der Klasse
     *        {@linkplain DynamicTransformGroup} sind (0 >= epf >= 1).
     * @throws IllegalArgumentException Wenn einer der Parameter ung&uuml;ltig
     *         ist.
     */
    public PhysicalEnvironment(
        CollisionHandler ch,
        int type,
        long maxDeltaTime,
        long maxCalcTime,
        float epf)
        throws IllegalArgumentException
    {
        if ((type != ALL_STEP && type != SINGLE_STEP)
            || maxDeltaTime <= 0
            || maxCalcTime <= 0
            || epf < 0
            || epf > 1)
        {
            throw new IllegalArgumentException();
        }

        this.collisionHandler = ch;
        this.type = type;
        this.maxDeltaTime = maxDeltaTime;
        this.maxCalculationTime = maxCalcTime * 1000000;
        this.defaultEnergyPushFactor = epf;

        dynamicObjectList = new ArrayList();
        body1 = new PhysicalBody();
        body2 = new PhysicalBody();
        collisionList = new ArrayList();
        lastTime = 0;

        deltaImpulse = new Vector3f();
        deltaVP1 = new Vector3f();
        deltaVP2 = new Vector3f();
        tmpTransform = new Transform3D();
        tmpVec = new Vector3f();
    }

    /**
     * Setzt den Sto&szlig;faktor f&uuml;r Objekte der Szene, mit denen die
     * registrierten K&ouml;rper zwar kollidieren k&ouml;nnen, die jedoch selbst
     * keine Instanzen der Klasse {@linkplain DynamicTransformGroup} sind.
     * 
     * @param value Ein Wert zwischen 0 und 1.
     * @throws IllegalArgumentException Wenn der Sto&szlig;faktor nicht zwischen
     *         0 und 1 liegt.
     */
    synchronized public void setDefaultEnergyPushFactor(float value)
        throws IllegalArgumentException
    {
        if (value < 0f || value > 1f)
            throw new IllegalArgumentException("Wert muss zwischen einschlielich 0 und 1 liegen.");

        this.defaultEnergyPushFactor = value;
    }

    /**
     * Liefert den Sto&szlig;faktor f&uuml;r Objekte der Szene, mit denen die
     * registrierten K&ouml;rper zwar kollidieren k&ouml;nnen, die jedoch selbst
     * keine Instanzen der Klasse {@linkplain DynamicTransformGroup} sind,
     * zur&uuml;ck.
     * 
     * @return Der Sto&szlig;faktor.
     */
    public float getDefaultEnergyPushFactor()
    {
        return defaultEnergyPushFactor;
    }

    /**
     * Registriert eine {@linkplain DynamicTransformGroup} f&uuml;r die
     * Simulation. 
     * 
     * @param dynTraGrp Die zu registrierende
     *        {@linkplain DynamicTransformGroup}.
     * @return <code>true</code>, wenn der K&ouml;rper erfolgreich registriert
     *         wurde, sonst <code>false</code>.
     */
    synchronized public boolean addDynamicTransformGroup(DynamicTransformGroup dynTraGrp)
    {
        if (dynamicObjectList.contains(dynTraGrp) == false)
            return dynamicObjectList.add(dynTraGrp);
        else
            return false;
    }

    /**
     * Entfernt einen K&ouml;rper aus der Simulation.
     * 
     * @param dynTraGrp Der zu entfernende K&ouml;rper.
     * @return <code>true</code>, wenn der K&ouml;rper erfolgreich entfernt
     *         wurde, sonst <code>false</code>.
     */
    synchronized public boolean removeDynamicTransformGroup(DynamicTransformGroup dynTraGrp)
    {
        return dynamicObjectList.remove(dynTraGrp);
    }

    //  Nach Kurtenbach_98
    private boolean calculateImpulseChanging(
        Node object1,
        Node object2,
        Vector3f collisionNormal,
        Point3f collisionPoint,
        Vector3f deltaImpuls)
    {
        assert object1 != null;

        // Koerper 1 belegen
         ((DynamicTransformGroup) object1).getSpeed(body1.speed);
        ((DynamicTransformGroup) object1).getAngleSpeed(body1.angleSpeed);
        body1.mass = ((DynamicTransformGroup) object1).getMass();
        body1.J = ((DynamicTransformGroup) object1).getI();
        ((DynamicTransformGroup) object1).getTransform(tmpTransform);
        tmpTransform.get(body1.collisionRadius);
        body1.collisionRadius.sub(collisionPoint, body1.collisionRadius);
        body1.energyReturn =
            ((DynamicTransformGroup) object1).getEnergyPushFactor();

        if (object2 != null && object2 instanceof DynamicTransformGroup)
        {
            // Koerper 2 belegen
             ((DynamicTransformGroup) object2).getSpeed(body2.speed);
            ((DynamicTransformGroup) object2).getAngleSpeed(body2.angleSpeed);
            body2.mass = ((DynamicTransformGroup) object2).getMass();
            body2.J = ((DynamicTransformGroup) object2).getI();
            ((DynamicTransformGroup) object2).getTransform(tmpTransform);
            tmpTransform.get(body2.collisionRadius);
            body2.collisionRadius.sub(collisionPoint, body2.collisionRadius);
            body2.energyReturn =
                ((DynamicTransformGroup) object2).getEnergyPushFactor();

        } else
        {
            body2.angleSpeed.x = body2.angleSpeed.y = body2.angleSpeed.z = 0;
            body2.speed.x = body2.speed.y = body2.speed.z = 0;
            body2.collisionRadius.x =
                body2.collisionRadius.y = body2.collisionRadius.z = 0;
            body2.mass = Float.POSITIVE_INFINITY;
            body2.J = Float.POSITIVE_INFINITY;
            body2.energyReturn = defaultEnergyPushFactor;
        }

        float A, B, scalarDeltaP;

        deltaVP2.cross(body2.angleSpeed, body2.collisionRadius);
        deltaVP2.add(body2.speed);
        deltaVP1.cross(body1.angleSpeed, body1.collisionRadius);
        deltaVP1.add(body1.speed);
        deltaVP1.sub(deltaVP2);

        A = deltaVP1.dot(collisionNormal);

        // Verringert sich der Abstand zwischen den Koerpern?
        if (A > 0)
        {
            tmpVec.cross(body1.collisionRadius, collisionNormal);
            B = tmpVec.lengthSquared() / body1.J;
            tmpVec.cross(body2.collisionRadius, collisionNormal);
            B = B + tmpVec.lengthSquared() / body2.J;
            B = B + 1f / body1.mass + 1f / body2.mass;

            scalarDeltaP = -A / B;
            if (Math
                .abs(
                    body1.speed.dot(collisionNormal)
                        + body2.speed.dot(collisionNormal))
                >= 0.9)
                scalarDeltaP =
                    (1f + body1.energyReturn * body2.energyReturn)
                        * scalarDeltaP;
            else
                scalarDeltaP = 2 * scalarDeltaP;

            deltaImpuls.set(collisionNormal);
            deltaImpuls.scale(scalarDeltaP);

            return true;
        } else
        {
            return false;
        }
    }

    private long getDeltaTime()
    {
        long delta;
        long currentTime = J3DTimer.getValue();

        if (lastTime == 0)
        {
            // Zeit initialisieren
            lastTime = J3DTimer.getValue();
        }

        delta = (currentTime - lastTime) / 1000000;
        lastTime = currentTime;

        if (delta > maxDeltaTime)
        {
            return maxDeltaTime;
        }

        return delta;
    }

    private void manageCollisions()
    {
        collisionHandler.getAllCollisions(
            this.dynamicObjectList,
            collisionList);
        for (int i = 0; i < collisionList.size(); ++i)
        {
            Collision collision = (Collision) collisionList.get(i);

            if (calculateImpulseChanging(collision.node1,
                collision.node2,
                collision.normal,
                collision.point,
                deltaImpulse)
                == false)
            {
                continue;
            }

            ((DynamicTransformGroup) collision.node1).getTransform(
                tmpTransform);
            tmpTransform.get(tmpVec);
            tmpVec.sub(collision.point, tmpVec);
            ((DynamicTransformGroup) collision.node1).setDeltaMomentum(
                deltaImpulse,
                tmpVec);

            if (collision.node2 != null
                && collision.node2 instanceof DynamicTransformGroup)
            {
                if (((DynamicTransformGroup) collision.node2).isSleep())
                     ((DynamicTransformGroup) collision.node2).setSleep(false);
                deltaImpulse.negate();
                ((DynamicTransformGroup) collision.node2).getTransform(
                    tmpTransform);
                tmpTransform.get(tmpVec);
                tmpVec.sub(collision.point, tmpVec);
                ((DynamicTransformGroup) collision.node2).setDeltaMomentum(
                    deltaImpulse,
                    tmpVec);
            }
        }
    }

    private void moveAllObjects(long delta)
    {
        DynamicTransformGroup dynTrGrp;
        for (Iterator objIt = dynamicObjectList.iterator(); objIt.hasNext();)
        {
            dynTrGrp = ((DynamicTransformGroup) objIt.next());
            if (dynTrGrp.isSleep() == false)
                dynTrGrp.nextStep(delta);
        }
    }

    /**
     * Setzt die maximale Zeitspanne, die f&uuml;r einen Simulationsschritt
     * verwendet werden soll.
     * 
     * @param maxValue Die maximale Zeitspanne.
     * @throws IllegalArgumentException Wenn der angegebene Wert kleiner gleich
     *         0 ist.
     */
    synchronized public void setMaxDeltaTime(long maxValue)
        throws IllegalArgumentException
    {
        if (maxValue <= 0)
            throw new IllegalArgumentException("Wert muss grer 0 sein.");
        this.maxDeltaTime = maxValue;
    }

    /**
     * Gibt die maxmale Zeitspanne zur&uuml;ck, die f&uuml;r einen
     * Simulationsschritt verwendet wird. 
     * 
     * @return Die maximale Zeitspanne.
     */
    public long getMaxDeltaTime()
    {
        return maxDeltaTime;
    }

    /**
     * Setzt die maximale Berechnungszeit im Modus <code>SINGLE_STEP</code>.
     * 
     * @param value Die maximale Berechnungsdauer.
     */
    synchronized public void setMaxCalculationTime(long value)
    {
        if (value <= 0)
            throw new IllegalArgumentException("Wert muss grer 0 sein.");
        this.maxCalculationTime = value * 1000000;
    }

    /**
     * Gibt die maximale Berechnungszeit im Modus <code>SINGLE_STEP</code>
     * zur&uuml;ck.
     * 
     * @return Die maximale Berechnungsdauer.
     */
    public long getMaxCalculationTime()
    {
        return this.maxCalculationTime / 1000000;
    }

    private void nextStepSingle()
    {
        // DeltaTime berechnen
        long startTime = J3DTimer.getValue();
        long delta = 0;

        DynamicTransformGroup dynTrGrp;
        int length = dynamicObjectList.size();
        int lastIndex = single_stepCurrentIndex;
        do
        {
            if (single_stepCurrentIndex == lastIndex)
            {
                delta = getDeltaTime();
            }
            if (single_stepCurrentIndex >= length)
            {
                single_stepCurrentIndex = single_stepCurrentIndex - length;
            }

            dynTrGrp =
                (DynamicTransformGroup) dynamicObjectList.get(
                    single_stepCurrentIndex);

            // Objekt bewegen
            if (dynTrGrp.isSleep() == false)
            {
                dynTrGrp.nextStep(delta);

                // Kollisionen dieses Objekts ermitteln
                collisionHandler.getCollisions(dynTrGrp, collisionList);
                for (int i = 0; i < collisionList.size(); ++i)
                {
                    Collision collision = (Collision) collisionList.get(i);

                    // Kollisionsreaktion
                    if (calculateImpulseChanging(collision.node1,
                        collision.node2,
                        collision.normal,
                        collision.point,
                        deltaImpulse)
                        == false)
                    {
                        continue;
                    }

                    ((DynamicTransformGroup) collision.node1).getTransform(
                        tmpTransform);
                    tmpTransform.get(tmpVec);
                    tmpVec.sub(collision.point, tmpVec);
                    ((DynamicTransformGroup) collision.node1).setDeltaMomentum(
                        deltaImpulse,
                        tmpVec);

                    if (collision.node2 != null
                        && collision.node2 instanceof DynamicTransformGroup)
                    {
                        if (((DynamicTransformGroup) collision.node2)
                            .isSleep())
                            ((DynamicTransformGroup) collision.node2).setSleep(
                                false);
                        deltaImpulse.negate();
                        ((DynamicTransformGroup) collision.node2).getTransform(
                            tmpTransform);
                        tmpTransform.get(tmpVec);
                        tmpVec.sub(collision.point, tmpVec);
                        (
                            (
                                DynamicTransformGroup) collision
                                    .node2)
                                    .setDeltaMomentum(
                            deltaImpulse,
                            tmpVec);
                    }
                } // for (int i = 0; i < collisionList.size(); ++i)
            } // if (dynTrGrp.isSleep() == false)

            ++single_stepCurrentIndex;
        } while ((J3DTimer.getValue() - startTime) < maxCalculationTime);
    }

    private void nextStepBlock()
    {
        long delta = getDeltaTime();

        // Alle Objekte bewegen
        moveAllObjects(delta);

        // Kollisionen ermitteln und verarbeiten
        manageCollisions();
    }

    /**
     * F&uuml;hrt einen Simulationsschritt durch. Je nach angegebenen
     * Simulationsmodus werden alle K&ouml;rper gleichzeitig bewegt und auf
     * Kollisionen gepr&uuml;ft oder jeweils einzeln hintereinander.
     * <p/>
     * Zur Berechnung der Bewegung wird die seit dem letzten Aufruf der Methode
     * vergangene Zeit herangezogen. &Uuml;bersteigt diese jedoch
     * <code>maxDeltaTime</code>, so wird <code>maxDeltaTime</code> genutzt.
     */
    synchronized public void nextStep()
    {
        if (type == PhysicalEnvironment.ALL_STEP)
            nextStepBlock();
        else
            nextStepSingle();
    }

    private class PhysicalBody
    {
        public Vector3f speed;
        public Vector3f angleSpeed;
        public float mass;
        public float J;
        public Vector3f collisionRadius;
        public float energyReturn;

        public PhysicalBody()
        {
            speed = new Vector3f();
            angleSpeed = new Vector3f();
            collisionRadius = new Vector3f();
        }
    }
}
