package physics.geometry;
import java.io.Serializable;

import javax.vecmath.Point3f;
import javax.vecmath.Vector3f;
import javax.vecmath.Vector4f;

/*
 * Created on 25.04.2004
 */

/**
 * Repr&auml;sentiert ein Dreieck im Raum. Auf die Punkte des Dreiecks kann
 * direkt zugegriffen werden. Sollten die Punkte direkt gesetzt werden, muss
 * vor Aufruf der Methode
 * {@linkplain #collided(Polygon, Point3f, Point3f, Vector3f)}, auch wenn dieses
 * Objekt nur &uuml;bergeben wird, die Ebenengleichung des Polygons mit der
 * Methode {@linkplain #calculatePlaneNormalForm()} berechnet werden.
 * 
 * @author Martin Hedler
 * @version 1.0 (05/04/25)
 */
public class Polygon implements Serializable
{
    public Point3f a, b, c;
    
    /** Normale der Dreiecksfl&auml;che. */
    public Vector3f normal;
    
    /** Normalenform der Ebenen des Dreiecks. */
    private Vector4f planeNormalForm;

    /**
     * Erstellt ein Polygon mit den angegeben Punkten. Die Werte werden kopiert.
     * Die Ebenengleichung wird berechnet.
     * 
     * @param a Punkt des Dreiecks.
     * @param b Punkt des Dreiecks.
     * @param c Punkt des Dreiecks.
     */
    public Polygon(Point3f a, Point3f b, Point3f c)
    {
        this.a = new Point3f(a);
        this.b = new Point3f(b);
        this.c = new Point3f(c);

        this.normal = new Vector3f();

        planeNormalForm = new Vector4f();
        
        // Ebenengleichung dieses Polygons ermitteln
        calculatePlaneNormalForm();
    }

    /**
     * Erstellt ein Dreieck, dessen Eckpunkte alle im Nullpunkt liegen.
     */
    public Polygon()
    {
        this(new Point3f(), new Point3f(), new Point3f());
    }

    /**
     * Setzt die Punkte des Dreiecks und berechnet die Ebenengleichung.
     * 
     * @param a Punkt des Dreiecks.
     * @param b Punkt des Dreiecks.
     * @param c Punkt des Dreiecks.
     */
    public void set(Point3f a, Point3f b, Point3f c)
    {
        this.a.set(a);
        this.b.set(b);
        this.c.set(c);

        // Ebenengleichung dieses Polygons ermitteln
        calculatePlaneNormalForm();
    }

    /*
     * (non-Javadoc)
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
        return a + ", " + b + ", " + c;
    }

    /**
     * Errechnet zu einem angegebenen Polygon die Ebenengleichung der Ebene,
     * in der das Polygon liegt.
     */
    public void calculatePlaneNormalForm()
    {
        float x1, y1, z1;
        float x2, y2, z2;
        float xn, yn, zn;

        x1 = (float) (b.x - c.x);
        y1 = (float) (b.y - c.y);
        z1 = (float) (b.z - c.z);

        x2 = (float) (a.x - c.x);
        y2 = (float) (a.y - c.y);
        z2 = (float) (a.z - c.z);

        xn = y1 * z2 - z1 * y2;
        yn = z1 * x2 - x1 * z2;
        zn = x1 * y2 - y1 * x2;

        planeNormalForm.x = xn;
        planeNormalForm.y = yn;
        planeNormalForm.z = zn;
        planeNormalForm.w = (float) (-xn * a.x - yn * a.y - zn * a.z);
    }

    /**
     * &Uuml;berpr&uuml;ft, ob das &uuml;bergebene Polygon die Ebene, in der
     * <code>this</code> liegt, schneidet. Sollte dies der Fall sein, so wird
     * in <code>basePoint</code> der Punkt von <code>other</code> gespeichert,
     * der sich auf einer anderen Seite befindet, als die beiden anderen. Diese
     * werden in den Parametern <code>dirPoint1</code> und
     * <code>dirPoint2</code> gespeichert. 
     * 
     * @param other Das Polygon, welches auf Schnitt gepr&uuml;ft werden soll.
     * @param basePoint Hier wird der Punkt gespeichert, der sich alleine auf
     *        einer Seite der Ebene befindet.
     * @param dirPoint1 Hier wird einer der Punkte gespeichert, der sich mit
     *        einem anderen Punkt auf einer Seite der Ebene befindet.
     * @param dirPoint2 Hier wird einer der Punkte gespeichert, der sich mit
     *        einem anderen Punkt auf einer Seite der Ebene befindet.
     * @return <code>true</code>, wenn <code>other</code> die Ebene von
     *         <code>this</code> schneidet, sonst <code>false</code>.
     */
    private boolean isPlaneCrossing(
        Polygon other,
        Point3f basePoint,
        Point3f dirPoint1,
        Point3f dirPoint2)
    {
        float planeDistanceA, planeDistanceB, planeDistanceC;

        // Punkte des zweiten Polygons in Ebenengleichung einsetzen und jeweilige
        // Entfernung berechnen
        planeDistanceA =
            (float) (planeNormalForm.x * other.a.x
                + planeNormalForm.y * other.a.y
                + planeNormalForm.z * other.a.z
                + planeNormalForm.w);
        planeDistanceB =
            (float) (planeNormalForm.x * other.b.x
                + planeNormalForm.y * other.b.y
                + planeNormalForm.z * other.b.z
                + planeNormalForm.w);
        planeDistanceC =
            (float) (planeNormalForm.x * other.c.x
                + planeNormalForm.y * other.c.y
                + planeNormalForm.z * other.c.z
                + planeNormalForm.w);

        // Nur die Vorzeichen sind von Bedeutung
        planeDistanceA =
            (planeDistanceA > 0) ? +1 : ((planeDistanceA < 0) ? -1 : 0);
        planeDistanceB =
            (planeDistanceB > 0) ? +1 : ((planeDistanceB < 0) ? -1 : 0);
        planeDistanceC =
            (planeDistanceC > 0) ? +1 : ((planeDistanceC < 0) ? -1 : 0);

        // Sind alle Punkte auf der selben Seite?
        if (planeDistanceA == planeDistanceB
            && planeDistanceA == planeDistanceC)
        {
            // Polygone schneiden sich nicht
            return false;
        } else
            if (planeDistanceA != planeDistanceB
                && planeDistanceA != planeDistanceC
                && (planeDistanceB == planeDistanceC || planeDistanceA != 0))
            {
                // A steht einzeln
                basePoint.set(other.a);
                dirPoint1.set(other.c);
                dirPoint2.set(other.b);
            } else
                if (planeDistanceB != planeDistanceA
                    && planeDistanceB != planeDistanceC
                    && (planeDistanceA == planeDistanceC || planeDistanceB != 0))
                {
                    // B steht einzeln
                    basePoint.set(other.b);
                    dirPoint1.set(other.c);
                    dirPoint2.set(other.a);
                } else
                    if (planeDistanceC != planeDistanceA
                        && planeDistanceC != planeDistanceB
                        && (planeDistanceA == planeDistanceB
                            || planeDistanceC != 0))
                    {
                        // C steht einzeln
                        basePoint.set(other.c);
                        dirPoint1.set(other.a);
                        dirPoint2.set(other.b);
                    } else
                    {
                        assert false;
                    }
        return true;
    }

    /* Temoraere Variablen */

    private Point3f basePointOther = new Point3f(),
        dirPoint1Other = new Point3f(),
        dirPoint2Other = new Point3f();
    private Point3f basePointThis = new Point3f(),
        dirPoint1This = new Point3f(),
        dirPoint2This = new Point3f();
    private Point3f crossPoint1Other = new Point3f(),
        crossPoint2Other = new Point3f();
    private Point3f crossPoint1This = new Point3f(),
        crossPoint2This = new Point3f();
    private Vector3f lineDirection = new Vector3f();

    /**
     * Ermittelt, ob dieses Polygon mit dem &uuml;bergebenen Polygon kollidiert.
     * Bei einer Kollision wird <code>true</code> zur&uuml;ck geliefert.
     * Algorithmus nach <a href="http://www.scherfgen-software.net">David
     * Scherfgen</a>.<br/>
     * <i>Achtung:</i> Der Algorithmus erkennt keine &Uuml;berschneidungen.
     * @param other Das Polygon, das auf Schnitt mit diesem Polygon gepr&uml;ft
     *        werden soll.
     * @param startPoint Punkt in dem der Startpunkt der Schnittstrecke
     *        gespeichert wird.
     * @param endPoint Punkt in dem der Endpunkt der Schnittstrecke
     *        gespeichert wird.
     * @param collisionNormal Vektor, in dem ie Kollisionsnormale
     *        gespeichert wird.
     * @return <code>true</code> wenn sich die beiden Polygone schneiden,
     *         ansonsten <code>false</code>.
     */
    public boolean collided(
        Polygon other,
        Point3f startPoint,
        Point3f endPoint,
        Vector3f collisionNormal)
    {
        /*
         * Verwendet die globalen Variablen:
         * basePointOther,
         * dirPoint1Other,
         * dirPoint2Other,
         * basePointThis,
         * dirPoint1This,
         * dirPoint2This,
         * crossPoint1This,
         * crossPoint2This,
         * crossPoint1Other,
         * crossPoint2Other,
         * lineDirection
         */
        
        // Schneidet other nicht die Ebene von this?
        if (this
            .isPlaneCrossing(
                other,
                basePointOther,
                dirPoint1Other,
                dirPoint2Other)
            == false)
        {
            return false;
        }

        // Schneidet this nicht die Ebene von other?
        if (other
            .isPlaneCrossing(this, basePointThis, dirPoint1This, dirPoint2This)
            == false)
        {
            return false;
        }

        /* Wenn wir bis hier gekommen sind, dann schneiden die beiden Polygone jeweils
         * die Ebene des anderen Polygons.
         */

        // Schnittpunkte berechnen
        getCrossPoint(
            other.planeNormalForm,
            dirPoint1This,
            basePointThis,
            crossPoint1This);
        getCrossPoint(
            other.planeNormalForm,
            dirPoint2This,
            basePointThis,
            crossPoint2This);
        getCrossPoint(
            this.planeNormalForm,
            dirPoint1Other,
            basePointOther,
            crossPoint1Other);
        getCrossPoint(
            this.planeNormalForm,
            dirPoint2Other,
            basePointOther,
            crossPoint2Other);

        // Geradengleichung der Geraden ermitteln, auf der die Schnittpunkte liegen
        lineDirection.sub(crossPoint1This, crossPoint1Other);

        if (lineDirection.lengthSquared() == 0)
        {
            lineDirection.sub(crossPoint2This, crossPoint2Other);

            if (lineDirection.lengthSquared() == 0)
            {
                startPoint.set(crossPoint1This);
                endPoint.set(crossPoint2This);
                return true;
            }
        }

        lineDirection.normalize();

        float startIndexThis, endIndexThis, startIndexOther, endIndexOther;

        // Groessten Nenner ermitteln
        if (Math.abs(lineDirection.x) > Math.abs(lineDirection.y)
            && Math.abs(lineDirection.x) > Math.abs(lineDirection.z))
        {
            // Nenner = x
            startIndexThis = 0;
            endIndexThis =
                (crossPoint2This.x - crossPoint1This.x) / lineDirection.x;
            startIndexOther =
                (crossPoint1Other.x - crossPoint1This.x) / lineDirection.x;
            endIndexOther =
                (crossPoint2Other.x - crossPoint1This.x) / lineDirection.x;
        } else
            if (Math.abs(lineDirection.y) > Math.abs(lineDirection.z))
            {
                // Nenner = y
                startIndexThis = 0;
                endIndexThis =
                    (crossPoint2This.y - crossPoint1This.y) / lineDirection.y;
                startIndexOther =
                    (crossPoint1Other.y - crossPoint1This.y) / lineDirection.y;
                endIndexOther =
                    (crossPoint2Other.y - crossPoint1This.y) / lineDirection.y;
            } else
            {
                // Nenner = z
                startIndexThis = 0;
                endIndexThis =
                    (crossPoint2This.z - crossPoint1This.z) / lineDirection.z;
                startIndexOther =
                    (crossPoint1Other.z - crossPoint1This.z) / lineDirection.z;
                endIndexOther =
                    (crossPoint2Other.z - crossPoint1This.z) / lineDirection.z;
            }

        // Werte sortieren
        if (startIndexThis > endIndexThis)
        {
            float temp;
            temp = startIndexThis;
            startIndexThis = endIndexThis;
            endIndexThis = temp;

            Point3f tempV;
            tempV = crossPoint1This;
            crossPoint1This = crossPoint2This;
            crossPoint2This = tempV;
        }
        if (startIndexOther > endIndexOther)
        {
            float temp;
            temp = startIndexOther;
            startIndexOther = endIndexOther;
            endIndexOther = temp;

            Point3f tempV;
            tempV = crossPoint1Other;
            crossPoint1Other = crossPoint2Other;
            crossPoint2Other = tempV;
        }

        // Ueberschneidung ermitteln        
        if (startIndexThis >= startIndexOther && endIndexThis <= endIndexOther)
        {
            // Schnittstrecke von Polygon1 liegt in Polygon2
            startPoint.set(crossPoint1This);
            endPoint.set(crossPoint2This);
        } else
            if (startIndexOther >= startIndexThis
                && endIndexOther <= endIndexThis)
            {
                // Schnittstrecke von Polygon2 liegt in Polygon1
                startPoint.set(crossPoint1Other);
                endPoint.set(crossPoint2Other);
            } else
                if (startIndexThis <= startIndexOther
                    && endIndexThis >= startIndexOther)
                {
                    startPoint.set(crossPoint1Other);
                    endPoint.set(crossPoint2This);
                } else
                    if (startIndexThis <= endIndexOther
                        && endIndexThis >= endIndexOther)
                    {
                        startPoint.set(crossPoint1This);
                        endPoint.set(crossPoint2Other);
                    } else
                    {
                        return false;
                    }

        // Kollisionsnormale berechnen
        collisionNormal.scaleAdd(-1, other.normal, this.normal);
        collisionNormal.normalize();

        // Zeigen beide Normalen in die selbe Richtung?
        if (Float.isNaN(collisionNormal.x))
        {
            collisionNormal.set(this.normal);
        }

        return true;
    }

    /**
     * Diese Funktion &uuml;berpr&uuml;ft, ob die durch die beiden Punkte
     * <code>p1</code> und <code>p2</code> beschriebene Gerade die Ebene
     * <code>plane</code> schneidet.
     * Diese Funktion dient nur dem Aufruf in einer <code>assert</code>
     * Anweisung.
     * 
     * @param plane Normalenform der Ebenengleichung.
     *        <pre>
     *           E: x * xp + y * yp + z * zp + w = 0
     *        </pre>
     * @param p1 Punkt auf der Geraden
     * @param p2 Punkt auf der Geraden
     * @return <code>true</code>, wenn die Gerade die Ebene schneidet, sonst
     *         <code>false</code>.
     */
    private boolean assert_IsLinePlaneCrossing(Vector4f plane, Point3f p1, Point3f p2)
    {
        Vector3f lineDirVec = new Vector3f(p1);
        lineDirVec.sub(p2);
        Vector3f planeNormal = new Vector3f(plane.x, plane.y, plane.z);
        
        planeNormal.cross(planeNormal, lineDirVec);
        
        return planeNormal.lengthSquared() != 0;
    }

    /**
     * Ermittelt den Schnittpunkt einer Ebenen mit einer Geraden.
     * Dabei wird davon ausgegangen, dass sich diese beiden auch wirklich
     * schneiden. Die Gerade wird durch zwei Punkte angegeben.
     * 
     * @param plane Ebenengleichung in der Hessische Normalform.
     *        <pre>
     *           E: x * xp + y * yp + z * zp + w = 0
     *        </pre>
     * @param point1 Erster Punkt auf der Geraden.
     * @param point2 Zweiter Punkt auf der Geraden.
     * @param crossPoint Wird mit den Wertden des Schnittpunkts von Ebene und Geraden
     *                   belegt und darf deshalb nicht <code>null</code> sein.
     */
    private void getCrossPoint(
        Vector4f plane,
        Point3f point1,
        Point3f point2,
        Point3f crossPoint)
    {
        assert assert_IsLinePlaneCrossing(plane, point1, point2);

        // Richtungsvektor berechnen
        crossPoint.set(point1);
        crossPoint.sub(point2);

        // Parameter der Geradengleichung g : sP + k * dir berechnen
        float k =
            (
                - plane.w
                - plane.x * point2.x
                - plane.y * point2.y
                - plane.z * point2.z)
                / (plane.x * crossPoint.x
                    + plane.y * crossPoint.y
                    + plane.z * crossPoint.z);

        crossPoint.scale(k);
        crossPoint.add(point2);
    }
}
